@@ -1,8 +1,8 @@
 /* ldapmodrdn.c - generic program to modify an entry's RDN using LDAP */
-/* $OpenLDAP: pkg/ldap/clients/tools/ldapmodrdn.c,v 1.99.2.3 2003/12/01 00:58:22 kurt Exp $ */
+/* $OpenLDAP: pkg/ldap/clients/tools/ldapmodrdn.c,v 1.106.2.1 2005/01/20 18:03:45 kurt Exp $ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 1998-2003 The OpenLDAP Foundation.
+ * Copyright 1998-2005 The OpenLDAP Foundation.
  * Portions Copyright 1998-2003 Kurt D. Zeilenga.
  * Portions Copyright 1998-2001 Net Boolean Incorporated.
  * Portions Copyright 2001-2003 IBM Corporation.
@@ -89,7 +89,7 @@ usage( void )
 
 
 const char options[] = "rs:"
-	"cCd:D:e:f:h:H:IkKMnO:p:P:QR:U:vVw:WxX:y:Y:Z";
+	"cd:D:e:f:h:H:IkKMnO:p:P:QR:U:vVw:WxX:y:Y:Z";
 
 int
 handle_private_option( int i )
@@ -98,7 +98,7 @@ handle_private_option( int i )
 #if 0
 		int crit;
 		char *control, *cvalue;
-	case 'E': /* modrdn controls */
+	case 'E': /* modrdn extensions */
 		if( protocol == LDAP_VERSION2 ) {
 			fprintf( stderr, _("%s: -E incompatible with LDAPv%d\n"),
 				prog, version );
@@ -120,7 +120,7 @@ handle_private_option( int i )
 		if ( (cvalue = strchr( control, '=' )) != NULL ) {
 			*cvalue++ = '\0';
 		}
-		fprintf( stderr, _("Invalid modrdn control name: %s\n"), control );
+		fprintf( stderr, _("Invalid modrdn extension name: %s\n"), control );
 		usage();
 #endif
 
@@ -227,7 +227,7 @@ main(int argc, char **argv)
 	}
     }
 
-    ldap_unbind( ld );
+	ldap_unbind_ext( ld, NULL, NULL );
 
     return( retval );
 }
