@@ -1,8 +1,8 @@
 /* ldapsearch -- a tool for searching LDAP directories */
-/* $OpenLDAP: pkg/ldap/clients/tools/ldapsearch.c,v 1.170.2.6 2003/12/01 00:58:22 kurt Exp $ */
+/* $OpenLDAP: pkg/ldap/clients/tools/ldapsearch.c,v 1.207.2.4 2005/03/14 22:24:55 kurt Exp $ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 1998-2003 The OpenLDAP Foundation.
+ * Copyright 1998-2005 The OpenLDAP Foundation.
  * Portions Copyright 1998-2003 Kurt D. Zeilenga.
  * Portions Copyright 1998-2001 Net Boolean Incorporated.
  * Portions Copyright 2001-2003 IBM Corporation.
@@ -48,6 +48,10 @@
 #include <ac/errno.h>
 #include <sys/stat.h>
 
+#if defined(TEST_ABANDON) || defined(TEST_CANCEL)
+#include <ac/signal.h>
+#endif
+
 #ifdef HAVE_FCNTL_H
 #include <fcntl.h>
 #endif
@@ -79,6 +83,12 @@ static int sizelimit = -1;
 static char *def_tmpdir;
 static char *def_urlpre;
 
+#if defined(__CYGWIN__) || defined(__MINGW32__)
+/* Turn off commandline globbing, otherwise you cannot search for
+ * attribute '*'
+ */
+int _CRT_glob = 0;
+#endif
 
 void
 usage( void )
@@ -96,13 +106,13 @@ usage( void )
 	fprintf( stderr, _("  -a deref   one of never (default), always, search, or find\n"));
 	fprintf( stderr, _("  -A         retrieve attribute names only (no values)\n"));
 	fprintf( stderr, _("  -b basedn  base dn for search\n"));
-	fprintf( stderr, _("  -E [!]<ctrl>[=<ctrlparam>] search controls (! indicates criticality)\n"));
+	fprintf( stderr, _("  -E [!]<ext>[=<extparam>] search extensions (! indicates criticality)\n"));
 #ifdef LDAP_CONTROL_X_DOMAIN_SCOPE
 	fprintf( stderr, _("             [!]domainScope              (domain scope)\n"));
 #endif
 	fprintf( stderr, _("             [!]mv=<filter>              (matched values filter)\n"));
 #ifdef LDAP_CONTROL_PAGEDRESULTS
-	fprintf( stderr, _("             [!]pr=<size>                (paged results)\n"));
+	fprintf( stderr, _("             [!]pr=<size>[/prompt|noprompt]   (paged results/prompt)\n"));
 #endif
 #ifdef LDAP_CONTROL_SUBENTRIES
 	fprintf( stderr, _("             [!]subentries[=true|false]  (subentries)\n"));
@@ -110,18 +120,22 @@ usage( void )
 	fprintf( stderr, _("             [!]sync=ro[/<cookie>]            (LDAP Sync refreshOnly)\n"));
 	fprintf( stderr, _("                     rp[/<cookie>][/<slimit>] (LDAP Sync refreshAndPersist)\n"));
 	fprintf( stderr, _("  -F prefix  URL prefix for files (default: %s)\n"), def_urlpre);
-	fprintf( stderr, _("  -l limit   time limit (in seconds) for search\n"));
+	fprintf( stderr, _("  -l limit   time limit (in seconds, or \"none\" or \"max\") for search\n"));
 	fprintf( stderr, _("  -L         print responses in LDIFv1 format\n"));
 	fprintf( stderr, _("  -LL        print responses in LDIF format without comments\n"));
 	fprintf( stderr, _("  -LLL       print responses in LDIF format without comments\n"));
 	fprintf( stderr, _("             and version\n"));
+#ifdef LDAP_SCOPE_SUBORDINATE
+	fprintf( stderr, _("  -s scope   one of base, one, sub or children (search scope)\n"));
+#else /* ! LDAP_SCOPE_SUBORDINATE */
 	fprintf( stderr, _("  -s scope   one of base, one, or sub (search scope)\n"));
+#endif /* ! LDAP_SCOPE_SUBORDINATE */
 	fprintf( stderr, _("  -S attr    sort the results by attribute `attr'\n"));
 	fprintf( stderr, _("  -t         write binary values to files in temporary directory\n"));
 	fprintf( stderr, _("  -tt        write all values to files in temporary directory\n"));
 	fprintf( stderr, _("  -T path    write files to directory specified by path (default: %s)\n"), def_tmpdir);
 	fprintf( stderr, _("  -u         include User Friendly entry names in the output\n"));
-	fprintf( stderr, _("  -z limit   size limit (in entries) for search\n"));
+	fprintf( stderr, _("  -z limit   size limit (in entries, or \"none\" or \"max\") for search\n"));
 	tool_common_usage();
 	exit( EXIT_FAILURE );
 }
@@ -189,10 +203,11 @@ static int sync_slimit = -1;
 
 #ifdef LDAP_CONTROL_PAGEDRESULTS
 static int pagedResults = 0;
+static int pagePrompt = 1;
 static ber_int_t pageSize = 0;
 static ber_int_t entriesLeft = 0;
 static ber_int_t morePagedResults = 1;
-static struct berval cookie = { 0, NULL };
+static struct berval page_cookie = { 0, NULL };
 static int npagedresponses;
 static int npagedentries;
 static int npagedreferences;
@@ -225,30 +240,31 @@ const char options[] = "a:Ab:E:F:l:Ls:S:tT:uz:"
 int
 handle_private_option( int i )
 {
-	int crit;
-	char *control, *cvalue;
+	int crit, ival;
+	char *control, *cvalue, *next;
 	switch ( i ) {
 	case 'a':	/* set alias deref option */
 		if ( strcasecmp( optarg, "never" ) == 0 ) {
-		deref = LDAP_DEREF_NEVER;
+			deref = LDAP_DEREF_NEVER;
 		} else if ( strncasecmp( optarg, "search", sizeof("search")-1 ) == 0 ) {
-		deref = LDAP_DEREF_SEARCHING;
+			deref = LDAP_DEREF_SEARCHING;
 		} else if ( strncasecmp( optarg, "find", sizeof("find")-1 ) == 0 ) {
-		deref = LDAP_DEREF_FINDING;
+			deref = LDAP_DEREF_FINDING;
 		} else if ( strcasecmp( optarg, "always" ) == 0 ) {
-		deref = LDAP_DEREF_ALWAYS;
+			deref = LDAP_DEREF_ALWAYS;
 		} else {
-		fprintf( stderr, _("alias deref should be never, search, find, or always\n") );
-		usage();
+			fprintf( stderr,
+				_("alias deref should be never, search, find, or always\n") );
+			usage();
 		}
 		break;
 	case 'A':	/* retrieve attribute names only -- no values */
 		++attrsonly;
 		break;
 	case 'b': /* search base */
-		base = strdup( optarg );
+		base = ber_strdup( optarg );
 		break;
-	case 'E': /* search controls */
+	case 'E': /* search extensions */
 		if( protocol == LDAP_VERSION2 ) {
 			fprintf( stderr, _("%s: -E incompatible with LDAPv%d\n"),
 				prog, protocol );
@@ -266,7 +282,7 @@ handle_private_option( int i )
 			optarg++;
 		}
 
-		control = strdup( optarg );
+		control = ber_strdup( optarg );
 		if ( (cvalue = strchr( control, '=' )) != NULL ) {
 			*cvalue++ = '\0';
 		}
@@ -274,7 +290,8 @@ handle_private_option( int i )
 		if ( strcasecmp( control, "mv" ) == 0 ) {
 			/* ValuesReturnFilter control */
 			if( valuesReturnFilter ) {
-				fprintf( stderr, _("ValuesReturnFilter previously specified\n"));
+				fprintf( stderr,
+					_("ValuesReturnFilter previously specified\n"));
 				exit( EXIT_FAILURE );
 			}
 			valuesReturnFilter= 1 + crit;
@@ -293,15 +310,39 @@ handle_private_option( int i )
 			int num, tmp;
 			/* PagedResults control */
 			if ( pagedResults != 0 ) {
-				fprintf( stderr, _("PagedResultsControl previously specified\n") );
+				fprintf( stderr,
+					_("PagedResultsControl previously specified\n") );
 				exit( EXIT_FAILURE );
 			}
-			
-			num = sscanf( cvalue, "%d", &tmp );
-			if ( num != 1 ) {
-				fprintf( stderr, _("Invalid value for PagedResultsControl, %s.\n"), cvalue);
-				exit( EXIT_FAILURE );
 
+			if( cvalue != NULL ) {
+				char *promptp;
+
+				promptp = strchr( cvalue, '/' );
+				if ( promptp != NULL ) {
+					*promptp++ = '\0';
+					if ( strcasecmp( promptp, "prompt" ) == 0 ) {
+						pagePrompt = 1;
+					} else if ( strcasecmp( promptp, "noprompt" ) == 0) {
+						pagePrompt = 0;
+					} else {
+						fprintf( stderr,
+							_("Invalid value for PagedResultsControl,"
+							" %s/%s.\n"), cvalue, promptp );
+						exit( EXIT_FAILURE );
+					}
+				}
+				num = sscanf( cvalue, "%d", &tmp );
+				if ( num != 1 ) {
+					fprintf( stderr,
+						_("Invalid value for PagedResultsControl, %s.\n"),
+						cvalue );
+					exit( EXIT_FAILURE );
+				}
+			} else {
+				fprintf( stderr, _("Invalid value for PagedResultsControl.\n"),
+					cvalue);
+				exit( EXIT_FAILURE );
 			}
 			pageSize = (ber_int_t) tmp;
 			pagedResults = 1 + crit;
@@ -325,7 +366,8 @@ handle_private_option( int i )
 #ifdef LDAP_CONTROL_SUBENTRIES
 		} else if ( strcasecmp( control, "subentries" ) == 0 ) {
 			if( subentries ) {
-				fprintf( stderr, _("subentries control previously specified\n"));
+				fprintf( stderr,
+					_("subentries control previously specified\n"));
 				exit( EXIT_FAILURE );
 			}
 			if( cvalue == NULL || strcasecmp( cvalue, "true") == 0 ) {
@@ -341,16 +383,15 @@ handle_private_option( int i )
 			if( crit ) subentries *= -1;
 #endif
 
-	} else if ( strcasecmp( control, "sync" ) == 0 ) {
+		} else if ( strcasecmp( control, "sync" ) == 0 ) {
 			char *cookiep;
 			char *slimitp;
 			if ( ldapsync ) {
-				fprintf( stderr, _("ldap sync control previously specified\n") );
+				fprintf( stderr, _("sync control previously specified\n") );
 				exit( EXIT_FAILURE );
 			}
 			if ( cvalue == NULL ) {
-				fprintf( stderr,
-					_("missing specification of ldap sync control\n"));
+				fprintf( stderr, _("missing specification of sync control\n"));
 				exit( EXIT_FAILURE );
 			}
 			if ( strncasecmp( cvalue, "ro", 2 ) == 0 ) {
@@ -375,18 +416,24 @@ handle_private_option( int i )
 				}
 				if ( cookiep != NULL && *cookiep != '\0' )
 					ber_str2bv( cookiep, 0, 0, &sync_cookie );
-				if ( slimitp != NULL && *slimitp != '\0' )
-					sync_slimit = atoi( slimitp );
+				if ( slimitp != NULL && *slimitp != '\0' ) {
+					ival = strtol( slimitp, &next, 10 );
+					if ( next == NULL || next[0] != '\0' ) {
+						fprintf( stderr, _("Unable to parse sync control value \"%s\"\n"), slimitp );
+						exit( EXIT_FAILURE );
+					}
+					sync_slimit = ival;
+				}
 			} else {
-				fprintf( stderr,
-					_("ldap sync control value \"%s\" invalid\n"),
+				fprintf( stderr, _("sync control value \"%s\" invalid\n"),
 					cvalue );
 				exit( EXIT_FAILURE );
 			}
 			if ( crit ) ldapsync *= -1;
 
 		} else {
-			fprintf( stderr, _("Invalid control name: %s\n"), control );
+			fprintf( stderr, _("Invalid search extension name: %s\n"),
+				control );
 			usage();
 		}
 		break;
@@ -395,8 +442,22 @@ handle_private_option( int i )
 		urlpre = strdup( optarg );
 		break;
 	case 'l':	/* time limit */
-		timelimit = atoi( optarg );
-		if( timelimit < 0 ) {
+		if ( strcasecmp( optarg, "none" ) == 0 ) {
+			timelimit = 0;
+
+		} else if ( strcasecmp( optarg, "max" ) == 0 ) {
+			timelimit = LDAP_MAXINT;
+
+		} else {
+			ival = strtol( optarg, &next, 10 );
+			if ( next == NULL || next[0] != '\0' ) {
+				fprintf( stderr,
+					_("Unable to parse time limit \"%s\"\n"), optarg );
+				exit( EXIT_FAILURE );
+			}
+			timelimit = ival;
+		}
+		if( timelimit < 0 || timelimit > LDAP_MAXINT ) {
 			fprintf( stderr, _("%s: invalid timelimit (%d) specified\n"),
 				prog, timelimit );
 			exit( EXIT_FAILURE );
@@ -406,15 +467,21 @@ handle_private_option( int i )
 		++ldif;
 		break;
 	case 's':	/* search scope */
-		if ( strcasecmp( optarg, "base" ) == 0 ) {
-		scope = LDAP_SCOPE_BASE;
+		if ( strncasecmp( optarg, "base", sizeof("base"-1) ) == 0 ) {
+			scope = LDAP_SCOPE_BASE;
 		} else if ( strncasecmp( optarg, "one", sizeof("one")-1 ) == 0 ) {
-		scope = LDAP_SCOPE_ONELEVEL;
+			scope = LDAP_SCOPE_ONELEVEL;
+#ifdef LDAP_SCOPE_SUBORDINATE
+		} else if (( strcasecmp( optarg, "subordinate" ) == 0 )
+			|| ( strcasecmp( optarg, "children" ) == 0 ))
+		{
+			scope = LDAP_SCOPE_SUBORDINATE;
+#endif
 		} else if ( strncasecmp( optarg, "sub", sizeof("sub")-1 ) == 0 ) {
-		scope = LDAP_SCOPE_SUBTREE;
+			scope = LDAP_SCOPE_SUBTREE;
 		} else {
-		fprintf( stderr, _("scope should be base, one, or sub\n") );
-		usage();
+			fprintf( stderr, _("scope should be base, one, or sub\n") );
+			usage();
 		}
 		break;
 	case 'S':	/* sort attribute */
@@ -431,7 +498,26 @@ handle_private_option( int i )
 		++includeufn;
 		break;
 	case 'z':	/* size limit */
-		sizelimit = atoi( optarg );
+		if ( strcasecmp( optarg, "none" ) == 0 ) {
+			sizelimit = 0;
+
+		} else if ( strcasecmp( optarg, "max" ) == 0 ) {
+			sizelimit = LDAP_MAXINT;
+
+		} else {
+			ival = strtol( optarg, &next, 10 );
+			if ( next == NULL || next[0] != '\0' ) {
+				fprintf( stderr,
+					_("Unable to parse size limit \"%s\"\n"), optarg );
+				exit( EXIT_FAILURE );
+			}
+			sizelimit = ival;
+		}
+		if( sizelimit < 0 || sizelimit > LDAP_MAXINT ) {
+			fprintf( stderr, _("%s: invalid sizelimit (%d) specified\n"),
+				prog, sizelimit );
+			exit( EXIT_FAILURE );
+		}
 		break;
 	default:
 		return 0;
@@ -444,25 +530,39 @@ static void
 private_conn_setup( LDAP *ld )
 {
 	if (deref != -1 &&
-		ldap_set_option( ld, LDAP_OPT_DEREF, (void *) &deref ) != LDAP_OPT_SUCCESS )
+		ldap_set_option( ld, LDAP_OPT_DEREF, (void *) &deref )
+			!= LDAP_OPT_SUCCESS )
 	{
 		fprintf( stderr, _("Could not set LDAP_OPT_DEREF %d\n"), deref );
 		exit( EXIT_FAILURE );
 	}
-	if (timelimit != -1 &&
-		ldap_set_option( ld, LDAP_OPT_TIMELIMIT, (void *) &timelimit ) != LDAP_OPT_SUCCESS )
+	if (timelimit > 0 &&
+		ldap_set_option( ld, LDAP_OPT_TIMELIMIT, (void *) &timelimit )
+			!= LDAP_OPT_SUCCESS )
 	{
-		fprintf( stderr, _("Could not set LDAP_OPT_TIMELIMIT %d\n"), timelimit );
+		fprintf( stderr,
+			_("Could not set LDAP_OPT_TIMELIMIT %d\n"), timelimit );
 		exit( EXIT_FAILURE );
 	}
-	if (sizelimit != -1 &&
-		ldap_set_option( ld, LDAP_OPT_SIZELIMIT, (void *) &sizelimit ) != LDAP_OPT_SUCCESS )
+	if (sizelimit > 0 &&
+		ldap_set_option( ld, LDAP_OPT_SIZELIMIT, (void *) &sizelimit )
+			!= LDAP_OPT_SUCCESS )
 	{
-		fprintf( stderr, _("Could not set LDAP_OPT_SIZELIMIT %d\n"), sizelimit );
+		fprintf( stderr,
+			_("Could not set LDAP_OPT_SIZELIMIT %d\n"), sizelimit );
 		exit( EXIT_FAILURE );
 	}
 }
 
+#if defined(TEST_ABANDON) || defined(TEST_CANCEL)
+static int gotintr;
+
+RETSIGTYPE
+do_sig( int sig )
+{
+	gotintr = 1;
+}
+#endif
 
 int
 main( int argc, char **argv )
@@ -485,8 +585,6 @@ main( int argc, char **argv )
 
 	prog = lutil_progname( "ldapsearch", argc, argv );
 
-	lutil_log_initialize(argc, argv);
-
 	if((def_tmpdir = getenv("TMPDIR")) == NULL &&
 	   (def_tmpdir = getenv("TMP")) == NULL &&
 	   (def_tmpdir = getenv("TEMP")) == NULL )
@@ -524,6 +622,10 @@ main( int argc, char **argv )
 		attrs = &argv[optind];
 	}
 
+#if defined(TEST_ABANDON) || defined(TEST_CANCEL)
+	SIGNAL( SIGINT, do_sig );
+#endif
+
 	if ( infile != NULL ) {
 		if ( infile[0] == '-' && infile[1] == '\0' ) {
 			fp = stdin;
@@ -577,33 +679,36 @@ getNextPage:
 		|| domainScope
 #endif
 #ifdef LDAP_CONTROL_PAGEDRESULTS
-		|| pageSize
+		|| pagedResults
 #endif
+#ifdef LDAP_CONTROL_X_CHAINING_BEHAVIOR
+		|| chaining
+#endif /* LDAP_CONTROL_X_CHAINING_BEHAVIOR */
 		|| ldapsync
 		|| subentries || valuesReturnFilter )
 	{
 		int err;
 		int i=0;
-		LDAPControl c[6];
+		LDAPControl c[10];
 
 #ifdef LDAP_CONTROL_X_DOMAIN_SCOPE
-	if ( domainScope ) {
-		c[i].ldctl_oid = LDAP_CONTROL_X_DOMAIN_SCOPE;
-		c[i].ldctl_value.bv_val = NULL;
-		c[i].ldctl_value.bv_len = 0;
-		c[i].ldctl_iscritical = domainScope > 1;
-		i++;
-	}
+		if ( domainScope ) {
+			c[i].ldctl_oid = LDAP_CONTROL_X_DOMAIN_SCOPE;
+			c[i].ldctl_value.bv_val = NULL;
+			c[i].ldctl_value.bv_len = 0;
+			c[i].ldctl_iscritical = domainScope > 1;
+			i++;
+		}
 #endif
 
 #ifdef LDAP_CONTROL_SUBENTRIES
 		if ( subentries ) {
-	        if (( seber = ber_alloc_t(LBER_USE_DER)) == NULL ) {
+			if (( seber = ber_alloc_t(LBER_USE_DER)) == NULL ) {
 				return EXIT_FAILURE;
 			}
 
-			err = ber_printf( seber, "{b}", abs(subentries) == 1 ? 0 : 1 );
-	    	if ( err == -1 ) {
+			err = ber_printf( seber, "b", abs(subentries) == 1 ? 0 : 1 );
+		    	if ( err == -1 ) {
 				ber_free( seber, 1 );
 				fprintf( stderr, _("Subentries control encoding error!\n") );
 				return EXIT_FAILURE;
@@ -673,10 +778,14 @@ getNextPage:
 				return EXIT_FAILURE;
 			}
 
-			ber_printf( prber, "{iO}", pageSize, &cookie );
+			ber_printf( prber, "{iO}", pageSize, &page_cookie );
 			if ( ber_flatten2( prber, &c[i].ldctl_value, 0 ) == -1 ) {
 				return EXIT_FAILURE;
 			}
+			if ( page_cookie.bv_val != NULL ) {
+				ber_memfree( page_cookie.bv_val );
+				page_cookie.bv_val = NULL;
+			}
 			
 			c[i].ldctl_oid = LDAP_CONTROL_PAGEDRESULTS;
 			c[i].ldctl_iscritical = pagedResults > 1;
@@ -686,9 +795,13 @@ getNextPage:
 
 		tool_server_controls( ld, c, i );
 
+#ifdef LDAP_CONTROL_SUBENTRIES
 		ber_free( seber, 1 );
+#endif
 		ber_free( vrber, 1 );
+#ifdef LDAP_CONTROL_PAGEDRESULTS
 		ber_free( prber, 1 );
+#endif
 	}
 	
 	if ( verbose ) {
@@ -716,8 +829,17 @@ getNextPage:
 		printf( "#\n" );
 		printf(_("# LDAPv%d\n"), protocol);
 		printf(_("# base <%s> with scope %s\n"),
-			base ? base : "", (scope == LDAP_SCOPE_BASE) ? "base"
-		       : ((scope == LDAP_SCOPE_ONELEVEL) ? "one" : "sub"));
+			base ? base : "",
+			((scope == LDAP_SCOPE_BASE) ? "baseObject"
+				: ((scope == LDAP_SCOPE_ONELEVEL) ? "oneLevel"
+#ifdef LDAP_SCOPE_SUBORDINATE
+				: ((scope == LDAP_SCOPE_SUBORDINATE) ? "children"
+#endif
+				: "subtree"
+#ifdef LDAP_SCOPE_SUBORDINATE
+				)
+#endif
+				)));
 		printf(_("# filter%s: %s\n"), infile != NULL ? _(" pattern") : "",
 		       filtpattern);
 		printf(_("# requesting: "));
@@ -748,7 +870,7 @@ getNextPage:
 				valuesReturnFilter > 1 ? _("critical ") : "", vrFilter );
 		}
 #ifdef LDAP_CONTROL_PAGEDRESULTS
-		if ( pageSize ) {
+		if ( pagedResults ) {
 			printf(_("\n# with pagedResults %scontrol: size=%d"),
 				(pagedResults > 1) ? _("critical ") : "", 
 				pageSize );
@@ -781,44 +903,46 @@ getNextPage:
 	}
 
 #ifdef LDAP_CONTROL_PAGEDRESULTS
-	if ( ( pageSize != 0 ) && ( morePagedResults != 0 ) ) {
+	if ( ( rc == LDAP_SUCCESS ) &&  ( pageSize != 0 ) && ( morePagedResults != 0 ) ) {
 		char	buf[6];
 		int	i, moreEntries, tmpSize;
 
 		/* Loop to get the next pages when 
 		 * enter is pressed on the terminal.
 		 */
-		if ( entriesLeft > 0 ) {
-			printf( _("Estimate entries: %d\n"), entriesLeft );
-		}
-		printf( _("Press [size] Enter for the next {%d|size} entries.\n"),
-			(int)pageSize ); 
-		i = 0;
-		moreEntries = getchar();
-		while ( moreEntries != EOF && moreEntries != '\n' ) { 
-			if ( i < (int)sizeof(buf) - 1 ) {
-				buf[i] = moreEntries;
-				i++;
+		if ( pagePrompt != 0 ) {
+			if ( entriesLeft > 0 ) {
+				printf( _("Estimate entries: %d\n"), entriesLeft );
 			}
+			printf( _("Press [size] Enter for the next {%d|size} entries.\n"),
+				(int)pageSize );
+			i = 0;
 			moreEntries = getchar();
-		}
-		buf[i] = '\0';
+			while ( moreEntries != EOF && moreEntries != '\n' ) { 
+				if ( i < (int)sizeof(buf) - 1 ) {
+					buf[i] = moreEntries;
+					i++;
+				}
+				moreEntries = getchar();
+			}
+			buf[i] = '\0';
 
-		if ( i > 0 && isdigit( (unsigned char)buf[0] ) ) {
-			int num = sscanf( buf, "%d", &tmpSize );
-			if ( num != 1 ) {
-				fprintf( stderr, _("Invalid value for PagedResultsControl, %s.\n"), buf);
-				return EXIT_FAILURE;
-
+			if ( i > 0 && isdigit( (unsigned char)buf[0] ) ) {
+				int num = sscanf( buf, "%d", &tmpSize );
+				if ( num != 1 ) {
+					fprintf( stderr, _("Invalid value for PagedResultsControl, %s.\n"), buf);
+					return EXIT_FAILURE;
+	
+				}
+				pageSize = (ber_int_t)tmpSize;
 			}
-			pageSize = (ber_int_t)tmpSize;
 		}
 
-		goto getNextPage;	
+		goto getNextPage;
 	}
 #endif
 
-	ldap_unbind( ld );
+	ldap_unbind_ext( ld, NULL, NULL );
 #ifdef HAVE_CYRUS_SASL
 	sasl_done();
 #endif
@@ -946,8 +1070,14 @@ static int dosearch(
 			case LDAP_RES_SEARCH_RESULT:
 				rc = print_result( ld, msg, 1 );
 #ifdef LDAP_CONTROL_PAGEDRESULTS
-				if ( pageSize != 0 ) { 
-					rc = parse_page_control( ld, msg, &cookie );
+				if ( pageSize != 0 ) {
+					if ( rc == LDAP_SUCCESS ) {
+						rc = parse_page_control( ld, msg, &page_cookie );
+					} else {
+						morePagedResults = 0;
+					}
+				} else {
+					morePagedResults = 0;
 				}
 #endif
 
@@ -992,6 +1122,20 @@ static int dosearch(
 
 		ldap_msgfree( res );
 	}
+#if defined(TEST_ABANDON) || defined(TEST_CANCEL)
+	if ( gotintr ) {
+#ifdef TEST_CANCEL
+		rc = ldap_cancel_s( ld, msgid, NULL, NULL );
+		fprintf( stderr, "got interrupt, cancel got %d\n", rc );
+		return -1;
+#endif
+#ifdef TEST_ABANDON
+		rc = ldap_abandon( ld, msgid );
+		fprintf( stderr, "got interrupt, abandon got %d\n", rc );
+		return -1;
+#endif
+	}
+#endif
 
 	if ( rc == -1 ) {
 		ldap_perror( ld, "ldap_result" );
@@ -1001,18 +1145,18 @@ static int dosearch(
 done:
 	ldap_msgfree( res );
 #ifdef LDAP_CONTROL_PAGEDRESULTS
-	if ( pageSize != 0 ) { 
-		npagedresponses = npagedresponses + nresponses;
-		npagedentries = npagedentries + nentries;
-		npagedreferences = npagedreferences + nreferences;
-		npagedextended = npagedextended + nextended;
-		npagedpartial = npagedpartial + npartial;
+	if ( pagedResults ) { 
+		npagedresponses += nresponses;
+		npagedentries += nentries;
+		npagedextended += nextended;
+		npagedpartial += npartial;
+		npagedreferences += nreferences;
 		if ( ( morePagedResults == 0 ) && ( ldif < 2 ) ) {
 			printf( _("\n# numResponses: %d\n"), npagedresponses );
-			if( nentries ) printf( _("# numEntries: %d\n"), npagedentries );
-			if( nextended ) printf( _("# numExtended: %d\n"), npagedextended );
-			if( npartial ) printf( _("# numPartial: %d\n"), npagedpartial );
-			if( nreferences ) printf( _("# numReferences: %d\n"), npagedreferences );
+			if( npagedentries ) printf( _("# numEntries: %d\n"), npagedentries );
+			if( npagedextended ) printf( _("# numExtended: %d\n"), npagedextended );
+			if( npagedpartial ) printf( _("# numPartial: %d\n"), npagedpartial );
+			if( npagedreferences ) printf( _("# numReferences: %d\n"), npagedreferences );
 		}
 	} else
 #endif
@@ -1027,116 +1171,6 @@ done:
 	return( rc );
 }
 
-#ifdef OLD_WAY
-/* This is the original version, the old way of doing things. */
-static void
-print_entry(
-	LDAP	*ld,
-	LDAPMessage	*entry,
-	int		attrsonly)
-{
-	char		*a, *dn, *ufn;
-	char	tmpfname[ 256 ];
-	char	url[ 256 ];
-	int			i, rc;
-	BerElement		*ber = NULL;
-	struct berval	**bvals;
-	LDAPControl **ctrls = NULL;
-	FILE		*tmpfp;
-
-	dn = ldap_get_dn( ld, entry );
-	ufn = NULL;
-
-	if ( ldif < 2 ) {
-		ufn = ldap_dn2ufn( dn );
-		write_ldif( LDIF_PUT_COMMENT, NULL, ufn, ufn ? strlen( ufn ) : 0 );
-	}
-	write_ldif( LDIF_PUT_VALUE, "dn", dn, dn ? strlen( dn ) : 0);
-
-	rc = ldap_get_entry_controls( ld, entry, &ctrls );
-
-	if( rc != LDAP_SUCCESS ) {
-		fprintf(stderr, _("print_entry: %d\n"), rc );
-		ldap_perror( ld, "ldap_get_entry_controls" );
-		exit( EXIT_FAILURE );
-	}
-
-	if( ctrls ) {
-		print_ctrls( ctrls );
-		ldap_controls_free( ctrls );
-	}
-
-	if ( includeufn ) {
-		if( ufn == NULL ) {
-			ufn = ldap_dn2ufn( dn );
-		}
-		write_ldif( LDIF_PUT_VALUE, "ufn", ufn, ufn ? strlen( ufn ) : 0 );
-	}
-
-	if( ufn != NULL ) ldap_memfree( ufn );
-	ldap_memfree( dn );
-
-	for ( a = ldap_first_attribute( ld, entry, &ber ); a != NULL;
-		a = ldap_next_attribute( ld, entry, ber ) )
-	{
-		if ( attrsonly ) {
-			write_ldif( LDIF_PUT_NOVALUE, a, NULL, 0 );
-
-		} else if (( bvals = ldap_get_values_len( ld, entry, a )) != NULL ) {
-			for ( i = 0; bvals[i] != NULL; i++ ) {
-				if ( vals2tmp > 1 || ( vals2tmp
-					&& ldif_is_not_printable( bvals[i]->bv_val, bvals[i]->bv_len ) ))
-				{
-					int tmpfd;
-					/* write value to file */
-					snprintf( tmpfname, sizeof tmpfname,
-						"%s" LDAP_DIRSEP "ldapsearch-%s-XXXXXX",
-						tmpdir, a );
-					tmpfp = NULL;
-
-					tmpfd = mkstemp( tmpfname );
-
-					if ( tmpfd < 0  ) {
-						perror( tmpfname );
-						continue;
-					}
-
-					if (( tmpfp = fdopen( tmpfd, "w")) == NULL ) {
-						perror( tmpfname );
-						continue;
-					}
-
-					if ( fwrite( bvals[ i ]->bv_val,
-						bvals[ i ]->bv_len, 1, tmpfp ) == 0 )
-					{
-						perror( tmpfname );
-						fclose( tmpfp );
-						continue;
-					}
-
-					fclose( tmpfp );
-
-					snprintf( url, sizeof url, "%s%s", urlpre,
-						&tmpfname[strlen(tmpdir) + sizeof(LDAP_DIRSEP) - 1] );
-
-					urlize( url );
-					write_ldif( LDIF_PUT_URL, a, url, strlen( url ));
-
-				} else {
-					write_ldif( LDIF_PUT_VALUE, a,
-						bvals[ i ]->bv_val, bvals[ i ]->bv_len );
-				}
-			}
-			ber_bvecfree( bvals );
-		}
-		ldap_memfree( a );
-	}
-
-	if( ber != NULL ) {
-		ber_free( ber, 0 );
-	}
-}
-#else
 /* This is the proposed new way of doing things.
  * It is more efficient, but the API is non-standard.
  */
@@ -1197,8 +1231,8 @@ print_entry(
 
 		} else if ( bvals ) {
 			for ( i = 0; bvals[i].bv_val != NULL; i++ ) {
-				if ( vals2tmp > 1 || ( vals2tmp
-					&& ldif_is_not_printable( bvals[i].bv_val, bvals[i].bv_len ) ))
+				if ( vals2tmp > 1 || ( vals2tmp &&
+					ldif_is_not_printable( bvals[i].bv_val, bvals[i].bv_len )))
 				{
 					int tmpfd;
 					/* write value to file */
@@ -1248,7 +1282,6 @@ print_entry(
 		ber_free( ber, 0 );
 	}
 }
-#endif
 
 static void print_reference(
 	LDAP *ld,
@@ -1536,7 +1569,6 @@ parse_page_control(
 	LDAPControl *ctrlp = NULL;
 	BerElement *ber;
 	ber_tag_t tag;
-	struct berval servercookie = { 0, NULL };
 
 	rc = ldap_parse_result( ld, result,
 		&err, NULL, NULL, NULL, &ctrl, 0 );
@@ -1565,8 +1597,7 @@ parse_page_control(
 			return EXIT_FAILURE;
 		}
 
-		tag = ber_scanf( ber, "{im}", &entriesLeft, &servercookie );
-		ber_dupbv( cookie, &servercookie );
+		tag = ber_scanf( ber, "{io}", &entriesLeft, cookie );
 		(void) ber_free( ber, 1 );
 
 		if( tag == LBER_ERROR ) {
@@ -1581,6 +1612,10 @@ parse_page_control(
 			return EXIT_FAILURE;
 		}
 
+		if ( cookie->bv_len == 0 ) {
+			morePagedResults = 0;
+		}
+
 		ldap_controls_free( ctrl );
 
 	} else {
