@@ -1,7 +1,7 @@
-/* $OpenLDAP: pkg/ldap/include/ldap_cdefs.h,v 1.23.2.1 2003/12/01 00:58:24 kurt Exp $ */
+/* $OpenLDAP: pkg/ldap/include/ldap_cdefs.h,v 1.26.2.1 2005/01/20 18:03:48 kurt Exp $ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  * 
- * Copyright 1998-2003 The OpenLDAP Foundation.
+ * Copyright 1998-2005 The OpenLDAP Foundation.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -158,13 +158,9 @@
 #endif
 
 /* LDBM library */
-#if defined(_WIN32) && defined(SLAPD_IMPORT)
-#	define LDAP_LDBM_F(type)	extern __declspec(dllimport) type
-#	define LDAP_LDBM_V(type)	extern __declspec(dllimport) type
-#else
+/* Not exported/imported any more */
 #	define LDAP_LDBM_F(type)	extern type
 #	define LDAP_LDBM_V(type)	extern type
-#endif
 
 /* LDIF library */
 #if defined(_WIN32) && defined(SLAPD_IMPORT)
