@@ -1,8 +1,8 @@
 /* os-ip.c -- platform-specific TCP & UDP related code */
-/* $OpenLDAP: pkg/ldap/libraries/libldap/os-ip.c,v 1.90.2.3 2003/12/01 00:58:26 kurt Exp $ */
+/* $OpenLDAP: pkg/ldap/libraries/libldap/os-ip.c,v 1.108.2.3 2005/02/02 17:32:43 kurt Exp $ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 1998-2003 The OpenLDAP Foundation.
+ * Copyright 1998-2005 The OpenLDAP Foundation.
  * Portions Copyright 1999 Lars Uffmann.
  * All rights reserved.
  *
@@ -39,8 +39,6 @@
 
 #include "ldap-int.h"
 
-int ldap_int_tblsize = 0;
-
 #if defined( HAVE_GETADDRINFO ) && defined( HAVE_INET_NTOP )
 #  ifdef LDAP_PF_INET6
 int ldap_int_inet4or6 = AF_UNSPEC;
@@ -212,10 +210,6 @@ ldap_pvt_connect(LDAP *ld, ber_socket_t s,
 {
 	int rc;
 	struct timeval	tv, *opt_tv=NULL;
-	fd_set		wfds, *z=NULL;
-#ifdef HAVE_WINSOCK
-	fd_set		efds;
-#endif
 
 #ifdef LDAP_CONNECTIONLESS
 	/* We could do a connect() but that would interfere with
@@ -258,56 +252,97 @@ ldap_pvt_connect(LDAP *ld, ber_socket_t s,
 	if ( async ) return ( -2 );
 #endif
 
-	FD_ZERO(&wfds);
-	FD_SET(s, &wfds );
+#ifdef HAVE_POLL
+	{
+		struct pollfd fd;
+		int timeout = INFTIM;
 
-#ifdef HAVE_WINSOCK
-	FD_ZERO(&efds);
-	FD_SET(s, &efds );
-#endif
+		if( opt_tv != NULL ) timeout = TV2MILLISEC( &tv );
 
-	do {
-		rc = select(ldap_int_tblsize, z, &wfds,
-#ifdef HAVE_WINSOCK
-			&efds,
+		fd.fd = s;
+		fd.events = POLL_WRITE;
+
+		do {
+			fd.revents = 0;
+			rc = poll( &fd, 1, timeout );
+		} while( rc == AC_SOCKET_ERROR && errno == EINTR &&
+			LDAP_BOOL_GET(&ld->ld_options, LDAP_BOOL_RESTART ));
+
+		if( rc == AC_SOCKET_ERROR ) return rc;
+
+		if( fd.revents & POLL_WRITE ) {
+			if ( ldap_pvt_is_socket_ready(ld, s) == -1 ) return -1;
+			if ( ldap_pvt_ndelay_off(ld, s) == -1 ) return -1;
+			return ( 0 );
+		}
+	}
 #else
-			z,
+	{
+		fd_set		wfds, *z=NULL;
+#ifdef HAVE_WINSOCK
+		fd_set		efds;
 #endif
-			opt_tv ? &tv : NULL);
-	} while( rc == AC_SOCKET_ERROR && errno == EINTR &&
-		LDAP_BOOL_GET(&ld->ld_options, LDAP_BOOL_RESTART ));
 
-	if( rc == AC_SOCKET_ERROR ) return rc;
+#ifdef FD_SETSIZE
+		if ( s >= FD_SETSIZE ) {
+			rc = AC_SOCKET_ERROR;
+			tcp_close( s );
+			ldap_pvt_set_errno( EMFILE );
+			return rc;
+		}
+#endif
+
+		do {
+			FD_ZERO(&wfds);
+			FD_SET(s, &wfds );
 
 #ifdef HAVE_WINSOCK
-	/* This means the connection failed */
-	if ( FD_ISSET(s, &efds) ) {
-	    int so_errno;
-	    int dummy = sizeof(so_errno);
-	    if ( getsockopt( s, SOL_SOCKET, SO_ERROR,
-			(char *) &so_errno, &dummy ) == AC_SOCKET_ERROR || !so_errno )
-	    {
-	    	/* impossible */
-	    	so_errno = WSAGetLastError();
-	    }
-	    ldap_pvt_set_errno(so_errno);
-	    osip_debug(ld, "ldap_pvt_connect: error on socket %d: "
-		       "errno: %d (%s)\n", s, errno, sock_errstr(errno));
-	    return -1;
-	}
+			FD_ZERO(&efds);
+			FD_SET(s, &efds );
 #endif
-	if ( FD_ISSET(s, &wfds) ) {
+
+			rc = select(ldap_int_tblsize, z, &wfds,
+#ifdef HAVE_WINSOCK
+				&efds,
+#else
+				z,
+#endif
+				opt_tv ? &tv : NULL);
+		} while( rc == AC_SOCKET_ERROR && errno == EINTR &&
+			LDAP_BOOL_GET(&ld->ld_options, LDAP_BOOL_RESTART ));
+
+		if( rc == AC_SOCKET_ERROR ) return rc;
+
+#ifdef HAVE_WINSOCK
+		/* This means the connection failed */
+		if ( FD_ISSET(s, &efds) ) {
+		    int so_errno;
+		    int dummy = sizeof(so_errno);
+		    if ( getsockopt( s, SOL_SOCKET, SO_ERROR,
+				(char *) &so_errno, &dummy ) == AC_SOCKET_ERROR || !so_errno )
+		    {
+		    	/* impossible */
+		    	so_errno = WSAGetLastError();
+		    }
+		    ldap_pvt_set_errno(so_errno);
+		    osip_debug(ld, "ldap_pvt_connect: error on socket %d: "
+			       "errno: %d (%s)\n", s, errno, sock_errstr(errno));
+		    return -1;
+		}
+#endif
+		if ( FD_ISSET(s, &wfds) ) {
 #ifndef HAVE_WINSOCK
-		if ( ldap_pvt_is_socket_ready(ld, s) == -1 )
-			return ( -1 );
+			if ( ldap_pvt_is_socket_ready(ld, s) == -1 ) return -1;
 #endif
-		if ( ldap_pvt_ndelay_off(ld, s) == -1 )
-			return ( -1 );
-		return ( 0 );
+			if ( ldap_pvt_ndelay_off(ld, s) == -1 ) return -1;
+			return 0;
+		}
 	}
+#endif
+
 	osip_debug(ld, "ldap_connect_timeout: timed out\n",0,0,0);
 	ldap_pvt_set_errno( ETIMEDOUT );
-	return ( -1 );
+	return -1;
 }
 
 #ifndef HAVE_INET_ATON
@@ -315,11 +350,14 @@ int
 ldap_pvt_inet_aton( const char *host, struct in_addr *in)
 {
 	unsigned long u = inet_addr( host );
-	if ( u != 0xffffffff || u != (unsigned long) -1 ) {
-		in->s_addr = u;
-		return 1;
-	}
-	return 0;
+
+#ifdef INADDR_NONE
+	if ( u == INADDR_NONE ) return 0;
+#endif
+	if ( u == 0xffffffffUL || u == (unsigned long) -1L ) return 0;
+
+	in->s_addr = u;
+	return 1;
 }
 #endif
 
@@ -368,9 +406,10 @@ ldap_connect_to_host(LDAP *ld, Sockbuf *sb,
 
 #if defined( HAVE_GETADDRINFO ) && defined( HAVE_INET_NTOP )
 	memset( &hints, '\0', sizeof(hints) );
-#ifdef AI_ADDRCONFIG
-	hints.ai_flags = AI_ADDRCONFIG;
-#endif	
+#ifdef USE_AI_ATTRCONFIG /* FIXME: configure test needed */
+	/* Use AI_ATTRCONFIG only on systems where its known to be needed. */
+	hints.ai_flags = AI_ATTRCONFIG;
+#endif
 	hints.ai_family = ldap_int_inet4or6;
 	hints.ai_socktype = socktype;
 	snprintf(serv, sizeof serv, "%d", port );
@@ -523,7 +562,6 @@ ldap_host_connected_to( Sockbuf *sb, const char *host )
 	struct sockaddr sabuf;
 #endif
 	struct sockaddr	*sa = (struct sockaddr *) &sabuf;
-	int rc;
 	ber_socket_t	sd;
 
 	(void)memset( (char *)sa, '\0', sizeof sabuf );
@@ -585,10 +623,15 @@ ldap_host_connected_to( Sockbuf *sb, const char *host )
 		break;
 	}
 
-#if 0
 	{
 		char *herr;
+#ifdef NI_MAXHOST
 		char hbuf[NI_MAXHOST];
+#elif defined( MAXHOSTNAMELEN
+		char hbuf[MAXHOSTNAMELEN];
+#else
+		char hbuf[256];
+#endif
 		hbuf[0] = 0;
 
 		if (ldap_pvt_get_hname( sa, len, hbuf, sizeof(hbuf), &herr ) == 0
@@ -597,22 +640,26 @@ ldap_host_connected_to( Sockbuf *sb, const char *host )
 			return LDAP_STRDUP( hbuf );   
 		}
 	}
-#endif
 
 	return host ? LDAP_STRDUP( host ) : NULL;
 }
 #endif
 
 
-/* for UNIX */
 struct selectinfo {
+#ifdef HAVE_POLL
+	/* for UNIX poll(2) */
+	int si_maxfd;
+	struct pollfd si_fds[FD_SETSIZE];
+#else
+	/* for UNIX select(2) */
 	fd_set	si_readfds;
 	fd_set	si_writefds;
 	fd_set	si_use_readfds;
 	fd_set	si_use_writefds;
+#endif
 };
 
-
 void
 ldap_mark_select_write( LDAP *ld, Sockbuf *sb )
 {
@@ -622,9 +669,39 @@ ldap_mark_select_write( LDAP *ld, Sockbuf *sb )
 	sip = (struct selectinfo *)ld->ld_selectinfo;
 	
 	ber_sockbuf_ctrl( sb, LBER_SB_OPT_GET_FD, &sd );
+
+#ifdef HAVE_POLL
+	/* for UNIX poll(2) */
+	{
+		int empty=-1;
+		int i;
+		for(i=0; i < sip->si_maxfd; i++) {
+			if( sip->si_fds[i].fd == sd ) {
+				sip->si_fds[i].events |= POLL_WRITE;
+				return;
+			}
+			if( empty==-1 && sip->si_fds[i].fd == -1 ) {
+				empty=i;
+			}
+		}
+
+		if( empty == -1 ) {
+			if( sip->si_maxfd >= FD_SETSIZE ) {
+				/* FIXME */
+				return;
+			}
+			empty = sip->si_maxfd++;
+		}
+
+		sip->si_fds[empty].fd = sd;
+		sip->si_fds[empty].events = POLL_WRITE;
+	}
+#else
+	/* for UNIX select(2) */
 	if ( !FD_ISSET( sd, &sip->si_writefds )) {
 		FD_SET( sd, &sip->si_writefds );
 	}
+#endif
 }
 
 
@@ -637,9 +714,39 @@ ldap_mark_select_read( LDAP *ld, Sockbuf *sb )
 	sip = (struct selectinfo *)ld->ld_selectinfo;
 
 	ber_sockbuf_ctrl( sb, LBER_SB_OPT_GET_FD, &sd );
+
+#ifdef HAVE_POLL
+	/* for UNIX poll(2) */
+	{
+		int empty=-1;
+		int i;
+		for(i=0; i < sip->si_maxfd; i++) {
+			if( sip->si_fds[i].fd == sd ) {
+				sip->si_fds[i].events |= POLL_READ;
+				return;
+			}
+			if( empty==-1 && sip->si_fds[i].fd == -1 ) {
+				empty=i;
+			}
+		}
+
+		if( empty == -1 ) {
+			if( sip->si_maxfd >= FD_SETSIZE ) {
+				/* FIXME */
+				return;
+			}
+			empty = sip->si_maxfd++;
+		}
+
+		sip->si_fds[empty].fd = sd;
+		sip->si_fds[empty].events = POLL_READ;
+	}
+#else
+	/* for UNIX select(2) */
 	if ( !FD_ISSET( sd, &sip->si_readfds )) {
 		FD_SET( sd, &sip->si_readfds );
 	}
+#endif
 }
 
 
@@ -652,8 +759,22 @@ ldap_mark_select_clear( LDAP *ld, Sockbuf *sb )
 	sip = (struct selectinfo *)ld->ld_selectinfo;
 
 	ber_sockbuf_ctrl( sb, LBER_SB_OPT_GET_FD, &sd );
+
+#ifdef HAVE_POLL
+	/* for UNIX poll(2) */
+	{
+		int i;
+		for(i=0; i < sip->si_maxfd; i++) {
+			if( sip->si_fds[i].fd == sd ) {
+				sip->si_fds[i].fd = -1;
+			}
+		}
+	}
+#else
+	/* for UNIX select(2) */
 	FD_CLR( sd, &sip->si_writefds );
 	FD_CLR( sd, &sip->si_readfds );
+#endif
 }
 
 
@@ -666,7 +787,23 @@ ldap_is_write_ready( LDAP *ld, Sockbuf *sb )
 	sip = (struct selectinfo *)ld->ld_selectinfo;
 
 	ber_sockbuf_ctrl( sb, LBER_SB_OPT_GET_FD, &sd );
+
+#ifdef HAVE_POLL
+	/* for UNIX poll(2) */
+	{
+		int i;
+		for(i=0; i < sip->si_maxfd; i++) {
+			if( sip->si_fds[i].fd == sd ) {
+				return sip->si_fds[i].revents & POLL_WRITE;
+			}
+		}
+
+		return 0;
+	}
+#else
+	/* for UNIX select(2) */
 	return( FD_ISSET( sd, &sip->si_use_writefds ));
+#endif
 }
 
 
@@ -679,7 +816,23 @@ ldap_is_read_ready( LDAP *ld, Sockbuf *sb )
 	sip = (struct selectinfo *)ld->ld_selectinfo;
 
 	ber_sockbuf_ctrl( sb, LBER_SB_OPT_GET_FD, &sd );
+
+#ifdef HAVE_POLL
+	/* for UNIX poll(2) */
+	{
+		int i;
+		for(i=0; i < sip->si_maxfd; i++) {
+			if( sip->si_fds[i].fd == sd ) {
+				return sip->si_fds[i].revents & POLL_READ;
+			}
+		}
+
+		return 0;
+	}
+#else
+	/* for UNIX select(2) */
 	return( FD_ISSET( sd, &sip->si_use_readfds ));
+#endif
 }
 
 
@@ -688,11 +841,18 @@ ldap_new_select_info( void )
 {
 	struct selectinfo	*sip;
 
-	if (( sip = (struct selectinfo *)LDAP_CALLOC( 1,
-	    sizeof( struct selectinfo ))) != NULL ) {
-		FD_ZERO( &sip->si_readfds );
-		FD_ZERO( &sip->si_writefds );
-	}
+	sip = (struct selectinfo *)LDAP_CALLOC( 1, sizeof( struct selectinfo ));
+
+	if ( sip == NULL ) return NULL;
+
+#ifdef HAVE_POLL
+	/* for UNIX poll(2) */
+	/* sip->si_maxfd=0 */
+#else
+	/* for UNIX select(2) */
+	FD_ZERO( &sip->si_readfds );
+	FD_ZERO( &sip->si_writefds );
+#endif
 
 	return( (void *)sip );
 }
@@ -705,10 +865,14 @@ ldap_free_select_info( void *sip )
 }
 
 
+#ifndef HAVE_POLL
+int ldap_int_tblsize = 0;
+
 void
 ldap_int_ip_init( void )
 {
 	int tblsize;
+
 #if defined( HAVE_SYSCONF )
 	tblsize = sysconf( _SC_OPEN_MAX );
 #elif defined( HAVE_GETDTABLESIZE )
@@ -718,32 +882,41 @@ ldap_int_ip_init( void )
 #endif /* !USE_SYSCONF */
 
 #ifdef FD_SETSIZE
-	if( tblsize > FD_SETSIZE )
-		tblsize = FD_SETSIZE;
-#endif	/* FD_SETSIZE*/
+	if( tblsize > FD_SETSIZE ) tblsize = FD_SETSIZE;
+#endif	/* FD_SETSIZE */
+
 	ldap_int_tblsize = tblsize;
 }
+#endif
 
 
 int
 ldap_int_select( LDAP *ld, struct timeval *timeout )
 {
+	int rc;
 	struct selectinfo	*sip;
 
-#ifdef NEW_LOGGING
-	LDAP_LOG ( CONNECTION, ENTRY, "ldap_int_select\n", 0, 0, 0 );
-#else
 	Debug( LDAP_DEBUG_TRACE, "ldap_int_select\n", 0, 0, 0 );
+
+#ifndef HAVE_POLL
+	if ( ldap_int_tblsize == 0 ) ldap_int_ip_init();
 #endif
 
-	if ( ldap_int_tblsize == 0 )
-		ldap_int_ip_init();
-
 	sip = (struct selectinfo *)ld->ld_selectinfo;
+
+#ifdef HAVE_POLL
+	{
+		int to = timeout ? TV2MILLISEC( timeout ) : INFTIM;
+		rc = poll( sip->si_fds, sip->si_maxfd, to );
+	}
+#else
 	sip->si_use_readfds = sip->si_readfds;
 	sip->si_use_writefds = sip->si_writefds;
 	
-	return( select( ldap_int_tblsize,
-	                &sip->si_use_readfds, &sip->si_use_writefds,
-	                NULL, timeout ));
+	rc = select( ldap_int_tblsize,
+		&sip->si_use_readfds, &sip->si_use_writefds,
+		NULL, timeout );
+#endif
+
+	return rc;
 }
