@@ -1,7 +1,7 @@
-/* $OpenLDAP: pkg/ldap/servers/slapd/add.c,v 1.162.2.4 2003/12/01 00:58:31 kurt Exp $ */
+/* $OpenLDAP: pkg/ldap/servers/slapd/add.c,v 1.208.2.4 2005/03/17 22:29:49 kurt Exp $ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 1998-2003 The OpenLDAP Foundation.
+ * Copyright 1998-2005 The OpenLDAP Foundation.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -30,15 +30,15 @@
 #include <ac/time.h>
 #include <ac/socket.h>
 
-#include "ldap_pvt.h"
 #include "slap.h"
 
 #ifdef LDAP_SLAPI
-#include "slapi.h"
-static void initAddPlugin( Operation *op,
-	struct berval *dn, Entry *e, int manageDSAit );
-static int doPreAddPluginFNs( Operation *op );
-static void doPostAddPluginFNs( Operation *op );
+#include "slapi/slapi.h"
+
+static void init_add_pblock( Operation *op, struct berval *dn, Entry *e,
+	int manageDSAit );
+static int call_add_preop_plugins( Operation *op );
+static void call_add_postop_plugins( Operation *op );
 #endif /* LDAP_SLAPI */
 
 int
@@ -46,20 +46,18 @@ do_add( Operation *op, SlapReply *rs )
 {
 	BerElement	*ber = op->o_ber;
 	char		*last;
-	struct berval dn = { 0, NULL };
+	struct berval	dn = BER_BVNULL;
 	ber_len_t	len;
 	ber_tag_t	tag;
 	Entry		*e;
 	Modifications	*modlist = NULL;
 	Modifications	**modtail = &modlist;
 	Modifications	tmp;
-	int	manageDSAit;
+	char		textbuf[ SLAP_TEXT_BUFLEN ];
+	size_t		textlen = sizeof( textbuf );
+	int		rc = 0;
 
-#ifdef NEW_LOGGING
-	LDAP_LOG( OPERATION, ENTRY, "do_add: conn %d enter\n", op->o_connid,0,0 );
-#else
 	Debug( LDAP_DEBUG_TRACE, "do_add\n", 0, 0, 0 );
-#endif
 	/*
 	 * Parse the add request.  It looks like this:
 	 *
@@ -74,27 +72,18 @@ do_add( Operation *op, SlapReply *rs )
 
 	/* get the name */
 	if ( ber_scanf( ber, "{m", /*}*/ &dn ) == LBER_ERROR ) {
-#ifdef NEW_LOGGING
-		LDAP_LOG( OPERATION, ERR, 
-			"do_add: conn %d ber_scanf failed\n", op->o_connid,0,0 );
-#else
 		Debug( LDAP_DEBUG_ANY, "do_add: ber_scanf failed\n", 0, 0, 0 );
-#endif
 		send_ldap_discon( op, rs, LDAP_PROTOCOL_ERROR, "decoding error" );
-		return -1;
+		return SLAPD_DISCONNECT;
 	}
 
 	e = (Entry *) ch_calloc( 1, sizeof(Entry) );
 
-	rs->sr_err = dnPrettyNormal( NULL, &dn, &op->o_req_dn, &op->o_req_ndn, op->o_tmpmemctx );
+	rs->sr_err = dnPrettyNormal( NULL, &dn, &op->o_req_dn, &op->o_req_ndn,
+		op->o_tmpmemctx );
 
-	if( rs->sr_err != LDAP_SUCCESS ) {
-#ifdef NEW_LOGGING
-		LDAP_LOG( OPERATION, ERR, 
-			"do_add: conn %d invalid dn (%s)\n", op->o_connid, dn.bv_val, 0 );
-#else
+	if ( rs->sr_err != LDAP_SUCCESS ) {
 		Debug( LDAP_DEBUG_ANY, "do_add: invalid dn (%s)\n", dn.bv_val, 0, 0 );
-#endif
 		send_ldap_error( op, rs, LDAP_INVALID_DN_SYNTAX, "invalid DN" );
 		goto done;
 	}
@@ -102,12 +91,7 @@ do_add( Operation *op, SlapReply *rs )
 	ber_dupbv( &e->e_name, &op->o_req_dn );
 	ber_dupbv( &e->e_nname, &op->o_req_ndn );
 
-#ifdef NEW_LOGGING
-	LDAP_LOG( OPERATION, ARGS, 
-		"do_add: conn %d  dn (%s)\n", op->o_connid, e->e_dn, 0 );
-#else
 	Debug( LDAP_DEBUG_ARGS, "do_add: dn (%s)\n", e->e_dn, 0, 0 );
-#endif
 
 	/* get the attrs */
 	for ( tag = ber_first_element( ber, &len, &last ); tag != LBER_DEFAULT;
@@ -121,26 +105,15 @@ do_add( Operation *op, SlapReply *rs )
 		rtag = ber_scanf( ber, "{m{W}}", &tmp.sml_type, &tmp.sml_values );
 
 		if ( rtag == LBER_ERROR ) {
-#ifdef NEW_LOGGING
-			LDAP_LOG( OPERATION, ERR, 
-				   "do_add: conn %d	 decoding error \n", op->o_connid, 0, 0 );
-#else
 			Debug( LDAP_DEBUG_ANY, "do_add: decoding error\n", 0, 0, 0 );
-#endif
 			send_ldap_discon( op, rs, LDAP_PROTOCOL_ERROR, "decoding error" );
-			rs->sr_err = -1;
+			rs->sr_err = SLAPD_DISCONNECT;
 			goto done;
 		}
 
 		if ( tmp.sml_values == NULL ) {
-#ifdef NEW_LOGGING
-			LDAP_LOG( OPERATION, INFO, 
-				"do_add: conn %d	 no values for type %s\n",
-				op->o_connid, tmp.sml_type.bv_val, 0 );
-#else
 			Debug( LDAP_DEBUG_ANY, "no values for type %s\n",
 				tmp.sml_type.bv_val, 0, 0 );
-#endif
 			send_ldap_error( op, rs, LDAP_PROTOCOL_ERROR,
 				"no values for attribute type" );
 			goto done;
@@ -159,24 +132,14 @@ do_add( Operation *op, SlapReply *rs )
 	}
 
 	if ( ber_scanf( ber, /*{*/ "}") == LBER_ERROR ) {
-#ifdef NEW_LOGGING
-		LDAP_LOG( OPERATION, ERR, 
-			"do_add: conn %d ber_scanf failed\n", op->o_connid, 0, 0 );
-#else
 		Debug( LDAP_DEBUG_ANY, "do_add: ber_scanf failed\n", 0, 0, 0 );
-#endif
 		send_ldap_discon( op, rs, LDAP_PROTOCOL_ERROR, "decoding error" );
-		rs->sr_err = -1;
+		rs->sr_err = SLAPD_DISCONNECT;
 		goto done;
 	}
 
-	if( get_ctrls( op, rs, 1 ) != LDAP_SUCCESS ) {
-#ifdef NEW_LOGGING
-		LDAP_LOG( OPERATION, INFO, 
-			"do_add: conn %d get_ctrls failed\n", op->o_connid, 0, 0 );
-#else
+	if ( get_ctrls( op, rs, 1 ) != LDAP_SUCCESS ) {
 		Debug( LDAP_DEBUG_ANY, "do_add: get_ctrls failed\n", 0, 0, 0 );
-#endif
 		goto done;
 	} 
 
@@ -186,21 +149,64 @@ do_add( Operation *op, SlapReply *rs )
 		goto done;
 	}
 
-	Statslog( LDAP_DEBUG_STATS, "conn=%lu op=%lu ADD dn=\"%s\"\n",
-	    op->o_connid, op->o_opid, e->e_name.bv_val, 0, 0 );
+	Statslog( LDAP_DEBUG_STATS, "%s ADD dn=\"%s\"\n",
+	    op->o_log_prefix, e->e_name.bv_val, 0, 0, 0 );
 
-	if( e->e_nname.bv_len == 0 ) {
+	if ( dn_match( &e->e_nname, &slap_empty_bv ) ) {
 		/* protocolError may be a more appropriate error */
 		send_ldap_error( op, rs, LDAP_ALREADY_EXISTS,
 			"root DSE already exists" );
 		goto done;
 
-	} else if ( bvmatch( &e->e_nname, &global_schemandn ) ) {
+	} else if ( dn_match( &e->e_nname, &frontendDB->be_schemandn ) ) {
 		send_ldap_error( op, rs, LDAP_ALREADY_EXISTS,
 			"subschema subentry already exists" );
 		goto done;
 	}
 
+	rs->sr_err = slap_mods_check( modlist, &rs->sr_text,
+			  textbuf, textlen, NULL );
+
+	if ( rs->sr_err != LDAP_SUCCESS ) {
+		send_ldap_result( op, rs );
+		goto done;
+	}
+
+	/* temporary; remove if not invoking backend function */
+	op->ora_e = e;
+	op->ora_modlist = modlist;
+
+	op->o_bd = frontendDB;
+	rc = frontendDB->be_add( op, rs );
+	if ( rc == 0 ) {
+		e = NULL;
+	}
+
+done:;
+	slap_graduate_commit_csn( op );
+
+	if ( modlist != NULL ) {
+		slap_mods_free( modlist );
+	}
+	if ( e != NULL ) {
+		entry_free( e );
+	}
+	op->o_tmpfree( op->o_req_dn.bv_val, op->o_tmpmemctx );
+	op->o_tmpfree( op->o_req_ndn.bv_val, op->o_tmpmemctx );
+
+	return rc;
+}
+
+int
+fe_op_add( Operation *op, SlapReply *rs )
+{
+	int		manageDSAit;
+	Entry		*e = op->ora_e;
+	Modifications	*modlist = op->ora_modlist;
+	Modifications	**modtail = &modlist;
+	int		rc = 0;
+	BackendDB *op_be;
+
 	manageDSAit = get_manageDSAit( op );
 
 	/*
@@ -208,25 +214,35 @@ do_add( Operation *op, SlapReply *rs )
 	 * appropriate one, or send a referral to our "referral server"
 	 * if we don't hold it.
 	 */
-	op->o_bd = select_backend( &e->e_nname, manageDSAit, 0 );
+	op->o_bd = select_backend( &e->e_nname, manageDSAit, 1 );
 	if ( op->o_bd == NULL ) {
 		rs->sr_ref = referral_rewrite( default_referral,
 			NULL, &e->e_name, LDAP_SCOPE_DEFAULT );
-		if (!rs->sr_ref) rs->sr_ref = default_referral;
-		if ( rs->sr_ref != NULL ) {
+		if ( !rs->sr_ref ) rs->sr_ref = default_referral;
+		if ( rs->sr_ref ) {
 			rs->sr_err = LDAP_REFERRAL;
+			op->o_bd = frontendDB;
 			send_ldap_result( op, rs );
+			op->o_bd = NULL;
 
 			if ( rs->sr_ref != default_referral ) {
 				ber_bvarray_free( rs->sr_ref );
 			}
 		} else {
+			op->o_bd = frontendDB;
 			send_ldap_error( op, rs, LDAP_UNWILLING_TO_PERFORM,
-					"referral missing" );
+				"no global superior knowledge" );
+			op->o_bd = NULL;
 		}
 		goto done;
 	}
 
+	/* If we've got a glued backend, check the real backend */
+	op_be = op->o_bd;
+	if ( SLAP_GLUE_INSTANCE( op->o_bd )) {
+		op->o_bd = select_backend( &e->e_nname, manageDSAit, 0 );
+	}
+
 	/* check restrictions */
 	if( backend_check_restrictions( op, rs, NULL ) != LDAP_SUCCESS ) {
 		send_ldap_result( op, rs );
@@ -239,7 +255,7 @@ do_add( Operation *op, SlapReply *rs )
 	}
 
 #ifdef LDAP_SLAPI
-	if ( op->o_pb ) initAddPlugin( op, &dn, e, manageDSAit );
+	if ( op->o_pb ) init_add_pblock( op, &op->o_req_dn, e, manageDSAit );
 #endif /* LDAP_SLAPI */
 
 	/*
@@ -250,45 +266,51 @@ do_add( Operation *op, SlapReply *rs )
 	 */
 	if ( op->o_bd->be_add ) {
 		/* do the update here */
-		int repl_user = be_isupdate(op->o_bd, &op->o_ndn );
+		int repl_user = be_isupdate( op );
 #ifndef SLAPD_MULTIMASTER
-		if ( LDAP_STAILQ_EMPTY( &op->o_bd->be_syncinfo ) &&
-			( !op->o_bd->be_update_ndn.bv_len || repl_user ))
-#else
-		if ( LDAP_STAILQ_EMPTY( &op->o_bd->be_syncinfo ))
+		if ( !SLAP_SHADOW(op->o_bd) || repl_user )
 #endif
 		{
-			int update = op->o_bd->be_update_ndn.bv_len;
-			char textbuf[SLAP_TEXT_BUFLEN];
-			size_t textlen = sizeof textbuf;
+			int		update = !BER_BVISEMPTY( &op->o_bd->be_update_ndn );
+			char		textbuf[ SLAP_TEXT_BUFLEN ];
+			size_t		textlen = sizeof( textbuf );
+			slap_callback	cb = { NULL, slap_replog_cb, NULL, NULL };
 
-			rs->sr_err = slap_mods_check( modlist, update, &rs->sr_text,
-										  textbuf, textlen, NULL );
+			op->o_bd = op_be;
 
-			if( rs->sr_err != LDAP_SUCCESS ) {
-				send_ldap_result( op, rs );
-				goto done;
+			if ( !update ) {
+				rs->sr_err = slap_mods_no_update_check( modlist,
+						&rs->sr_text,
+				  		textbuf, textlen );
+
+				if ( rs->sr_err != LDAP_SUCCESS ) {
+					send_ldap_result( op, rs );
+					goto done;
+				}
 			}
 
 			if ( !repl_user ) {
-				for( modtail = &modlist;
-					*modtail != NULL;
-					modtail = &(*modtail)->sml_next )
+				/* go to the last mod */
+				for ( modtail = &modlist;
+						*modtail != NULL;
+						modtail = &(*modtail)->sml_next )
 				{
 					assert( (*modtail)->sml_op == LDAP_MOD_ADD );
 					assert( (*modtail)->sml_desc != NULL );
 				}
-				rs->sr_err = slap_mods_opattrs( op, modlist, modtail,
-					&rs->sr_text, textbuf, textlen );
-				if( rs->sr_err != LDAP_SUCCESS ) {
+
+				rs->sr_err = slap_mods_opattrs( op, modlist,
+						modtail, &rs->sr_text,
+						textbuf, textlen, 1 );
+				if ( rs->sr_err != LDAP_SUCCESS ) {
 					send_ldap_result( op, rs );
 					goto done;
 				}
 			}
 
 			rs->sr_err = slap_mods2entry( modlist, &e, repl_user, 0,
-									&rs->sr_text, textbuf, textlen );
-			if( rs->sr_err != LDAP_SUCCESS ) {
+				&rs->sr_text, textbuf, textlen );
+			if ( rs->sr_err != LDAP_SUCCESS ) {
 				send_ldap_result( op, rs );
 				goto done;
 			}
@@ -299,7 +321,7 @@ do_add( Operation *op, SlapReply *rs )
 			 * will actually contain something.
 			 */
 			if ( op->o_pb ) {
-				rs->sr_err = doPreAddPluginFNs( op );
+				rs->sr_err = call_add_preop_plugins( op );
 				if ( rs->sr_err != LDAP_SUCCESS ) {
 					/* plugin will have sent result */
 					goto done;
@@ -308,16 +330,23 @@ do_add( Operation *op, SlapReply *rs )
 #endif /* LDAP_SLAPI */
 
 			op->ora_e = e;
-			if ( (op->o_bd->be_add)( op, rs ) == 0 ) {
 #ifdef SLAPD_MULTIMASTER
-				if ( !repl_user )
+			if ( !repl_user )
 #endif
-				{
-					replog( op );
-				}
+			{
+				cb.sc_next = op->o_callback;
+				op->o_callback = &cb;
+			}
+			rc = (op->o_bd->be_add)( op, rs );
+			if ( rc == 0 ) {
+				/* FIXME: be_entry_release_w() should be
+				 * called by do_add(), so that global
+				 * overlays on the way back can
+				 * at least read the entry */
 				be_entry_release_w( op, e );
 				e = NULL;
 			}
+			op->ora_e = NULL;
 
 #ifndef SLAPD_MULTIMASTER
 		} else {
@@ -328,7 +357,7 @@ do_add( Operation *op, SlapReply *rs )
 			 * on replicas (for now, it involves the minimum code intrusion).
 			 */
 			if ( op->o_pb ) {
-				rs->sr_err = doPreAddPluginFNs( op );
+				rs->sr_err = call_add_preop_plugins( op );
 				if ( rs->sr_err != LDAP_SUCCESS ) {
 					/* plugin will have sent result */
 					goto done;
@@ -336,17 +365,8 @@ do_add( Operation *op, SlapReply *rs )
 			}
 #endif /* LDAP_SLAPI */
 
-			if ( !LDAP_STAILQ_EMPTY( &op->o_bd->be_syncinfo )) {
-				syncinfo_t *si;
-				LDAP_STAILQ_FOREACH( si, &op->o_bd->be_syncinfo, si_next ) {
-					struct berval tmpbv;
-					ber_dupbv( &tmpbv, &si->si_provideruri_bv[0] );
-					ber_bvarray_add( &defref, &tmpbv );
-				}
-			} else {
-				defref = op->o_bd->be_update_refs
-					? op->o_bd->be_update_refs : default_referral;
-			}
+			defref = op->o_bd->be_update_refs
+				? op->o_bd->be_update_refs : default_referral;
 
 			if ( defref != NULL ) {
 				rs->sr_ref = referral_rewrite( defref,
@@ -356,52 +376,37 @@ do_add( Operation *op, SlapReply *rs )
 				if (!rs->sr_ref) rs->sr_ref = default_referral;
 				send_ldap_result( op, rs );
 
-				if ( rs->sr_ref != default_referral ) ber_bvarray_free( rs->sr_ref );
+				if ( rs->sr_ref != default_referral ) {
+					ber_bvarray_free( rs->sr_ref );
+				}
 			} else {
 				send_ldap_error( op, rs,
-						LDAP_UNWILLING_TO_PERFORM,
-						"referral missing" );
+					LDAP_UNWILLING_TO_PERFORM,
+					"shadow context; no update referral" );
 			}
 #endif /* SLAPD_MULTIMASTER */
 		}
 	} else {
 #ifdef LDAP_SLAPI
 		if ( op->o_pb ) {
-			rs->sr_err = doPreAddPluginFNs( op );
+			rs->sr_err = call_add_preop_plugins( op );
 			if ( rs->sr_err != LDAP_SUCCESS ) {
 				/* plugin will have sent result */
 				goto done;
 			}
 		}
 #endif
-#ifdef NEW_LOGGING
-	    LDAP_LOG( OPERATION, INFO, 
-		       "do_add: conn %d	 no backend support\n", op->o_connid, 0, 0 );
-#else
 	    Debug( LDAP_DEBUG_ARGS, "	 do_add: no backend support\n", 0, 0, 0 );
-#endif
 	    send_ldap_error( op, rs, LDAP_UNWILLING_TO_PERFORM,
 			"operation not supported within namingContext" );
 	}
 
 #ifdef LDAP_SLAPI
-	if ( op->o_pb ) doPostAddPluginFNs( op );
+	if ( op->o_pb ) call_add_postop_plugins( op );
 #endif /* LDAP_SLAPI */
 
-done:
-
-	slap_graduate_commit_csn( op );
-
-	if( modlist != NULL ) {
-		slap_mods_free( modlist );
-	}
-	if( e != NULL ) {
-		entry_free( e );
-	}
-	op->o_tmpfree( op->o_req_dn.bv_val, op->o_tmpmemctx );
-	op->o_tmpfree( op->o_req_ndn.bv_val, op->o_tmpmemctx );
-
-	return rs->sr_err;
+done:;
+	return rc;
 }
 
 int
@@ -453,24 +458,32 @@ slap_mods2entry(
 
 			/* should check for duplicates */
 
-			AC_MEMCPY( &attr->a_vals[i], mods->sml_values,
-				sizeof( struct berval ) * j );
-
-			/* trim the mods array */
-			ch_free( mods->sml_values );
-			mods->sml_values = NULL;
+			if ( dup ) {
+				for ( j = 0; mods->sml_values[j].bv_val; j++ ) {
+					ber_dupbv( &attr->a_vals[i+j], &mods->sml_values[j] );
+				}
+				BER_BVZERO( &attr->a_vals[i+j] );	
+			} else {
+				AC_MEMCPY( &attr->a_vals[i], mods->sml_values,
+					sizeof( struct berval ) * j );
+				ch_free( mods->sml_values );
+				mods->sml_values = NULL;
+			}
 
 			if( mods->sml_nvalues ) {
 				attr->a_nvals = ch_realloc( attr->a_nvals,
 					sizeof( struct berval ) * (i+j) );
-
-				AC_MEMCPY( &attr->a_nvals[i], mods->sml_nvalues,
-					sizeof( struct berval ) * j );
-
-				/* trim the mods array */
-				ch_free( mods->sml_nvalues );
-				mods->sml_nvalues = NULL;
-
+				if ( dup ) {
+					for ( j = 0; mods->sml_nvalues[j].bv_val; j++ ) {
+						ber_dupbv( &attr->a_nvals[i+j], &mods->sml_nvalues[j] );
+					}
+					BER_BVZERO( &attr->a_nvals[i+j] );	
+				} else {
+					AC_MEMCPY( &attr->a_nvals[i], mods->sml_nvalues,
+						sizeof( struct berval ) * j );
+					ch_free( mods->sml_nvalues );
+					mods->sml_nvalues = NULL;
+				}
 			} else {
 				attr->a_nvals = attr->a_vals;
 			}
@@ -491,11 +504,12 @@ slap_mods2entry(
 
 			/* check if the values we're adding already exist */
 			if( mr == NULL || !mr->smr_match ) {
-				for ( i = 0; mods->sml_bvalues[i].bv_val != NULL; i++ ) {
+				for ( i = 1; mods->sml_values[i].bv_val != NULL; i++ ) {
 					/* test asserted values against themselves */
 					for( j = 0; j < i; j++ ) {
-						if ( bvmatch( &mods->sml_bvalues[i],
-							&mods->sml_bvalues[j] ) ) {
+						if ( bvmatch( &mods->sml_values[i],
+							&mods->sml_values[j] ) )
+						{
 							/* value exists already */
 							snprintf( textbuf, textlen,
 								"%s: value #%d provided more than once",
@@ -506,14 +520,15 @@ slap_mods2entry(
 				}
 
 			} else {
-				int		rc = LDAP_SUCCESS;
+				int	rc;
 				int match;
 
-				for ( i = 0; mods->sml_values[i].bv_val != NULL; i++ ) {
+				for ( i = 1; mods->sml_values[i].bv_val != NULL; i++ ) {
 					/* test asserted values against themselves */
 					for( j = 0; j < i; j++ ) {
 						rc = value_match( &match, mods->sml_desc, mr,
-							SLAP_MR_EQUALITY | SLAP_MR_VALUE_OF_ATTRIBUTE_SYNTAX
+							SLAP_MR_EQUALITY
+							| SLAP_MR_VALUE_OF_ATTRIBUTE_SYNTAX
 							| SLAP_MR_ASSERTED_VALUE_NORMALIZED_MATCH
 							| SLAP_MR_ATTRIBUTE_VALUE_NORMALIZED_MATCH,
 							mods->sml_nvalues
@@ -523,18 +538,19 @@ slap_mods2entry(
 								? &mods->sml_nvalues[j]
 								: &mods->sml_values[j],
 							text );
+
 						if ( rc == LDAP_SUCCESS && match == 0 ) {
 							/* value exists already */
 							snprintf( textbuf, textlen,
 								"%s: value #%d provided more than once",
 								mods->sml_desc->ad_cname.bv_val, j );
 							return LDAP_TYPE_OR_VALUE_EXISTS;
+
+						} else if ( rc != LDAP_SUCCESS ) {
+							return rc;
 						}
 					}
 				}
-				if ( rc != LDAP_SUCCESS ) {
-					return rc;
-				}
 			}
 		}
 
@@ -542,19 +558,18 @@ slap_mods2entry(
 
 		/* move ad to attr structure */
 		attr->a_desc = mods->sml_desc;
-		if ( !dup )
-			mods->sml_desc = NULL;
+		if ( !dup ) mods->sml_desc = NULL;
 
 		/* move values to attr structure */
 		/*	should check for duplicates */
 		if ( dup ) { 
 			int i;
-			for ( i = 0; mods->sml_values[i].bv_val; i++ ) ;
+			for ( i = 0; mods->sml_values[i].bv_val; i++ ) /* EMPTY */;
 			attr->a_vals = (BerVarray) ch_calloc( i+1, sizeof( BerValue ));
-			for ( i = 0; mods->sml_values[i].bv_val; i++ )
+			for ( i = 0; mods->sml_values[i].bv_val; i++ ) {
 				ber_dupbv( &attr->a_vals[i], &mods->sml_values[i] );
-			attr->a_vals[i].bv_len = 0;
-			attr->a_vals[i].bv_val = NULL;
+			}
+			BER_BVZERO( &attr->a_vals[i] );
 		} else {
 			attr->a_vals = mods->sml_values;
 			mods->sml_values = NULL;
@@ -563,12 +578,12 @@ slap_mods2entry(
 		if ( mods->sml_nvalues ) {
 			if ( dup ) {
 				int i;
-				for ( i = 0; mods->sml_nvalues[i].bv_val; i++ ) ;
+				for ( i = 0; mods->sml_nvalues[i].bv_val; i++ ) /* EMPTY */;
 				attr->a_nvals = (BerVarray) ch_calloc( i+1, sizeof( BerValue ));
-				for ( i = 0; mods->sml_nvalues[i].bv_val; i++ )
+				for ( i = 0; mods->sml_nvalues[i].bv_val; i++ ) {
 					ber_dupbv( &attr->a_nvals[i], &mods->sml_nvalues[i] );
-				attr->a_nvals[i].bv_len = 0;
-				attr->a_nvals[i].bv_val = NULL;
+				}
+				BER_BVZERO( &attr->a_nvals[i] );
 			} else {
 				attr->a_nvals = mods->sml_nvalues;
 				mods->sml_nvalues = NULL;
@@ -581,6 +596,8 @@ slap_mods2entry(
 		tail = &attr->a_next;
 	}
 
+	*text = NULL;
+
 	return LDAP_SUCCESS;
 }
 
@@ -588,15 +605,15 @@ int
 slap_entry2mods(
 	Entry *e,
 	Modifications **mods,
-	const char **text
-)
+	const char **text,
+	char *textbuf, size_t textlen )
 {
 	Modifications	*modhead = NULL;
 	Modifications	*mod;
 	Modifications	**modtail = &modhead;
 	Attribute		*a_new;
 	AttributeDescription	*a_new_desc;
-	int				i, count, rc;
+	int				i, count;
 
 	a_new = e->e_attrs;
 
@@ -604,17 +621,14 @@ slap_entry2mods(
 		a_new_desc = a_new->a_desc;
 		mod = (Modifications *) malloc( sizeof( Modifications ));
 		
-		if ( a_new_desc != slap_schema.si_ad_queryid )
-			mod->sml_op = LDAP_MOD_REPLACE;
-		else
-			mod->sml_op = LDAP_MOD_ADD;
+		mod->sml_op = LDAP_MOD_REPLACE;
 
-		ber_dupbv( &mod->sml_type, &a_new_desc->ad_cname );
+		mod->sml_type = a_new_desc->ad_cname;
 
-		for ( count = 0; a_new->a_vals[count].bv_val; count++ );
+		for ( count = 0; a_new->a_vals[count].bv_val; count++ ) /* EMPTY */;
 
-		mod->sml_bvalues = (struct berval*) malloc(
-				(count+1) * sizeof( struct berval) );
+		mod->sml_values = (struct berval*) malloc(
+			(count+1) * sizeof( struct berval) );
 
 		/* see slap_mods_check() comments...
 		 * if a_vals == a_nvals, there is no normalizer.
@@ -628,17 +642,17 @@ slap_entry2mods(
 		}
 
 		for ( i = 0; i < count; i++ ) {
-			ber_dupbv(mod->sml_bvalues+i, a_new->a_vals+i); 
+			ber_dupbv(mod->sml_values+i, a_new->a_vals+i); 
 			if ( mod->sml_nvalues ) {
-				ber_dupbv( mod->sml_nvalues+i, a_new->a_vals+i ); 
+				ber_dupbv( mod->sml_nvalues+i, a_new->a_nvals+i ); 
 			} 
 		}
 
-		mod->sml_bvalues[count].bv_val = 0; 
-		mod->sml_bvalues[count].bv_len = 0; 
+		mod->sml_values[count].bv_val = NULL; 
+		mod->sml_values[count].bv_len = 0; 
 
 		if ( mod->sml_nvalues ) {
-			mod->sml_nvalues[count].bv_val = 0; 
+			mod->sml_nvalues[count].bv_val = NULL; 
 			mod->sml_nvalues[count].bv_len = 0; 
 		}
 
@@ -655,36 +669,34 @@ slap_entry2mods(
 }
 
 #ifdef LDAP_SLAPI
-static void initAddPlugin( Operation *op,
+static void init_add_pblock( Operation *op,
 	struct berval *dn, Entry *e, int manageDSAit )
 {
-	slapi_x_pblock_set_operation( op->o_pb, op );
+	slapi_int_pblock_set_operation( op->o_pb, op );
 	slapi_pblock_set( op->o_pb, SLAPI_ADD_TARGET, (void *)dn->bv_val );
 	slapi_pblock_set( op->o_pb, SLAPI_ADD_ENTRY, (void *)e );
 	slapi_pblock_set( op->o_pb, SLAPI_MANAGEDSAIT, (void *)manageDSAit );
 }
 
-static int doPreAddPluginFNs( Operation *op )
+static int call_add_preop_plugins( Operation *op )
 {
 	int rc;
 
-	rc = doPluginFNs( op->o_bd, SLAPI_PLUGIN_PRE_ADD_FN, op->o_pb );
+	rc = slapi_int_call_plugins( op->o_bd, SLAPI_PLUGIN_PRE_ADD_FN, op->o_pb );
 	if ( rc < 0 ) {
 		/*
 		 * A preoperation plugin failure will abort the
 		 * entire operation.
 		 */
-#ifdef NEW_LOGGING
-		LDAP_LOG( OPERATION, INFO, "do_add: add preoperation plugin failed\n",
-				0, 0, 0);
-#else
-		Debug(LDAP_DEBUG_TRACE, "do_add: add preoperation plugin failed.\n",
-				0, 0, 0);
-		if ( ( slapi_pblock_get( op->o_pb, SLAPI_RESULT_CODE, (void *)&rc ) != 0 ) ||
-		     rc == LDAP_SUCCESS ) {
+		Debug(LDAP_DEBUG_TRACE,
+			"do_add: add preoperation plugin failed.\n",
+			0, 0, 0);
+
+		if (( slapi_pblock_get( op->o_pb, SLAPI_RESULT_CODE,
+			(void *)&rc ) != 0 ) || rc == LDAP_SUCCESS )
+		{
 			rc = LDAP_OTHER;
 		}
-#endif
 	} else {
 		rc = LDAP_SUCCESS;
 	}
@@ -692,19 +704,15 @@ static int doPreAddPluginFNs( Operation *op )
 	return rc;
 }
 
-static void doPostAddPluginFNs( Operation *op )
+static void call_add_postop_plugins( Operation *op )
 {
 	int rc;
 
-	rc = doPluginFNs( op->o_bd, SLAPI_PLUGIN_POST_ADD_FN, op->o_pb );
+	rc = slapi_int_call_plugins( op->o_bd, SLAPI_PLUGIN_POST_ADD_FN, op->o_pb );
 	if ( rc < 0 ) {
-#ifdef NEW_LOGGING
-		LDAP_LOG( OPERATION, INFO, "do_add: add postoperation plugin failed\n",
-				0, 0, 0);
-#else
-		Debug(LDAP_DEBUG_TRACE, "do_add: add postoperation plugin failed.\n",
-				0, 0, 0);
-#endif
+		Debug(LDAP_DEBUG_TRACE,
+			"do_add: add postoperation plugin failed\n",
+			0, 0, 0);
 	}
 }
 #endif /* LDAP_SLAPI */
