@@ -1,8 +1,8 @@
 /* attr.c - routines for dealing with attributes */
-/* $OpenLDAP: pkg/ldap/servers/slapd/attr.c,v 1.88.2.3 2003/12/01 00:58:31 kurt Exp $ */
+/* $OpenLDAP: pkg/ldap/servers/slapd/attr.c,v 1.100.2.3 2005/01/24 21:28:51 kurt Exp $ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 1998-2003 The OpenLDAP Foundation.
+ * Copyright 1998-2005 The OpenLDAP Foundation.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -38,17 +38,35 @@
 #include <ac/string.h>
 #include <ac/time.h>
 
-#include "ldap_pvt.h"
 #include "slap.h"
 
 void
 attr_free( Attribute *a )
 {
+	if ( a->a_nvals && a->a_nvals != a->a_vals ) {
+		ber_bvarray_free( a->a_nvals );
+	}
 	ber_bvarray_free( a->a_vals );
-	if (a->a_nvals != a->a_vals) ber_bvarray_free( a->a_nvals );
 	free( a );
 }
 
+#ifdef LDAP_COMP_MATCH
+void
+comp_tree_free( Attribute *a )
+{
+	Attribute *next;
+
+	for( ; a != NULL ; a = next ) {
+		next = a->a_next;
+		if ( component_destructor && a->a_comp_data ) {
+			if ( a->a_comp_data->cd_mem_op )
+				component_destructor( a->a_comp_data->cd_mem_op );
+			free ( a->a_comp_data );
+		}
+	}
+}
+#endif
+
 void
 attrs_free( Attribute *a )
 {
@@ -60,35 +78,41 @@ attrs_free( Attribute *a )
 	}
 }
 
-Attribute *attr_dup( Attribute *a )
+Attribute *
+attr_dup( Attribute *a )
 {
 	Attribute *tmp;
 
-	if( a == NULL) return NULL;
+	if ( a == NULL) return NULL;
 
 	tmp = ch_malloc( sizeof(Attribute) );
 
-	if( a->a_vals != NULL ) {
+	if ( a->a_vals != NULL ) {
 		int i;
 
-		for( i=0; a->a_vals[i].bv_val != NULL; i++ ) {
+		for ( i = 0; !BER_BVISNULL( &a->a_vals[i] ); i++ ) {
 			/* EMPTY */ ;
 		}
 
-		tmp->a_vals = ch_malloc((i+1) * sizeof(struct berval));
-		for( i=0; a->a_vals[i].bv_val != NULL; i++ ) {
+		tmp->a_vals = ch_malloc( (i + 1) * sizeof(struct berval) );
+		for ( i = 0; !BER_BVISNULL( &a->a_vals[i] ); i++ ) {
 			ber_dupbv( &tmp->a_vals[i], &a->a_vals[i] );
-			if( tmp->a_vals[i].bv_val == NULL ) break;
+			if ( BER_BVISNULL( &tmp->a_vals[i] ) ) break;
+			/* FIXME: error? */
 		}
-		tmp->a_vals[i].bv_val = NULL;
+		BER_BVZERO( &tmp->a_vals[i] );
 
-		if( a->a_nvals != a->a_vals ) {
-			tmp->a_nvals = ch_malloc((i+1) * sizeof(struct berval));
-			for( i=0; a->a_nvals[i].bv_val != NULL; i++ ) {
+		/* a_nvals must be non null; it may be equal to a_vals */
+		assert( a->a_nvals );
+
+		if ( a->a_nvals != a->a_vals ) {
+			tmp->a_nvals = ch_malloc( (i + 1) * sizeof(struct berval) );
+			for ( i = 0; !BER_BVISNULL( &a->a_nvals[i] ); i++ ) {
 				ber_dupbv( &tmp->a_nvals[i], &a->a_nvals[i] );
-				if( tmp->a_nvals[i].bv_val == NULL ) break;
+				if ( BER_BVISNULL( &tmp->a_nvals[i] ) ) break;
+				/* FIXME: error? */
 			}
-			tmp->a_nvals[i].bv_val = NULL;
+			BER_BVZERO( &tmp->a_nvals[i] );
 
 		} else {
 			tmp->a_nvals = tmp->a_vals;
@@ -102,11 +126,15 @@ Attribute *attr_dup( Attribute *a )
 	tmp->a_desc = a->a_desc;
 	tmp->a_next = NULL;
 	tmp->a_flags = 0;
+#ifdef LDAP_COMP_MATCH
+	tmp->a_comp_data = NULL;
+#endif
 
 	return tmp;
 }
 
-Attribute *attrs_dup( Attribute *a )
+Attribute *
+attrs_dup( Attribute *a )
 {
 	Attribute *tmp, **next;
 
@@ -142,14 +170,14 @@ attr_merge(
 	Entry		*e,
 	AttributeDescription *desc,
 	BerVarray	vals,
-	BerVarray	nvals
-) {
+	BerVarray	nvals )
+{
 	int rc;
 
 	Attribute	**a;
 
 	for ( a = &e->e_attrs; *a != NULL; a = &(*a)->a_next ) {
-		if ( ad_cmp( (*a)->a_desc, desc ) == 0 ) {
+		if (  (*a)->a_desc == desc ) {
 			break;
 		}
 	}
@@ -161,12 +189,21 @@ attr_merge(
 		(*a)->a_nvals = NULL;
 		(*a)->a_next = NULL;
 		(*a)->a_flags = 0;
+#ifdef LDAP_COMP_MATCH
+		(*a)->a_comp_data = NULL;
+#endif
 	}
 
 	rc = value_add( &(*a)->a_vals, vals );
 
-	if( !rc && nvals ) rc = value_add( &(*a)->a_nvals, nvals );
-	else (*a)->a_nvals = (*a)->a_vals;
+	if ( rc == LDAP_SUCCESS ) {
+		if ( nvals ) {
+			rc = value_add( &(*a)->a_nvals, nvals );
+			/* FIXME: what if rc != LDAP_SUCCESS ? */
+		} else {
+			(*a)->a_nvals = (*a)->a_vals;
+		}
+	}
 
 	return rc;
 }
@@ -176,8 +213,8 @@ attr_merge_normalize(
 	Entry		*e,
 	AttributeDescription *desc,
 	BerVarray	vals,
-	void	 *memctx
-) {
+	void	 *memctx )
+{
 	BerVarray	nvals = NULL;
 	int		rc;
 
@@ -186,10 +223,10 @@ attr_merge_normalize(
 	{
 		int	i;
 		
-		for ( i = 0; vals[i].bv_val; i++ );
+		for ( i = 0; !BER_BVISNULL( &vals[i] ); i++ );
 
-		nvals = sl_calloc( sizeof(struct berval), i + 1, memctx );
-		for ( i = 0; vals[i].bv_val; i++ ) {
+		nvals = slap_sl_calloc( sizeof(struct berval), i + 1, memctx );
+		for ( i = 0; !BER_BVISNULL( &vals[i] ); i++ ) {
 			rc = (*desc->ad_type->sat_equality->smr_normalize)(
 					SLAP_MR_VALUE_OF_ATTRIBUTE_SYNTAX,
 					desc->ad_type->sat_syntax,
@@ -197,11 +234,11 @@ attr_merge_normalize(
 					&vals[i], &nvals[i], memctx );
 
 			if ( rc != LDAP_SUCCESS ) {
-				nvals[i+1].bv_val = NULL;
+				BER_BVZERO( &nvals[i + 1] );
 				goto error_return;
 			}
 		}
-		nvals[i].bv_val = NULL;
+		BER_BVZERO( &nvals[i] );
 	}
 
 	rc = attr_merge( e, desc, vals, nvals );
@@ -218,13 +255,13 @@ attr_merge_one(
 	Entry		*e,
 	AttributeDescription *desc,
 	struct berval	*val,
-	struct berval	*nval
-) {
+	struct berval	*nval )
+{
 	int rc;
 	Attribute	**a;
 
 	for ( a = &e->e_attrs; *a != NULL; a = &(*a)->a_next ) {
-		if ( ad_cmp( (*a)->a_desc, desc ) == 0 ) {
+		if ( (*a)->a_desc == desc ) {
 			break;
 		}
 	}
@@ -236,12 +273,21 @@ attr_merge_one(
 		(*a)->a_nvals = NULL;
 		(*a)->a_next = NULL;
 		(*a)->a_flags = 0;
+#ifdef LDAP_COMP_MATCH
+		(*a)->a_comp_data = NULL;
+#endif
 	}
 
 	rc = value_add_one( &(*a)->a_vals, val );
 
-	if( !rc && nval ) rc = value_add_one( &(*a)->a_nvals, nval );
-	else (*a)->a_nvals = (*a)->a_vals;
+	if ( rc == LDAP_SUCCESS ) {
+		if ( nval ) {
+			rc = value_add_one( &(*a)->a_nvals, nval );
+			/* FIXME: what if rc != LDAP_SUCCESS ? */
+		} else {
+			(*a)->a_nvals = (*a)->a_vals;
+		}
+	}
 	return rc;
 }
 
@@ -250,10 +296,10 @@ attr_merge_normalize_one(
 	Entry		*e,
 	AttributeDescription *desc,
 	struct berval	*val,
-	void		*memctx
-) {
+	void		*memctx )
+{
 	struct berval	nval;
-	struct berval	*nvalp;
+	struct berval	*nvalp = NULL;
 	int		rc;
 
 	if ( desc->ad_type->sat_equality &&
@@ -269,13 +315,11 @@ attr_merge_normalize_one(
 			return rc;
 		}
 		nvalp = &nval;
-	} else {
-		nvalp = NULL;
 	}
 
 	rc = attr_merge_one( e, desc, val, nvalp );
 	if ( nvalp != NULL ) {
-		sl_free( nval.bv_val, memctx );
+		slap_sl_free( nval.bv_val, memctx );
 	}
 	return rc;
 }
@@ -288,8 +332,7 @@ attr_merge_normalize_one(
 Attribute *
 attrs_find(
     Attribute	*a,
-	AttributeDescription *desc
-)
+	AttributeDescription *desc )
 {
 	for ( ; a != NULL; a = a->a_next ) {
 		if ( is_ad_subtype( a->a_desc, desc ) ) {
@@ -307,11 +350,10 @@ attrs_find(
 Attribute *
 attr_find(
     Attribute	*a,
-	AttributeDescription *desc
-)
+	AttributeDescription *desc )
 {
 	for ( ; a != NULL; a = a->a_next ) {
-		if ( ad_cmp( a->a_desc, desc ) == 0 ) {
+		if ( a->a_desc == desc ) {
 			return( a );
 		}
 	}
@@ -329,13 +371,12 @@ attr_find(
 int
 attr_delete(
     Attribute	**attrs,
-	AttributeDescription *desc
-)
+	AttributeDescription *desc )
 {
 	Attribute	**a;
 
 	for ( a = attrs; *a != NULL; a = &(*a)->a_next ) {
-		if ( ad_cmp( (*a)->a_desc, desc ) == 0 ) {
+		if ( (*a)->a_desc == desc ) {
 			Attribute	*save = *a;
 			*a = (*a)->a_next;
 			attr_free( save );
