@@ -1,8 +1,8 @@
 /* idl.c - ldap id list handling routines */
-/* $OpenLDAP: pkg/ldap/servers/slapd/back-bdb/idl.c,v 1.76.2.1 2003/12/01 00:58:38 kurt Exp $ */
+/* $OpenLDAP: pkg/ldap/servers/slapd/back-bdb/idl.c,v 1.94.2.4 2005/03/14 22:24:59 kurt Exp $ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 2000-2003 The OpenLDAP Foundation.
+ * Copyright 2000-2005 The OpenLDAP Foundation.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -27,7 +27,6 @@
 
 #define IDL_CMP(x,y)	( x < y ? -1 : ( x > y ? 1 : 0 ) )
 
-#ifdef SLAP_IDL_CACHE
 #define IDL_LRU_DELETE( bdb, e ) do { 					\
 	if ( e->idl_lru_prev != NULL ) {				\
 		e->idl_lru_prev->idl_lru_next = e->idl_lru_next; 	\
@@ -59,11 +58,10 @@ bdb_idl_entry_cmp( const void *v_idl1, const void *v_idl2 )
 	const bdb_idl_cache_entry_t *idl1 = v_idl1, *idl2 = v_idl2;
 	int rc;
 
-	if ((rc = idl1->db - idl2->db )) return rc;
+	if ((rc = SLAP_PTRCMP( idl1->db, idl2->db ))) return rc;
 	if ((rc = idl1->kstr.bv_len - idl2->kstr.bv_len )) return rc;
 	return ( memcmp ( idl1->kstr.bv_val, idl2->kstr.bv_val , idl1->kstr.bv_len ) );
 }
-#endif
 
 #if IDL_DEBUG > 0
 static void idl_check( ID *ids )
@@ -82,34 +80,20 @@ static void idl_check( ID *ids )
 static void idl_dump( ID *ids )
 {
 	if( BDB_IDL_IS_RANGE( ids ) ) {
-#ifdef NEW_LOGGING
-		LDAP_LOG( INDEX, INFO, "IDL: range (%ld - %ld)\n",
-			(long) BDB_IDL_RANGE_FIRST( ids ),
-			(long) BDB_IDL_RANGE_LAST( ids ), 0 );
-#else
 		Debug( LDAP_DEBUG_ANY,
 			"IDL: range ( %ld - %ld )\n",
 			(long) BDB_IDL_RANGE_FIRST( ids ),
 			(long) BDB_IDL_RANGE_LAST( ids ) );
-#endif
 
 	} else {
 		ID i;
-#ifdef NEW_LOGGING
-		LDAP_LOG( INDEX, INFO, "IDL: size %ld", (long) ids[0], 0, 0 );
-#else
 		Debug( LDAP_DEBUG_ANY, "IDL: size %ld", (long) ids[0], 0, 0 );
-#endif
 
 		for( i=1; i<=ids[0]; i++ ) {
 			if( i % 16 == 1 ) {
 				Debug( LDAP_DEBUG_ANY, "\n", 0, 0, 0 );
 			}
-#ifdef NEW_LOGGING
-			LDAP_LOG( INDEX, INFO, "%02lx",(long)ids[i], 0, 0 );
-#else
 			Debug( LDAP_DEBUG_ANY, "  %02lx", (long) ids[i], 0, 0 );
-#endif
 		}
 
 		Debug( LDAP_DEBUG_ANY, "\n", 0, 0, 0 );
@@ -184,12 +168,8 @@ int bdb_idl_insert( ID *ids, ID id )
 	unsigned x;
 
 #if IDL_DEBUG > 1
-#ifdef NEW_LOGGING
-	LDAP_LOG( INDEX, DETAIL1, "insert: %04lx at %d\n", (long) id, x, 0 );
-#else
 	Debug( LDAP_DEBUG_ANY, "insert: %04lx at %d\n", (long) id, x, 0 );
 	idl_dump( ids );
-#endif
 #elif IDL_DEBUG > 0
 	idl_check( ids );
 #endif
@@ -250,12 +230,8 @@ static int idl_delete( ID *ids, ID id )
 	unsigned x = bdb_idl_search( ids, id );
 
 #if IDL_DEBUG > 1
-#ifdef NEW_LOGGING
-	LDAP_LOG( INDEX, DETAIL1, "delete: %04lx at %d\n", (long) id, x, 0 );
-#else
 	Debug( LDAP_DEBUG_ANY, "delete: %04lx at %d\n", (long) id, x, 0 );
 	idl_dump( ids );
-#endif
 #elif IDL_DEBUG > 0
 	idl_check( ids );
 #endif
@@ -295,7 +271,7 @@ bdb_show_key(
 	DBT		*key,
 	char		*buf )
 {
-	if ( key->size == sizeof( ID ) ) {
+	if ( key->size == 4 /* LUTIL_HASH_BYTES */ ) {
 		unsigned char *c = key->data;
 		sprintf( buf, "[%02x%02x%02x%02x]", c[0], c[1], c[2], c[3] );
 		return buf;
@@ -304,8 +280,6 @@ bdb_show_key(
 	}
 }
 
-#ifdef SLAP_IDL_CACHE
-
 /* Find a db/key pair in the IDL cache. If ids is non-NULL,
  * copy the cached IDL into it, otherwise just return the status.
  */
@@ -318,6 +292,7 @@ bdb_idl_cache_get(
 {
 	bdb_idl_cache_entry_t idl_tmp;
 	bdb_idl_cache_entry_t *matched_idl_entry;
+	int rc = LDAP_NO_SUCH_OBJECT;
 
 	DBT2bv( key, &idl_tmp.kstr );
 	idl_tmp.db = db;
@@ -327,19 +302,18 @@ bdb_idl_cache_get(
 	if ( matched_idl_entry != NULL ) {
 		if ( matched_idl_entry->idl && ids )
 			BDB_IDL_CPY( ids, matched_idl_entry->idl );
-		ldap_pvt_thread_rdwr_runlock( &bdb->bi_idl_tree_rwlock );
 		ldap_pvt_thread_mutex_lock( &bdb->bi_idl_tree_lrulock );
 		IDL_LRU_DELETE( bdb, matched_idl_entry );
 		IDL_LRU_ADD( bdb, matched_idl_entry );
 		ldap_pvt_thread_mutex_unlock( &bdb->bi_idl_tree_lrulock );
 		if ( matched_idl_entry->idl )
-			return LDAP_SUCCESS;
+			rc = LDAP_SUCCESS;
 		else
-			return DB_NOTFOUND;
+			rc = DB_NOTFOUND;
 	}
 	ldap_pvt_thread_rdwr_runlock( &bdb->bi_idl_tree_rwlock );
 
-	return LDAP_NO_SUCH_OBJECT;
+	return rc;
 }
 
 void
@@ -385,15 +359,9 @@ bdb_idl_cache_put(
 			ee = bdb->bi_idl_lru_tail;
 			if ( avl_delete( &bdb->bi_idl_tree, (caddr_t) ee,
 				    bdb_idl_entry_cmp ) == NULL ) {
-#ifdef NEW_LOGGING
-				LDAP_LOG( INDEX, ERR, 
-					"bdb_idl_cache_put: AVL delete failed\n", 
-					0, 0, 0 );
-#else
 				Debug( LDAP_DEBUG_ANY, "=> bdb_idl_cache_put: "
 					"AVL delete failed\n",
 					0, 0, 0 );
-#endif
 			}
 			IDL_LRU_DELETE( bdb, ee );
 			i++;
@@ -423,15 +391,9 @@ bdb_idl_cache_del(
 	if ( matched_idl_entry != NULL ) {
 		if ( avl_delete( &bdb->bi_idl_tree, (caddr_t) matched_idl_entry,
 				    bdb_idl_entry_cmp ) == NULL ) {
-#ifdef NEW_LOGGING
-			LDAP_LOG( INDEX, ERR, 
-				"bdb_idl_cache_del: AVL delete failed\n", 
-				0, 0, 0 );
-#else
 			Debug( LDAP_DEBUG_ANY, "=> bdb_idl_cache_del: "
 				"AVL delete failed\n",
 				0, 0, 0 );
-#endif
 		}
 		--bdb->bi_idl_cache_size;
 		ldap_pvt_thread_mutex_lock( &bdb->bi_idl_tree_lrulock );
@@ -444,7 +406,6 @@ bdb_idl_cache_del(
 	}
 	ldap_pvt_thread_rdwr_wunlock( &bdb->bi_idl_tree_rwlock );
 }
-#endif
 
 int
 bdb_idl_fetch_key(
@@ -452,17 +413,20 @@ bdb_idl_fetch_key(
 	DB			*db,
 	DB_TXN		*tid,
 	DBT			*key,
-	ID			*ids )
+	ID			*ids,
+	DBC                     **saved_cursor,
+	int                     get_flag )
 {
 	struct bdb_info *bdb = (struct bdb_info *) be->be_private;
 	int rc;
-	DBT data;
+	DBT data, key2, *kptr;
 	DBC *cursor;
 	ID *i;
 	void *ptr;
 	size_t len;
 	int rc2;
 	int flags = bdb->bi_db_opflags | DB_MULTIPLE;
+	int opflag;
 
 	/* If using BerkeleyDB 4.0, the buf must be large enough to
 	 * grab the entire IDL in one get(), otherwise BDB will leak
@@ -474,33 +438,43 @@ bdb_idl_fetch_key(
 	 *
 	 * configure now requires Berkeley DB 4.1.
 	 */
-#if (DB_VERSION_MAJOR == 4) && (DB_VERSION_MINOR == 0)
+#if DB_VERSION_FULL < 0x04010000
 #	define BDB_ENOUGH 5
 #else
+	/* We sometimes test with tiny IDLs, and BDB always wants buffers
+	 * that are at least one page in size.
+	 */
+# if BDB_IDL_DB_SIZE < 4096
+#   define BDB_ENOUGH 2048
+# else
 #	define BDB_ENOUGH 1
+# endif
 #endif
 	ID buf[BDB_IDL_DB_SIZE*BDB_ENOUGH];
 
 	char keybuf[16];
 
-#ifdef NEW_LOGGING
-	LDAP_LOG( INDEX, ARGS,
-		"bdb_idl_fetch_key: %s\n", 
-		bdb_show_key( key, keybuf ), 0, 0 );
-#else
 	Debug( LDAP_DEBUG_ARGS,
 		"bdb_idl_fetch_key: %s\n", 
 		bdb_show_key( key, keybuf ), 0, 0 );
-#endif
 
 	assert( ids != NULL );
 
-#ifdef SLAP_IDL_CACHE
-	if ( bdb->bi_idl_cache_size ) {
+	if ( saved_cursor && *saved_cursor ) {
+		opflag = DB_NEXT;
+	} else if ( get_flag == LDAP_FILTER_GE ) {
+		opflag = DB_SET_RANGE;
+	} else if ( get_flag == LDAP_FILTER_LE ) {
+		opflag = DB_FIRST;
+	} else {
+		opflag = DB_SET;
+	}
+
+	/* only non-range lookups can use the IDL cache */
+	if ( bdb->bi_idl_cache_size && opflag == DB_SET ) {
 		rc = bdb_idl_cache_get( bdb, db, key, ids );
 		if ( rc != LDAP_NO_SUCH_OBJECT ) return rc;
 	}
-#endif
 
 	DBTzero( &data );
 
@@ -508,22 +482,47 @@ bdb_idl_fetch_key(
 	data.ulen = sizeof(buf);
 	data.flags = DB_DBT_USERMEM;
 
-	if ( tid ) flags |= DB_RMW;
-
-	rc = db->cursor( db, tid, &cursor, bdb->bi_db_opflags );
-	if( rc != 0 ) {
-#ifdef NEW_LOGGING
-		LDAP_LOG( INDEX, ERR, 
-			"bdb_idl_fetch_key: cursor failed: %s (%d)\n", 
-			db_strerror(rc), rc, 0 );
-#else
-		Debug( LDAP_DEBUG_ANY, "=> bdb_idl_fetch_key: "
-			"cursor failed: %s (%d)\n", db_strerror(rc), rc, 0 );
-#endif
-		return rc;
+	/* If we're not reusing an existing cursor, get a new one */
+	if( opflag != DB_NEXT ) {
+		rc = db->cursor( db, tid, &cursor, bdb->bi_db_opflags );
+		if( rc != 0 ) {
+			Debug( LDAP_DEBUG_ANY, "=> bdb_idl_fetch_key: "
+				"cursor failed: %s (%d)\n", db_strerror(rc), rc, 0 );
+			return rc;
+		}
+	} else {
+		cursor = *saved_cursor;
 	}
+	
+	/* If this is a LE lookup, save original key so we can determine
+	 * when to stop. If this is a GE lookup, save the key since it
+	 * will be overwritten.
+	 */
+	if ( get_flag == LDAP_FILTER_LE || get_flag == LDAP_FILTER_GE ) {
+		DBTzero( &key2 );
+		key2.flags = DB_DBT_USERMEM;
+		key2.ulen = sizeof(keybuf);
+		key2.data = keybuf;
+		key2.size = key->size;
+		AC_MEMCPY( keybuf, key->data, key->size );
+		kptr = &key2;
+	} else {
+		kptr = key;
+	}
+	len = key->size;
+	rc = cursor->c_get( cursor, kptr, &data, flags | opflag );
 
-	rc = cursor->c_get( cursor, key, &data, flags | DB_SET );
+	/* skip presence key on range inequality lookups */
+	while (rc == 0 && kptr->size != len) {
+		rc = cursor->c_get( cursor, kptr, &data, flags | DB_NEXT_NODUP );
+	}
+	/* If we're doing a LE compare and the new key is greater than
+	 * our search key, we're done
+	 */
+	if (rc == 0 && get_flag == LDAP_FILTER_LE && memcmp( kptr->data,
+		key->data, key->size ) > 0 ) {
+		rc = DB_NOTFOUND;
+	}
 	if (rc == 0) {
 		i = ids;
 		while (rc == 0) {
@@ -534,7 +533,7 @@ bdb_idl_fetch_key(
 				DB_MULTIPLE_NEXT(ptr, &data, j, len);
 				if (j) {
 					++i;
-					AC_MEMCPY( i, j, sizeof(ID) );
+					BDB_DISK2ID( j, i );
 				}
 			}
 			rc = cursor->c_get( cursor, key, &data, flags | DB_NEXT_DUP );
@@ -544,16 +543,9 @@ bdb_idl_fetch_key(
 		/* On disk, a range is denoted by 0 in the first element */
 		if (ids[1] == 0) {
 			if (ids[0] != BDB_IDL_RANGE_SIZE) {
-#ifdef NEW_LOGGING
-				LDAP_LOG( INDEX, ERR, 
-					"=> bdb_idl_fetch_key: range size mismatch: "
-					"expected %ld, got %ld\n", 
-					BDB_IDL_RANGE_SIZE, ids[0], 0 );
-#else
 				Debug( LDAP_DEBUG_ANY, "=> bdb_idl_fetch_key: "
 					"range size mismatch: expected %d, got %ld\n",
 					BDB_IDL_RANGE_SIZE, ids[0], 0 );
-#endif
 				cursor->c_close( cursor );
 				return -1;
 			}
@@ -562,68 +554,46 @@ bdb_idl_fetch_key(
 		data.size = BDB_IDL_SIZEOF(ids);
 	}
 
-	rc2 = cursor->c_close( cursor );
+	if ( saved_cursor && rc == 0 ) {
+		if ( !*saved_cursor )
+			*saved_cursor = cursor;
+		rc2 = 0;
+	}
+	else
+		rc2 = cursor->c_close( cursor );
 	if (rc2) {
-#ifdef NEW_LOGGING
-		LDAP_LOG( INDEX, ERR, 
-			"bdb_idl_fetch_key: close failed: %s (%d)\n", 
-			db_strerror(rc2), rc2, 0 );
-#else
 		Debug( LDAP_DEBUG_ANY, "=> bdb_idl_fetch_key: "
 			"close failed: %s (%d)\n", db_strerror(rc2), rc2, 0 );
-#endif
 		return rc2;
 	}
 
 	if( rc == DB_NOTFOUND ) {
-#ifndef SLAP_IDL_CACHE
 		return rc;
-#endif
 
 	} else if( rc != 0 ) {
-#ifdef NEW_LOGGING
-		LDAP_LOG( INDEX, ERR, 
-			"bdb_idl_fetch_key: get failed: %s (%d)\n", 
-			db_strerror(rc), rc, 0 );
-#else
 		Debug( LDAP_DEBUG_ANY, "=> bdb_idl_fetch_key: "
 			"get failed: %s (%d)\n",
 			db_strerror(rc), rc, 0 );
-#endif
 		return rc;
 
 	} else if ( data.size == 0 || data.size % sizeof( ID ) ) {
 		/* size not multiple of ID size */
-#ifdef NEW_LOGGING
-		LDAP_LOG( INDEX, ERR, 
-			"bdb_idl_fetch_key: odd size: expected %ld multiple, got %ld\n", 
-			(long) sizeof( ID ), (long) data.size, 0 );
-#else
 		Debug( LDAP_DEBUG_ANY, "=> bdb_idl_fetch_key: "
 			"odd size: expected %ld multiple, got %ld\n",
 			(long) sizeof( ID ), (long) data.size, 0 );
-#endif
 		return -1;
 
 	} else if ( data.size != BDB_IDL_SIZEOF(ids) ) {
 		/* size mismatch */
-#ifdef NEW_LOGGING
-		LDAP_LOG( INDEX, ERR, 
-			"bdb_idl_fetch_key: get size mismatch: expected %ld, got %ld\n", 
-			(long) ((1 + ids[0]) * sizeof( ID )), (long) data.size, 0 );
-#else
 		Debug( LDAP_DEBUG_ANY, "=> bdb_idl_fetch_key: "
 			"get size mismatch: expected %ld, got %ld\n",
 			(long) ((1 + ids[0]) * sizeof( ID )), (long) data.size, 0 );
-#endif
 		return -1;
 	}
 
-#ifdef SLAP_IDL_CACHE
 	if ( bdb->bi_idl_cache_max_size ) {
 		bdb_idl_cache_put( bdb, db, key, ids, rc );
 	}
-#endif
 
 	return rc;
 }
@@ -641,55 +611,43 @@ bdb_idl_insert_key(
 	int	rc;
 	DBT data;
 	DBC *cursor;
-	ID lo, hi, tmp;
+	ID lo, hi, nlo, nhi, nid;
 	char *err;
 
 	{
 		char buf[16];
-#ifdef NEW_LOGGING
-		LDAP_LOG( INDEX, ARGS,
-			"bdb_idl_insert_key: %lx %s\n", 
-			(long) id, bdb_show_key( key, buf ), 0 );
-#else
 		Debug( LDAP_DEBUG_ARGS,
 			"bdb_idl_insert_key: %lx %s\n", 
 			(long) id, bdb_show_key( key, buf ), 0 );
-#endif
 	}
 
 	assert( id != NOID );
 
-#ifdef SLAP_IDL_CACHE
 	if ( bdb->bi_idl_cache_size ) {
 		bdb_idl_cache_del( bdb, db, key );
 	}
-#endif
 
 	DBTzero( &data );
 	data.size = sizeof( ID );
 	data.ulen = data.size;
 	data.flags = DB_DBT_USERMEM;
 
+	BDB_ID2DISK( id, &nid );
+
 	rc = db->cursor( db, tid, &cursor, bdb->bi_db_opflags );
 	if ( rc != 0 ) {
-#ifdef NEW_LOGGING
-		LDAP_LOG( INDEX, ERR, 
-			"bdb_idl_insert_key: cursor failed: %s (%d)\n", 
-			db_strerror(rc), rc, 0 );
-#else
 		Debug( LDAP_DEBUG_ANY, "=> bdb_idl_insert_key: "
 			"cursor failed: %s (%d)\n", db_strerror(rc), rc, 0 );
-#endif
 		return rc;
 	}
-	data.data = &tmp;
+	data.data = &nlo;
 	/* Fetch the first data item for this key, to see if it
 	 * exists and if it's a range.
 	 */
-	rc = cursor->c_get( cursor, key, &data, DB_SET | DB_RMW );
+	rc = cursor->c_get( cursor, key, &data, DB_SET );
 	err = "c_get";
 	if ( rc == 0 ) {
-		if ( tmp != 0 ) {
+		if ( nlo != 0 ) {
 			/* not a range, count the number of items */
 			db_recno_t count;
 			rc = cursor->c_count( cursor, &count, 0 );
@@ -700,12 +658,14 @@ bdb_idl_insert_key(
 			if ( count >= BDB_IDL_DB_MAX ) {
 			/* No room, convert to a range */
 				DBT key2 = *key;
+				db_recno_t i;
 
 				key2.dlen = key2.ulen;
 				key2.flags |= DB_DBT_PARTIAL;
 
-				lo = tmp;
-				data.data = &hi;
+				BDB_DISK2ID( &nlo, &lo );
+				data.data = &nhi;
+
 				rc = cursor->c_get( cursor, &key2, &data, DB_NEXT_NODUP );
 				if ( rc != 0 && rc != DB_NOTFOUND ) {
 					err = "c_get next_nodup";
@@ -724,29 +684,60 @@ bdb_idl_insert_key(
 						goto fail;
 					}
 				}
-				if ( id < lo )
+				BDB_DISK2ID( &nhi, &hi );
+				/* Update hi/lo if needed, then delete all the items
+				 * between lo and hi
+				 */
+				if ( id < lo ) {
 					lo = id;
-				else if ( id > hi )
+					nlo = nid;
+				} else if ( id > hi ) {
 					hi = id;
-				rc = db->del( db, tid, key, 0 );
+					nhi = nid;
+				}
+				data.data = &nid;
+				/* Don't fetch anything, just position cursor */
+				data.flags = DB_DBT_USERMEM | DB_DBT_PARTIAL;
+				data.dlen = data.ulen = 0;
+				rc = cursor->c_get( cursor, key, &data, DB_SET );
 				if ( rc != 0 ) {
-					err = "del";
+					err = "c_get 2";
 					goto fail;
 				}
-				data.data = &id;
-				id = 0;
+				rc = cursor->c_del( cursor, 0 );
+				if ( rc != 0 ) {
+					err = "c_del range1";
+					goto fail;
+				}
+				/* Delete all the records */
+				for ( i=1; i<count; i++ ) {
+					rc = cursor->c_get( cursor, &key2, &data, DB_NEXT_DUP );
+					if ( rc != 0 ) {
+						err = "c_get next_dup";
+						goto fail;
+					}
+					rc = cursor->c_del( cursor, 0 );
+					if ( rc != 0 ) {
+						err = "c_del range";
+						goto fail;
+					}
+				}
+				/* Store the range marker */
+				data.size = data.ulen = sizeof(ID);
+				data.flags = DB_DBT_USERMEM;
+				nid = 0;
 				rc = cursor->c_put( cursor, key, &data, DB_KEYFIRST );
 				if ( rc != 0 ) {
-					err = "c_put 0";
+					err = "c_put range";
 					goto fail;
 				}
-				id = lo;
+				nid = nlo;
 				rc = cursor->c_put( cursor, key, &data, DB_KEYLAST );
 				if ( rc != 0 ) {
 					err = "c_put lo";
 					goto fail;
 				}
-				id = hi;
+				nid = nhi;
 				rc = cursor->c_put( cursor, key, &data, DB_KEYLAST );
 				if ( rc != 0 ) {
 					err = "c_put hi";
@@ -761,19 +752,21 @@ bdb_idl_insert_key(
 			 * the boundaries
 			 */
 			hi = id;
-			data.data = &lo;
+			data.data = &nlo;
 			rc = cursor->c_get( cursor, key, &data, DB_NEXT_DUP );
 			if ( rc != 0 ) {
 				err = "c_get lo";
 				goto fail;
 			}
+			BDB_DISK2ID( &nlo, &lo );
 			if ( id > lo ) {
-				data.data = &hi;
+				data.data = &nhi;
 				rc = cursor->c_get( cursor, key, &data, DB_NEXT_DUP );
 				if ( rc != 0 ) {
 					err = "c_get hi";
 					goto fail;
 				}
+				BDB_DISK2ID( &nhi, &hi );
 			}
 			if ( id < lo || id > hi ) {
 				/* Delete the current lo/hi */
@@ -782,7 +775,7 @@ bdb_idl_insert_key(
 					err = "c_del";
 					goto fail;
 				}
-				data.data = &id;
+				data.data = &nid;
 				rc = cursor->c_put( cursor, key, &data, DB_KEYFIRST );
 				if ( rc != 0 ) {
 					err = "c_put lo/hi";
@@ -791,7 +784,7 @@ bdb_idl_insert_key(
 			}
 		}
 	} else if ( rc == DB_NOTFOUND ) {
-put1:		data.data = &id;
+put1:		data.data = &nid;
 		rc = cursor->c_put( cursor, key, &data, DB_NODUPDATA );
 		/* Don't worry if it's already there */
 		if ( rc != 0 && rc != DB_KEYEXIST ) {
@@ -801,28 +794,16 @@ put1:		data.data = &id;
 	} else {
 		/* initial c_get failed, nothing was done */
 fail:
-#ifdef NEW_LOGGING
-		LDAP_LOG( INDEX, ERR, 
-			"bdb_idl_insert_key: %s failed: %s (%d)\n", 
-			err, db_strerror(rc), rc );
-#else
 		Debug( LDAP_DEBUG_ANY, "=> bdb_idl_insert_key: "
 			"%s failed: %s (%d)\n", err, db_strerror(rc), rc );
-#endif
 		cursor->c_close( cursor );
 		return rc;
 	}
 	rc = cursor->c_close( cursor );
 	if( rc != 0 ) {
-#ifdef NEW_LOGGING
-		LDAP_LOG( INDEX, ERR, 
-			"bdb_idl_insert_key: c_close failed: %s (%d)\n", 
-			db_strerror(rc), rc, 0 );
-#else
 		Debug( LDAP_DEBUG_ANY, "=> bdb_idl_insert_key: "
 			"c_close failed: %s (%d)\n",
 			db_strerror(rc), rc, 0 );
-#endif
 	}
 	return rc;
 }
@@ -839,28 +820,22 @@ bdb_idl_delete_key(
 	int	rc;
 	DBT data;
 	DBC *cursor;
-	ID lo, hi, tmp;
+	ID lo, hi, tmp, nid, nlo, nhi;
 	char *err;
 
 	{
 		char buf[16];
-#ifdef NEW_LOGGING
-		LDAP_LOG( INDEX, ARGS,
-			"bdb_idl_delete_key: %lx %s\n", 
-			(long) id, bdb_show_key( key, buf ), 0 );
-#else
 		Debug( LDAP_DEBUG_ARGS,
 			"bdb_idl_delete_key: %lx %s\n", 
 			(long) id, bdb_show_key( key, buf ), 0 );
-#endif
 	}
 	assert( id != NOID );
 
-#ifdef SLAP_IDL_CACHE
 	if ( bdb->bi_idl_cache_max_size ) {
 		bdb_idl_cache_del( bdb, db, key );
 	}
-#endif
+
+	BDB_ID2DISK( id, &nid );
 
 	DBTzero( &data );
 	data.data = &tmp;
@@ -870,29 +845,22 @@ bdb_idl_delete_key(
 
 	rc = db->cursor( db, tid, &cursor, bdb->bi_db_opflags );
 	if ( rc != 0 ) {
-#ifdef NEW_LOGGING
-		LDAP_LOG( INDEX, ERR, 
-			"bdb_idl_delete_key: cursor failed: %s (%d)\n", 
-			db_strerror(rc), rc, 0 );
-#else
 		Debug( LDAP_DEBUG_ANY, "=> bdb_idl_delete_key: "
 			"cursor failed: %s (%d)\n", db_strerror(rc), rc, 0 );
-#endif
 		return rc;
 	}
 	/* Fetch the first data item for this key, to see if it
 	 * exists and if it's a range.
 	 */
-	rc = cursor->c_get( cursor, key, &data, DB_SET | DB_RMW );
+	rc = cursor->c_get( cursor, key, &data, DB_SET );
 	err = "c_get";
 	if ( rc == 0 ) {
 		if ( tmp != 0 ) {
 			/* Not a range, just delete it */
-			if (tmp != id) {
+			if (tmp != nid) {
 				/* position to correct item */
-				tmp = id;
-				rc = cursor->c_get( cursor, key, &data, 
-					DB_GET_BOTH | DB_RMW  );
+				tmp = nid;
+				rc = cursor->c_get( cursor, key, &data, DB_GET_BOTH );
 				if ( rc != 0 ) {
 					err = "c_get id";
 					goto fail;
@@ -907,18 +875,20 @@ bdb_idl_delete_key(
 			/* It's a range, see if we need to rewrite
 			 * the boundaries
 			 */
-			data.data = &lo;
+			data.data = &nlo;
 			rc = cursor->c_get( cursor, key, &data, DB_NEXT_DUP );
 			if ( rc != 0 ) {
 				err = "c_get lo";
 				goto fail;
 			}
-			data.data = &hi;
+			BDB_DISK2ID( &nlo, &lo );
+			data.data = &nhi;
 			rc = cursor->c_get( cursor, key, &data, DB_NEXT_DUP );
 			if ( rc != 0 ) {
 				err = "c_get hi";
 				goto fail;
 			}
+			BDB_DISK2ID( &nhi, &hi );
 			if ( id == lo || id == hi ) {
 				if ( id == lo ) {
 					id++;
@@ -937,9 +907,8 @@ bdb_idl_delete_key(
 				} else {
 					if ( id == lo ) {
 						/* reposition on lo slot */
-						data.data = &lo;
+						data.data = &nlo;
 						cursor->c_get( cursor, key, &data, DB_PREV );
-						lo = id;
 					}
 					rc = cursor->c_del( cursor, 0 );
 					if ( rc != 0 ) {
@@ -948,7 +917,8 @@ bdb_idl_delete_key(
 					}
 				}
 				if ( lo <= hi ) {
-					data.data = &id;
+					BDB_ID2DISK( id, &nid );
+					data.data = &nid;
 					rc = cursor->c_put( cursor, key, &data, DB_KEYFIRST );
 					if ( rc != 0 ) {
 						err = "c_put lo/hi";
@@ -961,28 +931,17 @@ bdb_idl_delete_key(
 		/* initial c_get failed, nothing was done */
 fail:
 		if ( rc != DB_NOTFOUND ) {
-#ifdef NEW_LOGGING
-		LDAP_LOG( INDEX, ERR, 
-			"bdb_idl_delete_key: %s failed: %s (%d)\n", 
-			err, db_strerror(rc), rc );
-#else
 		Debug( LDAP_DEBUG_ANY, "=> bdb_idl_delete_key: "
 			"%s failed: %s (%d)\n", err, db_strerror(rc), rc );
-#endif
 		}
 		cursor->c_close( cursor );
 		return rc;
 	}
 	rc = cursor->c_close( cursor );
 	if( rc != 0 ) {
-#ifdef NEW_LOGGING
-		LDAP_LOG( INDEX, ERR, "bdb_idl_delete_key: c_close failed: %s (%d)\n", 
-			db_strerror(rc), rc, 0 );
-#else
 		Debug( LDAP_DEBUG_ANY,
 			"=> bdb_idl_delete_key: c_close failed: %s (%d)\n",
 			db_strerror(rc), rc, 0 );
-#endif
 	}
 
 	return rc;
