@@ -1,8 +1,8 @@
 /* init.c - initialize bdb backend */
-/* $OpenLDAP: pkg/ldap/servers/slapd/back-bdb/init.c,v 1.132.2.4 2003/12/01 00:58:38 kurt Exp $ */
+/* $OpenLDAP: pkg/ldap/servers/slapd/back-bdb/init.c,v 1.177.2.7 2005/03/17 22:29:52 kurt Exp $ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 2000-2003 The OpenLDAP Foundation.
+ * Copyright 2000-2005 The OpenLDAP Foundation.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -22,8 +22,8 @@
 #include <ac/stdlib.h>
 
 #include "back-bdb.h"
-#include "external.h"
 #include <lutil.h>
+#include <ldap_rq.h>
 
 static const struct bdbi_database {
 	char *file;
@@ -36,55 +36,17 @@ static const struct bdbi_database {
 	{ NULL, NULL, 0, 0 }
 };
 
-struct berval bdb_uuid = { 0, NULL };
-
 typedef void * db_malloc(size_t);
 typedef void * db_realloc(void *, size_t);
 
-#if 0
-static int
-bdb_open( BackendInfo *bi )
-{
-	return 0;
-}
-
-static int
-bdb_destroy( BackendInfo *bi )
-{
-	return 0;
-}
-
-static int
-bdb_close( BackendInfo *bi )
-{
-	/* terminate the underlying database system */
-	return 0;
-}
-#endif
-
 static int
 bdb_db_init( BackendDB *be )
 {
 	struct bdb_info	*bdb;
 
-#ifdef NEW_LOGGING
-	LDAP_LOG( BACK_BDB, ENTRY, "bdb_db_init", 0, 0, 0 );
-#else
 	Debug( LDAP_DEBUG_ANY,
-		"bdb_db_init: Initializing BDB database\n",
-		0, 0, 0 );
-#endif
-
-	/* indicate system schema supported */
-	be->be_flags |=
-		SLAP_BFLAG_INCREMENT |
-#ifdef BDB_SUBENTRIES
-		SLAP_BFLAG_SUBENTRIES |
-#endif
-#ifdef BDB_ALIASES
-		SLAP_BFLAG_ALIASES |
-#endif
-		SLAP_BFLAG_REFERRALS;
+		LDAP_XSTRING(bdb_db_init) ": Initializing "
+		BDB_UCTYPE " database\n", 0, 0, 0 );
 
 	/* allocate backend-database-specific stuff */
 	bdb = (struct bdb_info *) ch_calloc( 1, sizeof(struct bdb_info) );
@@ -100,8 +62,6 @@ bdb_db_init( BackendDB *be )
 	bdb->bi_search_stack_depth = DEFAULT_SEARCH_STACK_DEPTH;
 	bdb->bi_search_stack = NULL;
 
-	LDAP_LIST_INIT (&bdb->bi_psearch_list);
-
 	ldap_pvt_thread_mutex_init( &bdb->bi_database_mutex );
 	ldap_pvt_thread_mutex_init( &bdb->bi_lastid_mutex );
 	ldap_pvt_thread_mutex_init( &bdb->bi_cache.lru_mutex );
@@ -109,34 +69,20 @@ bdb_db_init( BackendDB *be )
 	ldap_pvt_thread_rdwr_init ( &bdb->bi_cache.c_rwlock );
 
 	be->be_private = bdb;
+	be->be_cf_table = be->bd_info->bi_cf_table;
 
 	return 0;
 }
 
-int
-bdb_bt_compare(
-	DB *db, 
-	const DBT *usrkey,
-	const DBT *curkey
-)
+static void *
+bdb_checkpoint( void *ctx, void *arg )
 {
-	unsigned char *u, *c;
-	int i, x;
-
-	u = usrkey->data;
-	c = curkey->data;
-
-#ifdef WORDS_BIGENDIAN
-	for( i = 0; i < (int)sizeof(ID); i++)
-#else
-	for( i = sizeof(ID)-1; i >= 0; i--)
-#endif
-	{
-		x = u[i] - c[i];
-		if( x ) return x;
-	}
-
-	return 0;
+	struct re_s *rtask = arg;
+	struct bdb_info *bdb = rtask->arg;
+	
+	TXN_CHECKPOINT( bdb->bi_dbenv, bdb->bi_txn_cp_kbyte,
+		bdb->bi_txn_cp_min, 0 );
+	return NULL;
 }
 
 static int
@@ -149,24 +95,14 @@ bdb_db_open( BackendDB *be )
 	char path[MAXPATHLEN];
 #endif
 
-#ifdef NEW_LOGGING
-	LDAP_LOG( BACK_BDB, ARGS, 
-		"bdb_db_open: %s\n", be->be_suffix[0].bv_val, 0, 0 );
-#else
 	Debug( LDAP_DEBUG_ARGS,
 		"bdb_db_open: %s\n",
 		be->be_suffix[0].bv_val, 0, 0 );
-#endif
 
 #ifndef BDB_MULTIPLE_SUFFIXES
 	if ( be->be_suffix[1].bv_val ) {
-#ifdef NEW_LOGGING
-	LDAP_LOG( BACK_BDB, ERR, 
-		"bdb_db_open: only one suffix allowed\n", 0, 0, 0 );
-#else
 	Debug( LDAP_DEBUG_ANY,
 		"bdb_db_open: only one suffix allowed\n", 0, 0, 0 );
-#endif
 		return -1;
 	}
 #endif
@@ -174,20 +110,16 @@ bdb_db_open( BackendDB *be )
 
 	rc = db_env_create( &bdb->bi_dbenv, 0 );
 	if( rc != 0 ) {
-#ifdef NEW_LOGGING
-		LDAP_LOG( BACK_BDB, ERR, 
-			"bdb_db_open: db_env_create failed: %s (%d)\n", 
-			db_strerror(rc), rc, 0 );
-#else
 		Debug( LDAP_DEBUG_ANY,
 			"bdb_db_open: db_env_create failed: %s (%d)\n",
 			db_strerror(rc), rc, 0 );
-#endif
 		return rc;
 	}
 
-	flags = DB_INIT_MPOOL | DB_THREAD | DB_CREATE
-		| DB_INIT_LOCK | DB_INIT_LOG | DB_INIT_TXN;
+	flags = DB_INIT_MPOOL | DB_THREAD | DB_CREATE;
+
+	if ( !( slapMode & SLAP_TOOL_QUICK ))
+		flags |= DB_INIT_LOCK | DB_INIT_LOG | DB_INIT_TXN;
 	
 #if 0
 	/* Never do automatic recovery, must perform it manually.
@@ -207,14 +139,25 @@ bdb_db_open( BackendDB *be )
 	bdb->bi_dbenv->set_errcall( bdb->bi_dbenv, bdb_errcall );
 	bdb->bi_dbenv->set_lk_detect( bdb->bi_dbenv, bdb->bi_lock_detect );
 
-#ifdef SLAP_IDL_CACHE
+	/* One long-lived TXN per thread, two TXNs per write op */
+	bdb->bi_dbenv->set_tx_max( bdb->bi_dbenv, connection_pool_max * 3 );
+
+#ifdef SLAP_ZONE_ALLOC
+	if ( bdb->bi_cache.c_maxsize ) {
+		bdb->bi_cache.c_zctx = slap_zn_mem_create(
+								SLAP_ZONE_INITSIZE,
+								SLAP_ZONE_MAXSIZE,
+								SLAP_ZONE_DELTA,
+								SLAP_ZONE_SIZE);
+	}
+#endif
+
 	if ( bdb->bi_idl_cache_max_size ) {
 		bdb->bi_idl_tree = NULL;
 		ldap_pvt_thread_rdwr_init( &bdb->bi_idl_tree_rwlock );
 		ldap_pvt_thread_mutex_init( &bdb->bi_idl_tree_lrulock );
 		bdb->bi_idl_cache_size = 0;
 	}
-#endif
 
 #ifdef BDB_SUBDIRS
 	{
@@ -241,15 +184,9 @@ bdb_db_open( BackendDB *be )
 #endif
 		rc = bdb->bi_dbenv->set_tmp_dir( bdb->bi_dbenv, dir );
 		if( rc != 0 ) {
-#ifdef NEW_LOGGING
-			LDAP_LOG( BACK_BDB, ERR, 
-				"bdb_db_open: set_tmp_dir(%s) failed: %s (%d)\n", 
-				dir, db_strerror(rc), rc );
-#else
 			Debug( LDAP_DEBUG_ANY,
 				"bdb_db_open: set_tmp_dir(%s) failed: %s (%d)\n",
 				dir, db_strerror(rc), rc );
-#endif
 			return rc;
 		}
 
@@ -259,15 +196,9 @@ bdb_db_open( BackendDB *be )
 #endif
 		rc = bdb->bi_dbenv->set_lg_dir( bdb->bi_dbenv, dir );
 		if( rc != 0 ) {
-#ifdef NEW_LOGGING
-			LDAP_LOG( BACK_BDB, ERR, 
-				"bdb_db_open: set_lg_dir(%s) failed: %s (%d)\n", 
-				dir, db_strerror(rc), rc );
-#else
 			Debug( LDAP_DEBUG_ANY,
 				"bdb_db_open: set_lg_dir(%s) failed: %s (%d)\n",
 				dir, db_strerror(rc), rc );
-#endif
 			return rc;
 		}
 
@@ -277,28 +208,28 @@ bdb_db_open( BackendDB *be )
 #endif
 		rc = bdb->bi_dbenv->set_data_dir( bdb->bi_dbenv, dir );
 		if( rc != 0 ) {
-#ifdef NEW_LOGGING
-			LDAP_LOG( BACK_BDB, ERR, 
-				"bdb_db_open: set_data_dir(%s) failed: %s (%d)\n",
-				dir, db_strerror(rc), rc );
-#else
 			Debug( LDAP_DEBUG_ANY,
 				"bdb_db_open: set_data_dir(%s) failed: %s (%d)\n",
 				dir, db_strerror(rc), rc );
-#endif
 			return rc;
 		}
 	}
 #endif
 
-#ifdef NEW_LOGGING
-	LDAP_LOG( BACK_BDB, DETAIL1, 
-		"bdb_db_open: dbenv_open %s\n", bdb->bi_dbenv_home, 0, 0 );
-#else
+	if( bdb->bi_dbenv_xflags != 0 ) {
+		rc = bdb->bi_dbenv->set_flags( bdb->bi_dbenv,
+			bdb->bi_dbenv_xflags, 1);
+		if( rc != 0 ) {
+			Debug( LDAP_DEBUG_ANY,
+				"bdb_db_open: dbenv_set_flags failed: %s (%d)\n",
+				db_strerror(rc), rc, 0 );
+			return rc;
+		}
+	}
+
 	Debug( LDAP_DEBUG_TRACE,
 		"bdb_db_open: dbenv_open(%s)\n",
 		bdb->bi_dbenv_home, 0, 0);
-#endif
 
 #ifdef HAVE_EBCDIC
 	strcpy( path, bdb->bi_dbenv_home );
@@ -314,37 +245,19 @@ bdb_db_open( BackendDB *be )
 		bdb->bi_dbenv_mode );
 #endif
 	if( rc != 0 ) {
-#ifdef NEW_LOGGING
-		LDAP_LOG( BACK_BDB, ERR, 
-			"bdb_db_open: dbenv_open failed: %s (%d)\n", 
-			db_strerror(rc), rc, 0 );
-#else
 		Debug( LDAP_DEBUG_ANY,
 			"bdb_db_open: dbenv_open failed: %s (%d)\n",
 			db_strerror(rc), rc, 0 );
-#endif
 		return rc;
 	}
 
-	if( bdb->bi_dbenv_xflags != 0 ) {
-		rc = bdb->bi_dbenv->set_flags( bdb->bi_dbenv,
-			bdb->bi_dbenv_xflags, 1);
-		if( rc != 0 ) {
-#ifdef NEW_LOGGING
-			LDAP_LOG( BACK_BDB, ERR, 
-				"bdb_db_open: dbenv_set_flags failed: %s (%d)\n", 
-				db_strerror(rc), rc, 0 );
-#else
-			Debug( LDAP_DEBUG_ANY,
-				"bdb_db_open: dbenv_set_flags failed: %s (%d)\n",
-				db_strerror(rc), rc, 0 );
-#endif
-			return rc;
-		}
-	}
-
 	flags = DB_THREAD | bdb->bi_db_opflags;
 
+#ifdef DB_AUTO_COMMIT
+	if ( !( slapMode & SLAP_TOOL_QUICK ))
+		flags |= DB_AUTO_COMMIT;
+#endif
+
 	bdb->bi_databases = (struct bdb_db_info **) ch_malloc(
 		BDB_INDICES * sizeof(struct bdb_db_info *) );
 
@@ -356,21 +269,13 @@ bdb_db_open( BackendDB *be )
 
 		rc = db_create( &db->bdi_db, bdb->bi_dbenv, 0 );
 		if( rc != 0 ) {
-#ifdef NEW_LOGGING
-			LDAP_LOG( BACK_BDB, ERR, 
-				"bdb_db_open: db_create(%s) failed: %s (%d)\n", 
-				bdb->bi_dbenv_home, db_strerror(rc), rc );
-#else
 			Debug( LDAP_DEBUG_ANY,
 				"bdb_db_open: db_create(%s) failed: %s (%d)\n",
 				bdb->bi_dbenv_home, db_strerror(rc), rc );
-#endif
 			return rc;
 		}
 
 		if( i == BDB_ID2ENTRY ) {
-			rc = db->bdi_db->set_bt_compare( db->bdi_db,
-				bdb_bt_compare );
 			rc = db->bdi_db->set_pagesize( db->bdi_db,
 				BDB_ID2ENTRY_PAGESIZE );
 			if ( slapMode & SLAP_TOOL_READMAIN ) {
@@ -382,18 +287,12 @@ bdb_db_open( BackendDB *be )
 			rc = db->bdi_db->set_flags( db->bdi_db, 
 				DB_DUP | DB_DUPSORT );
 #ifndef BDB_HIER
-			rc = db->bdi_db->set_dup_compare( db->bdi_db,
-				bdb_bt_compare );
 			if ( slapMode & SLAP_TOOL_READONLY ) {
 				flags |= DB_RDONLY;
 			} else {
 				flags |= DB_CREATE;
 			}
 #else
-			rc = db->bdi_db->set_dup_compare( db->bdi_db,
-				bdb_dup_compare );
-			rc = db->bdi_db->set_bt_compare( db->bdi_db,
-				bdb_bt_compare );
 			if ( slapMode & (SLAP_TOOL_READONLY|SLAP_TOOL_READMAIN) ) {
 				flags |= DB_RDONLY;
 			} else {
@@ -411,27 +310,25 @@ bdb_db_open( BackendDB *be )
 			path,
 		/*	bdbi_databases[i].name, */ NULL,
 			bdbi_databases[i].type,
-			bdbi_databases[i].flags | flags | DB_AUTO_COMMIT,
+			bdbi_databases[i].flags | flags,
 			bdb->bi_dbenv_mode );
 #else
 		rc = DB_OPEN( db->bdi_db,
 			bdbi_databases[i].file,
 		/*	bdbi_databases[i].name, */ NULL,
 			bdbi_databases[i].type,
-			bdbi_databases[i].flags | flags | DB_AUTO_COMMIT,
+			bdbi_databases[i].flags | flags,
 			bdb->bi_dbenv_mode );
 #endif
 
-		if( rc != 0 ) {
-#ifdef NEW_LOGGING
-			LDAP_LOG( BACK_BDB, ERR, 
-				"bdb_db_open: db_create(%s) failed: %s (%d)\n", 
-				bdb->bi_dbenv_home, db_strerror(rc), rc );
-#else
+		if ( rc != 0 ) {
+			char	buf[SLAP_TEXT_BUFLEN];
+
+			snprintf( buf, sizeof(buf), "%s/%s", 
+				bdb->bi_dbenv_home, bdbi_databases[i].file );
 			Debug( LDAP_DEBUG_ANY,
 				"bdb_db_open: db_open(%s) failed: %s (%d)\n",
-				bdb->bi_dbenv_home, db_strerror(rc), rc );
-#endif
+				buf, db_strerror(rc), rc );
 			return rc;
 		}
 
@@ -446,19 +343,27 @@ bdb_db_open( BackendDB *be )
 	/* get nextid */
 	rc = bdb_last_id( be, NULL );
 	if( rc != 0 ) {
-#ifdef NEW_LOGGING
-			LDAP_LOG( BACK_BDB, ERR, 
-				"bdb_db_open: last_id(%s) failed: %s (%d)\n", 
-				bdb->bi_dbenv_home, db_strerror(rc), rc );
-#else
 		Debug( LDAP_DEBUG_ANY,
 			"bdb_db_open: last_id(%s) failed: %s (%d)\n",
 			bdb->bi_dbenv_home, db_strerror(rc), rc );
-#endif
 		return rc;
 	}
 
-	/* <insert> open (and create) index databases */
+	if ( !( slapMode & SLAP_TOOL_QUICK )) {
+		XLOCK_ID(bdb->bi_dbenv, &bdb->bi_cache.c_locker);
+	}
+
+	/* If we're in server mode and time-based checkpointing is enabled,
+	 * submit a task to perform periodic checkpoints.
+	 */
+	if ( slapMode & SLAP_SERVER_MODE && bdb->bi_txn_cp &&
+		bdb->bi_txn_cp_min )  {
+		ldap_pvt_thread_mutex_lock( &slapd_rq.rq_mutex );
+		ldap_pvt_runqueue_insert( &slapd_rq, bdb->bi_txn_cp_min*60,
+			bdb_checkpoint, bdb );
+		ldap_pvt_thread_mutex_unlock( &slapd_rq.rq_mutex );
+	}
+
 	return 0;
 }
 
@@ -468,9 +373,7 @@ bdb_db_close( BackendDB *be )
 	int rc;
 	struct bdb_info *bdb = (struct bdb_info *) be->be_private;
 	struct bdb_db_info *db;
-#ifdef SLAP_IDL_CACHE
 	bdb_idl_cache_entry_t *entry, *next_entry;
-#endif
 
 	while( bdb->bi_ndatabases-- ) {
 		db = bdb->bi_databases[bdb->bi_ndatabases];
@@ -485,7 +388,6 @@ bdb_db_close( BackendDB *be )
 
 	bdb_cache_release_all (&bdb->bi_cache);
 
-#ifdef SLAP_IDL_CACHE
 	if ( bdb->bi_idl_cache_max_size ) {
 		ldap_pvt_thread_rdwr_wlock ( &bdb->bi_idl_tree_rwlock );
 		avl_free( bdb->bi_idl_tree, NULL );
@@ -500,7 +402,10 @@ bdb_db_close( BackendDB *be )
 		}
 		ldap_pvt_thread_rdwr_wunlock ( &bdb->bi_idl_tree_rwlock );
 	}
-#endif
+
+	if ( !( slapMode & SLAP_TOOL_QUICK )) {
+		XLOCK_ID_FREE(bdb->bi_dbenv, bdb->bi_cache.c_locker);
+	}
 
 	return 0;
 }
@@ -514,31 +419,21 @@ bdb_db_destroy( BackendDB *be )
 	/* close db environment */
 	if( bdb->bi_dbenv ) {
 		/* force a checkpoint */
-		rc = TXN_CHECKPOINT( bdb->bi_dbenv, 0, 0, DB_FORCE );
-		if( rc != 0 ) {
-#ifdef NEW_LOGGING
-			LDAP_LOG( BACK_BDB, ERR, 
-				"bdb_db_destroy: txn_checkpoint failed: %s (%d)\n",
-				db_strerror(rc), rc, 0 );
-#else
-			Debug( LDAP_DEBUG_ANY,
-				"bdb_db_destroy: txn_checkpoint failed: %s (%d)\n",
-				db_strerror(rc), rc, 0 );
-#endif
+		if ( !( slapMode & SLAP_TOOL_QUICK )) {
+			rc = TXN_CHECKPOINT( bdb->bi_dbenv, 0, 0, DB_FORCE );
+			if( rc != 0 ) {
+				Debug( LDAP_DEBUG_ANY,
+					"bdb_db_destroy: txn_checkpoint failed: %s (%d)\n",
+					db_strerror(rc), rc, 0 );
+			}
 		}
 
 		rc = bdb->bi_dbenv->close( bdb->bi_dbenv, 0 );
 		bdb->bi_dbenv = NULL;
 		if( rc != 0 ) {
-#ifdef NEW_LOGGING
-			LDAP_LOG( BACK_BDB, ERR, 
-				"bdb_db_destroy: close failed: %s (%d)\n", 
-				db_strerror(rc), rc, 0 );
-#else
 			Debug( LDAP_DEBUG_ANY,
 				"bdb_db_destroy: close failed: %s (%d)\n",
 				db_strerror(rc), rc, 0 );
-#endif
 			return rc;
 		}
 	}
@@ -550,12 +445,10 @@ bdb_db_destroy( BackendDB *be )
 	ldap_pvt_thread_mutex_destroy( &bdb->bi_cache.c_dntree.bei_kids_mutex );
 	ldap_pvt_thread_mutex_destroy( &bdb->bi_lastid_mutex );
 	ldap_pvt_thread_mutex_destroy( &bdb->bi_database_mutex );
-#ifdef SLAP_IDL_CACHE
 	if ( bdb->bi_idl_cache_max_size ) {
 		ldap_pvt_thread_rdwr_destroy( &bdb->bi_idl_tree_rwlock );
 		ldap_pvt_thread_mutex_destroy( &bdb->bi_idl_tree_lrulock );
 	}
-#endif
 
 	ch_free( bdb );
 	be->be_private = NULL;
@@ -563,55 +456,43 @@ bdb_db_destroy( BackendDB *be )
 	return 0;
 }
 
-#if	(defined(SLAPD_BDB_DYNAMIC) && !defined(BDB_HIER)) || \
-	(defined(SLAPD_HDB_DYNAMIC) && defined(BDB_HIER))
-int init_module( int argc, char *argv[] ) {
-	BackendInfo bi;
-
-	memset( &bi, '\0', sizeof(bi) );
-#ifdef BDB_HIER
-	bi.bi_type = "hdb";
-#else
-	bi.bi_type = "bdb";
-#endif
-	bi.bi_init = bdb_initialize;
-
-	backend_add( &bi );
-	return 0;
-}
-#endif /* SLAPD_BDB_DYNAMIC */
-
 int
-bdb_initialize(
-	BackendInfo	*bi
-)
+bdb_back_initialize(
+	BackendInfo	*bi )
 {
+	int rc;
+
 	static char *controls[] = {
 		LDAP_CONTROL_ASSERT,
 		LDAP_CONTROL_MANAGEDSAIT,
 		LDAP_CONTROL_NOOP,
-#ifdef LDAP_CONTROL_PAGEDRESULTS
 		LDAP_CONTROL_PAGEDRESULTS,
-#endif
 #ifdef LDAP_CONTROL_SUBENTRIES
 		LDAP_CONTROL_SUBENTRIES,
 #endif
- 		LDAP_CONTROL_VALUESRETURNFILTER,
+#ifdef LDAP_CONTROL_X_PERMISSIVE_MODIFY
+		LDAP_CONTROL_X_PERMISSIVE_MODIFY,
+#endif
 		NULL
 	};
 
+	/* initialize the underlying database system */
+	Debug( LDAP_DEBUG_TRACE,
+		LDAP_XSTRING(bdb_back_initialize) ": initialize " 
+		BDB_UCTYPE " backend\n", 0, 0, 0 );
+
+	bi->bi_flags |=
+		SLAP_BFLAG_INCREMENT |
+#ifdef BDB_SUBENTRIES
+		SLAP_BFLAG_SUBENTRIES |
+#endif
+		SLAP_BFLAG_ALIASES |
+		SLAP_BFLAG_REFERRALS;
+
 	bi->bi_controls = controls;
 
-	/* initialize the underlying database system */
-#ifdef NEW_LOGGING
-	LDAP_LOG( BACK_BDB, ENTRY, "bdb_db_initialize\n", 0, 0, 0 );
-#else
-	Debug( LDAP_DEBUG_TRACE, "bdb_initialize: initialize BDB backend\n",
-		0, 0, 0 );
-#endif
-
 	{	/* version check */
-		int major, minor, patch;
+		int major, minor, patch, ver;
 		char *version = db_version( &major, &minor, &patch );
 #ifdef HAVE_EBCDIC
 		char v2[1024];
@@ -625,30 +506,18 @@ bdb_initialize(
 		version = v2;
 #endif
 
-		if( major != DB_VERSION_MAJOR ||
-			minor != DB_VERSION_MINOR ||
-			patch < DB_VERSION_PATCH )
+		ver = (major << 24) | (minor << 16) | patch;
+		if( ver < DB_VERSION_FULL )
 		{
-#ifdef NEW_LOGGING
-			LDAP_LOG( BACK_BDB, ERR, 
-				"bdb_initialize: BDB library version mismatch:"
-				" expected " DB_VERSION_STRING ","
-				" got %s\n", version, 0, 0 );
-#else
 			Debug( LDAP_DEBUG_ANY,
-				"bdb_initialize: BDB library version mismatch:"
+				LDAP_XSTRING(bdb_back_initialize) ": "
+				"BDB library version mismatch:"
 				" expected " DB_VERSION_STRING ","
 				" got %s\n", version, 0, 0 );
-#endif
 		}
 
-#ifdef NEW_LOGGING
-		LDAP_LOG( BACK_BDB, DETAIL1, 
-			"bdb_db_initialize: %s\n", version, 0, 0 );
-#else
-		Debug( LDAP_DEBUG_ANY, "bdb_initialize: %s\n",
-			version, 0, 0 );
-#endif
+		Debug( LDAP_DEBUG_TRACE, LDAP_XSTRING(bdb_back_initialize)
+			": %s\n", version, 0, 0 );
 	}
 
 	db_env_set_func_free( ber_memfree );
@@ -661,20 +530,13 @@ bdb_initialize(
 	db_env_set_func_yield( ldap_pvt_thread_yield );
 #endif
 
-	{
-		static char uuidbuf[ LDAP_LUTIL_UUIDSTR_BUFSIZE ];
-
-		bdb_uuid.bv_len = lutil_uuidstr( uuidbuf, sizeof( uuidbuf ));
-		bdb_uuid.bv_val = uuidbuf;
-	}
-
 	bi->bi_open = 0;
 	bi->bi_close = 0;
 	bi->bi_config = 0;
 	bi->bi_destroy = 0;
 
 	bi->bi_db_init = bdb_db_init;
-	bi->bi_db_config = bdb_db_config;
+	bi->bi_db_config = config_generic_wrapper;
 	bi->bi_db_open = bdb_db_open;
 	bi->bi_db_close = bdb_db_close;
 	bi->bi_db_destroy = bdb_db_destroy;
@@ -689,9 +551,6 @@ bdb_initialize(
 
 	bi->bi_op_unbind = 0;
 
-	bi->bi_op_abandon = bdb_abandon;
-	bi->bi_op_cancel = bdb_cancel;
-
 	bi->bi_extended = bdb_extended;
 
 	bi->bi_chk_referrals = bdb_referrals;
@@ -718,5 +577,20 @@ bdb_initialize(
 	bi->bi_connection_init = 0;
 	bi->bi_connection_destroy = 0;
 
-	return 0;
+	rc = bdb_back_init_cf(bi);
+
+	return rc;
 }
+
+#if	(SLAPD_BDB == SLAPD_MOD_DYNAMIC && !defined(BDB_HIER)) || \
+	(SLAPD_HDB == SLAPD_MOD_DYNAMIC && defined(BDB_HIER))
+
+/* conditionally define the init_module() function */
+#ifdef BDB_HIER
+SLAP_BACKEND_INIT_MODULE( hdb )
+#else /* !BDB_HIER */
+SLAP_BACKEND_INIT_MODULE( bdb )
+#endif /* !BDB_HIER */
+
+#endif /* SLAPD_[BH]DB == SLAPD_MOD_DYNAMIC */
+
