@@ -1,8 +1,8 @@
 /* modrdn.c - bdb backend modrdn routine */
-/* $OpenLDAP: pkg/ldap/servers/slapd/back-bdb/modrdn.c,v 1.110.2.6 2003/12/01 00:58:38 kurt Exp $ */
+/* $OpenLDAP: pkg/ldap/servers/slapd/back-bdb/modrdn.c,v 1.160.2.2 2005/03/14 22:24:59 kurt Exp $ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 2000-2003 The OpenLDAP Foundation.
+ * Copyright 2000-2005 The OpenLDAP Foundation.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -20,7 +20,6 @@
 #include <ac/string.h>
 
 #include "back-bdb.h"
-#include "external.h"
 
 int
 bdb_modrdn( Operation	*op, SlapReply *rs )
@@ -40,10 +39,8 @@ bdb_modrdn( Operation	*op, SlapReply *rs )
 	char textbuf[SLAP_TEXT_BUFLEN];
 	size_t textlen = sizeof textbuf;
 	DB_TXN		*ltid = NULL, *lt2;
-	struct bdb_op_info opinfo;
-	Entry dummy, *save;
-
-	ID			id;
+	struct bdb_op_info opinfo = {0};
+	Entry dummy = {0};
 
 	Entry		*np = NULL;			/* newSuperior Entry */
 	struct berval	*np_dn = NULL;			/* newSuperior dn */
@@ -58,32 +55,30 @@ bdb_modrdn( Operation	*op, SlapReply *rs )
 	u_int32_t	locker = 0;
 	DB_LOCK		lock, plock, nplock;
 
-	int		noop = 0;
-
 	int		num_retries = 0;
 
+	LDAPControl **preread_ctrl = NULL;
+	LDAPControl **postread_ctrl = NULL;
 	LDAPControl *ctrls[SLAP_MAX_RESPONSE_CONTROLS];
 	int num_ctrls = 0;
 
-	Operation *ps_list;
-	struct psid_entry *pm_list, *pm_prev;
 	int	rc;
-	EntryInfo	*suffix_ei;
-	Entry		*ctxcsn_e;
-	int			ctxcsn_added = 0;
 
-#ifdef NEW_LOGGING
-	LDAP_LOG ( OPERATION, ENTRY, "==>bdb_modrdn(%s,%s,%s)\n", 
+	int parent_is_glue = 0;
+	int parent_is_leaf = 0;
+
+	ctrls[num_ctrls] = NULL;
+
+	Debug( LDAP_DEBUG_TRACE, "==>" LDAP_XSTRING(bdb_modrdn) "(%s,%s,%s)\n",
 		op->o_req_dn.bv_val,op->oq_modrdn.rs_newrdn.bv_val,
 		op->oq_modrdn.rs_newSup ? op->oq_modrdn.rs_newSup->bv_val : "NULL" );
-#else
-	Debug( LDAP_DEBUG_TRACE, "==>bdb_modrdn(%s,%s,%s)\n",
-		op->o_req_dn.bv_val,op->oq_modrdn.rs_newrdn.bv_val,
-		op->oq_modrdn.rs_newSup ? op->oq_modrdn.rs_newSup->bv_val : "NULL" );
-#endif
 
 	if( 0 ) {
 retry:	/* transaction retry */
+		if ( dummy.e_attrs ) {
+			attrs_free( dummy.e_attrs );
+			dummy.e_attrs = NULL;
+		}
 		if (e != NULL) {
 			bdb_unlocked_cache_return_entry_w(&bdb->bi_cache, e);
 			e = NULL;
@@ -96,18 +91,8 @@ retry:	/* transaction retry */
 			bdb_unlocked_cache_return_entry_r(&bdb->bi_cache, np);
 			np = NULL;
 		}
-#ifdef NEW_LOGGING
-		LDAP_LOG ( OPERATION, DETAIL1, "==>bdb_modrdn: retrying...\n", 0, 0, 0);
-#else
-		Debug( LDAP_DEBUG_TRACE, "==>bdb_modrdn: retrying...\n", 0, 0, 0 );
-#endif
-		pm_list = LDAP_LIST_FIRST(&op->o_pm_list);
-		while ( pm_list != NULL ) {
-			LDAP_LIST_REMOVE ( pm_list, ps_link );
-			pm_prev = pm_list;
-			pm_list = LDAP_LIST_NEXT ( pm_list, ps_link );
-			ch_free( pm_prev );
-		}
+		Debug( LDAP_DEBUG_TRACE, "==>" LDAP_XSTRING(bdb_modrdn)
+				": retrying...\n", 0, 0, 0 );
 
 		rs->sr_err = TXN_ABORT( ltid );
 		ltid = NULL;
@@ -118,6 +103,8 @@ retry:	/* transaction retry */
 			rs->sr_text = "internal error";
 			goto return_results;
 		}
+		parent_is_glue = 0;
+		parent_is_leaf = 0;
 		ldap_pvt_thread_yield();
 		bdb_trans_backoff( ++num_retries );
 	}
@@ -127,15 +114,9 @@ retry:	/* transaction retry */
 		bdb->bi_db_opflags );
 	rs->sr_text = NULL;
 	if( rs->sr_err != 0 ) {
-#ifdef NEW_LOGGING
-		LDAP_LOG ( OPERATION, ERR, 
-			"==>bdb_modrdn: txn_begin failed: %s (%d)\n", 
-			db_strerror(rs->sr_err), rs->sr_err, 0 );
-#else
 		Debug( LDAP_DEBUG_TRACE,
-			"bdb_delete: txn_begin failed: %s (%d)\n",
-			db_strerror(rs->sr_err), rs->sr_err, 0 );
-#endif
+			LDAP_XSTRING(bdb_modrdn) ": txn_begin failed: "
+			"%s (%d)\n", db_strerror(rs->sr_err), rs->sr_err, 0 );
 		rs->sr_err = LDAP_OTHER;
 		rs->sr_text = "internal error";
 		goto return_results;
@@ -172,7 +153,9 @@ retry:	/* transaction retry */
 
 	e = ei->bei_e;
 	/* FIXME: dn2entry() should return non-glue entry */
-	if (( rs->sr_err == DB_NOTFOUND ) || ( !manageDSAit && e && is_entry_glue( e ))) {
+	if (( rs->sr_err == DB_NOTFOUND ) ||
+		( !manageDSAit && e && is_entry_glue( e )))
+	{
 		if( e != NULL ) {
 			rs->sr_matched = ch_strdup( e->e_dn );
 			rs->sr_ref = is_entry_referral( e )
@@ -182,19 +165,8 @@ retry:	/* transaction retry */
 			e = NULL;
 
 		} else {
-			BerVarray deref = NULL;
-			if ( !LDAP_STAILQ_EMPTY( &op->o_bd->be_syncinfo )) {
-				syncinfo_t *si;
-				LDAP_STAILQ_FOREACH( si, &op->o_bd->be_syncinfo, si_next ) {
-					struct berval tmpbv;
-					ber_dupbv( &tmpbv, &si->si_provideruri_bv[0] );
-					ber_bvarray_add( &deref, &tmpbv );
-                }
-			} else {
-				deref = default_referral;
-			}
-			rs->sr_ref = referral_rewrite( deref, NULL, &op->o_req_dn,
-				LDAP_SCOPE_DEFAULT );
+			rs->sr_ref = referral_rewrite( default_referral, NULL,
+					&op->o_req_dn, LDAP_SCOPE_DEFAULT );
 		}
 
 		rs->sr_err = LDAP_REFERRAL;
@@ -224,13 +196,8 @@ retry:	/* transaction retry */
 			goto retry;
 		}
 
-#ifdef NEW_LOGGING
-		LDAP_LOG ( OPERATION, ERR, 
-			"==>bdb_modrdn: no access to entry\n", 0, 0, 0 );
-#else
 		Debug( LDAP_DEBUG_TRACE, "no access to entry\n", 0,
 			0, 0 );
-#endif
 		rs->sr_text = "no write access to old entry";
 		rs->sr_err = LDAP_INSUFFICIENT_ACCESS;
 		goto return_results;
@@ -244,27 +211,18 @@ retry:	/* transaction retry */
 		case DB_LOCK_NOTGRANTED:
 			goto retry;
 		case 0:
-#ifdef NEW_LOGGING
-			LDAP_LOG ( OPERATION, DETAIL1, 
-				"<=- bdb_modrdn: non-leaf %s\n", op->o_req_dn.bv_val, 0, 0 );
-#else
 			Debug(LDAP_DEBUG_ARGS,
-				"<=- bdb_modrdn: non-leaf %s\n",
+				"<=- " LDAP_XSTRING(bdb_modrdn)
+				": non-leaf %s\n",
 				op->o_req_dn.bv_val, 0, 0);
-#endif
 			rs->sr_err = LDAP_NOT_ALLOWED_ON_NONLEAF;
 			rs->sr_text = "subtree rename not supported";
 			break;
 		default:
-#ifdef NEW_LOGGING
-			LDAP_LOG ( OPERATION, ERR, 
-				"<=- bdb_modrdn: has_children failed %s (%d)\n",
-				db_strerror(rs->sr_err), rs->sr_err, 0 );
-#else
 			Debug(LDAP_DEBUG_ARGS,
-				"<=- bdb_modrdn: has_children failed: %s (%d)\n",
+				"<=- " LDAP_XSTRING(bdb_modrdn)
+				": has_children failed: %s (%d)\n",
 				db_strerror(rs->sr_err), rs->sr_err, 0 );
-#endif
 			rs->sr_err = LDAP_OTHER;
 			rs->sr_text = "internal error";
 		}
@@ -272,17 +230,13 @@ retry:	/* transaction retry */
 	}
 	ei->bei_state |= CACHE_ENTRY_NO_KIDS;
 #endif
+
 	if (!manageDSAit && is_entry_referral( e ) ) {
 		/* parent is a referral, don't allow add */
 		rs->sr_ref = get_entry_referrals( op, e );
 
-#ifdef NEW_LOGGING
-		LDAP_LOG ( OPERATION, DETAIL1, 
-			"==>bdb_modrdn: entry %s is referral \n", e->e_dn, 0, 0 );
-#else
-		Debug( LDAP_DEBUG_TRACE, "bdb_modrdn: entry %s is referral\n",
-			e->e_dn, 0, 0 );
-#endif
+		Debug( LDAP_DEBUG_TRACE, LDAP_XSTRING(bdb_modrdn)
+			": entry %s is referral\n", e->e_dn, 0, 0 );
 
 		rs->sr_err = LDAP_REFERRAL,
 		rs->sr_matched = e->e_name.bv_val;
@@ -326,13 +280,8 @@ retry:	/* transaction retry */
 
 		p = eip->bei_e;
 		if( p == NULL) {
-#ifdef NEW_LOGGING
-			LDAP_LOG ( OPERATION, ERR, 
-				"==>bdb_modrdn: parent does not exist\n", 0, 0, 0 );
-#else
-			Debug( LDAP_DEBUG_TRACE, "bdb_modrdn: parent does not exist\n",
-				0, 0, 0);
-#endif
+			Debug( LDAP_DEBUG_TRACE, LDAP_XSTRING(bdb_modrdn)
+				": parent does not exist\n", 0, 0, 0);
 			rs->sr_err = LDAP_OTHER;
 			rs->sr_text = "old entry's parent does not exist";
 			goto return_results;
@@ -350,25 +299,15 @@ retry:	/* transaction retry */
 			}
 
 			rs->sr_err = LDAP_INSUFFICIENT_ACCESS;
-#ifdef NEW_LOGGING
-			LDAP_LOG ( OPERATION, ERR, 
-				"==>bdb_modrdn: no access to parent\n", 0, 0, 0 );
-#else
 			Debug( LDAP_DEBUG_TRACE, "no access to parent\n", 0,
 				0, 0 );
-#endif
 			rs->sr_text = "no write access to old parent's children";
 			goto return_results;
 		}
 
-#ifdef NEW_LOGGING
-		LDAP_LOG ( OPERATION, DETAIL1, 
-			"==>bdb_modrdn: wr to children %s is OK\n", p_ndn.bv_val, 0, 0 );
-#else
 		Debug( LDAP_DEBUG_TRACE,
-			"bdb_modrdn: wr to children of entry %s OK\n",
-			p_ndn.bv_val, 0, 0 );
-#endif
+			LDAP_XSTRING(bdb_modrdn) ": wr to children "
+			"of entry %s OK\n", p_ndn.bv_val, 0, 0 );
 		
 		if ( p_ndn.bv_val == slap_empty_bv.bv_val ) {
 			p_dn = slap_empty_bv;
@@ -376,21 +315,16 @@ retry:	/* transaction retry */
 			dnParent( &e->e_name, &p_dn );
 		}
 
-#ifdef NEW_LOGGING
-		LDAP_LOG ( OPERATION, DETAIL1, 
-			"==>bdb_modrdn: parent dn=%s\n", p_dn.bv_val, 0, 0 );
-#else
 		Debug( LDAP_DEBUG_TRACE,
-			"bdb_modrdn: parent dn=%s\n",
+			LDAP_XSTRING(bdb_modrdn) ": parent dn=%s\n",
 			p_dn.bv_val, 0, 0 );
-#endif
 
 	} else {
 		/* no parent, modrdn entry directly under root */
-		isroot = be_isroot( op->o_bd, &op->o_ndn );
+		isroot = be_isroot( op );
 		if ( ! isroot ) {
 			if ( be_issuffix( op->o_bd, (struct berval *)&slap_empty_bv )
-				|| be_isupdate( op->o_bd, &op->o_ndn ) ) {
+				|| be_shadow_update( op ) ) {
 
 				p = (Entry *)&slap_entry_root;
 
@@ -408,51 +342,32 @@ retry:	/* transaction retry */
 					}
 
 					rs->sr_err = LDAP_INSUFFICIENT_ACCESS;
-#ifdef NEW_LOGGING
-					LDAP_LOG ( OPERATION, ERR, 
-						"==>bdb_modrdn: no access to parent\n", 0, 0, 0 );
-#else
 					Debug( LDAP_DEBUG_TRACE, 
 						"no access to parent\n", 
 						0, 0, 0 );
-#endif
 					rs->sr_text = "no write access to old parent";
 					goto return_results;
 				}
 
-#ifdef NEW_LOGGING
-				LDAP_LOG ( OPERATION, DETAIL1, 
-					"==>bdb_modrdn: wr to children of entry \"%s\" OK\n", 
-					p_dn.bv_val, 0, 0 );
-#else
 				Debug( LDAP_DEBUG_TRACE,
-					"bdb_modrdn: wr to children of entry \"\" OK\n",
+					LDAP_XSTRING(bdb_modrdn)
+					": wr to children of entry \"\" OK\n",
 					0, 0, 0 );
-#endif
 		
 				p_dn.bv_val = "";
 				p_dn.bv_len = 0;
 
-#ifdef NEW_LOGGING
-				LDAP_LOG ( OPERATION, DETAIL1, 
-					"==>bdb_modrdn: parent dn=\"\" \n", 0, 0, 0 );
-#else
 				Debug( LDAP_DEBUG_TRACE,
-					"bdb_modrdn: parent dn=\"\"\n",
+					LDAP_XSTRING(bdb_modrdn)
+					": parent dn=\"\"\n",
 					0, 0, 0 );
-#endif
 
 			} else {
-#ifdef NEW_LOGGING
-				LDAP_LOG ( OPERATION, ERR, 
-					"==>bdb_modrdn: no parent, not root &\"\" is not "
-					"suffix\n", 0, 0, 0 );
-#else
 				Debug( LDAP_DEBUG_TRACE,
-					"bdb_modrdn: no parent, not root "
+					LDAP_XSTRING(bdb_modrdn)
+					": no parent, not root "
 					"& \"\" is not suffix\n",
 					0, 0, 0);
-#endif
 				rs->sr_text = "no write access to old parent";
 				rs->sr_err = LDAP_INSUFFICIENT_ACCESS;
 				goto return_results;
@@ -463,27 +378,16 @@ retry:	/* transaction retry */
 	new_parent_dn = &p_dn;	/* New Parent unless newSuperior given */
 
 	if ( op->oq_modrdn.rs_newSup != NULL ) {
-#ifdef NEW_LOGGING
-		LDAP_LOG ( OPERATION, DETAIL1, 
-			"==>bdb_modrdn: new parent \"%s\" requested...\n", 
-			op->oq_modrdn.rs_newSup->bv_val, 0, 0 );
-#else
 		Debug( LDAP_DEBUG_TRACE, 
-			"bdb_modrdn: new parent \"%s\" requested...\n",
+			LDAP_XSTRING(bdb_modrdn)
+			": new parent \"%s\" requested...\n",
 			op->oq_modrdn.rs_newSup->bv_val, 0, 0 );
-#endif
 
 		/*  newSuperior == oldParent? */
 		if( dn_match( &p_ndn, op->oq_modrdn.rs_nnewSup ) ) {
-#ifdef NEW_LOGGING
-			LDAP_LOG( BACK_BDB, INFO, "bdb_back_modrdn: "
-				"new parent \"%s\" same as the old parent \"%s\"\n",
-				op->oq_modrdn.rs_newSup->bv_val, p_dn.bv_val, 0 );
-#else
 			Debug( LDAP_DEBUG_TRACE, "bdb_back_modrdn: "
 				"new parent \"%s\" same as the old parent \"%s\"\n",
 				op->oq_modrdn.rs_newSup->bv_val, p_dn.bv_val, 0 );
-#endif      
 			op->oq_modrdn.rs_newSup = NULL; /* ignore newSuperior */
 		}
 	}
@@ -517,29 +421,19 @@ retry:	/* transaction retry */
 			}
 
 			if( np == NULL) {
-#ifdef NEW_LOGGING
-				LDAP_LOG ( OPERATION, DETAIL1, 
-					"==>bdb_modrdn: newSup(ndn=%s) not here!\n", 
-					np_ndn->bv_val, 0, 0 );
-#else
 				Debug( LDAP_DEBUG_TRACE,
-					"bdb_modrdn: newSup(ndn=%s) not here!\n",
+					LDAP_XSTRING(bdb_modrdn)
+					": newSup(ndn=%s) not here!\n",
 					np_ndn->bv_val, 0, 0);
-#endif
 				rs->sr_text = "new superior not found";
 				rs->sr_err = LDAP_OTHER;
 				goto return_results;
 			}
 
-#ifdef NEW_LOGGING
-			LDAP_LOG ( OPERATION, DETAIL1, 
-				"==>bdb_modrdn: wr to new parent OK np=%p, id=%ld\n", 
-				(void *) np, (long) np->e_id, 0 );
-#else
 			Debug( LDAP_DEBUG_TRACE,
-				"bdb_modrdn: wr to new parent OK np=%p, id=%ld\n",
+				LDAP_XSTRING(bdb_modrdn)
+				": wr to new parent OK np=%p, id=%ld\n",
 				(void *) np, (long) np->e_id, 0 );
-#endif
 
 			/* check newSuperior for "children" acl */
 			rs->sr_err = access_allowed( op, np, children,
@@ -552,44 +446,32 @@ retry:	/* transaction retry */
 					goto retry;
 				}
 
-#ifdef NEW_LOGGING
-				LDAP_LOG ( OPERATION, DETAIL1, 
-					"==>bdb_modrdn: no wr to newSup children\n", 0, 0, 0 );
-#else
 				Debug( LDAP_DEBUG_TRACE,
-					"bdb_modrdn: no wr to newSup children\n",
+					LDAP_XSTRING(bdb_modrdn)
+					": no wr to newSup children\n",
 					0, 0, 0 );
-#endif
 				rs->sr_text = "no write access to new superior's children";
 				rs->sr_err = LDAP_INSUFFICIENT_ACCESS;
 				goto return_results;
 			}
 
-#ifdef BDB_ALIASES
 			if ( is_entry_alias( np ) ) {
 				/* parent is an alias, don't allow add */
-#ifdef NEW_LOGGING
-				LDAP_LOG ( OPERATION, DETAIL1, 
-					"==>bdb_modrdn: entry is alias\n", 0, 0, 0 );
-#else
-				Debug( LDAP_DEBUG_TRACE, "bdb_modrdn: entry is alias\n",
+				Debug( LDAP_DEBUG_TRACE,
+					LDAP_XSTRING(bdb_modrdn)
+					": entry is alias\n",
 					0, 0, 0 );
-#endif
 				rs->sr_text = "new superior is an alias";
 				rs->sr_err = LDAP_ALIAS_PROBLEM;
 				goto return_results;
 			}
-#endif
 
 			if ( is_entry_referral( np ) ) {
 				/* parent is a referral, don't allow add */
-#ifdef NEW_LOGGING
-				LDAP_LOG ( OPERATION, DETAIL1, 
-					"==>bdb_modrdn: entry is referral\n", 0, 0, 0 );
-#else
-				Debug( LDAP_DEBUG_TRACE, "bdb_modrdn: entry is referral\n",
+				Debug( LDAP_DEBUG_TRACE,
+					LDAP_XSTRING(bdb_modrdn)
+					": entry is referral\n",
 					0, 0, 0 );
-#endif
 				rs->sr_text = "new superior is a referral";
 				rs->sr_err = LDAP_OTHER;
 				goto return_results;
@@ -597,7 +479,7 @@ retry:	/* transaction retry */
 
 		} else {
 			if ( isroot == -1 ) {
-				isroot = be_isroot( op->o_bd, &op->o_ndn );
+				isroot = be_isroot( op );
 			}
 			
 			np_dn = NULL;
@@ -605,7 +487,7 @@ retry:	/* transaction retry */
 			/* no parent, modrdn entry directly under root */
 			if ( ! isroot ) {
 				if ( be_issuffix( op->o_bd, (struct berval *)&slap_empty_bv )
-					|| be_isupdate( op->o_bd, &op->o_ndn ) ) {
+					|| be_isupdate( op ) ) {
 					np = (Entry *)&slap_entry_root;
 
 					/* check parent for "children" acl */
@@ -622,68 +504,50 @@ retry:	/* transaction retry */
 						}
 
 						rs->sr_err = LDAP_INSUFFICIENT_ACCESS;
-#ifdef NEW_LOGGING
-						LDAP_LOG ( OPERATION, ERR, 
-							"==>bdb_modrdn: no access to superior\n", 0, 0, 0 );
-#else
 						Debug( LDAP_DEBUG_TRACE, 
 							"no access to new superior\n", 
 							0, 0, 0 );
-#endif
-						rs->sr_text = "no write access to new superior's children";
+						rs->sr_text =
+							"no write access to new superior's children";
 						goto return_results;
 					}
 
-#ifdef NEW_LOGGING
-					LDAP_LOG ( OPERATION, DETAIL1, 
-						"bdb_modrdn: wr to children entry \"\" OK\n", 0, 0, 0 );
-#else
 					Debug( LDAP_DEBUG_TRACE,
-						"bdb_modrdn: wr to children of entry \"\" OK\n",
+						LDAP_XSTRING(bdb_modrdn)
+						": wr to children "
+						"of entry \"\" OK\n",
 						0, 0, 0 );
-#endif
 		
 				} else {
-#ifdef NEW_LOGGING
-					LDAP_LOG ( OPERATION, ERR, 
-						"bdb_modrdn: new superior=\"\", not root & \"\" "
-						"is not suffix\n", 0, 0, 0 );
-#else
 					Debug( LDAP_DEBUG_TRACE,
-						"bdb_modrdn: new superior=\"\", not root "
+						LDAP_XSTRING(bdb_modrdn)
+						": new superior=\"\", not root "
 						"& \"\" is not suffix\n",
-						0, 0, 0);
-#endif
+						0, 0, 0 );
 					rs->sr_text = "no write access to new superior's children";
 					rs->sr_err = LDAP_INSUFFICIENT_ACCESS;
 					goto return_results;
 				}
 			}
 
-#ifdef NEW_LOGGING
-			LDAP_LOG ( OPERATION, DETAIL1, 
-				"bdb_modrdn: new superior=\"\"\n", 0, 0, 0 );
-#else
 			Debug( LDAP_DEBUG_TRACE,
-				"bdb_modrdn: new superior=\"\"\n",
+				LDAP_XSTRING(bdb_modrdn)
+				": new superior=\"\"\n",
 				0, 0, 0 );
-#endif
 		}
 
-#ifdef NEW_LOGGING
-		LDAP_LOG ( OPERATION, DETAIL1, 
-			"bdb_modrdn: wr to new parent's children OK\n", 0, 0, 0 );
-#else
 		Debug( LDAP_DEBUG_TRACE,
-			"bdb_modrdn: wr to new parent's children OK\n",
+			LDAP_XSTRING(bdb_modrdn)
+			": wr to new parent's children OK\n",
 			0, 0, 0 );
-#endif
 
 		new_parent_dn = np_dn;
 	}
 
 	/* Build target dn and make sure target entry doesn't exist already. */
-	if (!new_dn.bv_val) build_new_dn( &new_dn, new_parent_dn, &op->oq_modrdn.rs_newrdn, NULL ); 
+	if (!new_dn.bv_val) {
+		build_new_dn( &new_dn, new_parent_dn, &op->oq_modrdn.rs_newrdn, NULL ); 
+	}
 
 	if (!new_ndn.bv_val) {
 		struct berval bv = {0, NULL};
@@ -691,14 +555,8 @@ retry:	/* transaction retry */
 		ber_dupbv( &new_ndn, &bv );
 	}
 
-#ifdef NEW_LOGGING
-	LDAP_LOG ( OPERATION, RESULTS, 
-		"bdb_modrdn: new ndn=%s\n", new_ndn.bv_val, 0, 0 );
-#else
-	Debug( LDAP_DEBUG_TRACE, "bdb_modrdn: new ndn=%s\n",
+	Debug( LDAP_DEBUG_TRACE, LDAP_XSTRING(bdb_modrdn) ": new ndn=%s\n",
 		new_ndn.bv_val, 0, 0 );
-#endif
-
 
 	/* Shortcut the search */
 	nei = neip ? neip : eip;
@@ -722,54 +580,32 @@ retry:	/* transaction retry */
 	/* Get attribute type and attribute value of our new rdn, we will
 	 * need to add that to our new entry
 	 */
-	if ( !new_rdn && ldap_bv2rdn_x( &op->oq_modrdn.rs_newrdn, &new_rdn, (char **)&rs->sr_text,
-		LDAP_DN_FORMAT_LDAP, op->o_tmpmemctx ) )
+	if ( !new_rdn && ldap_bv2rdn_x( &op->oq_modrdn.rs_newrdn, &new_rdn,
+		(char **)&rs->sr_text, LDAP_DN_FORMAT_LDAP, op->o_tmpmemctx ) )
 	{
-#ifdef NEW_LOGGING
-		LDAP_LOG ( OPERATION, ERR, 
-			"bdb_modrdn: can't figure out "
-			"type(s)/values(s) of newrdn\n", 
-			0, 0, 0 );
-#else
 		Debug( LDAP_DEBUG_TRACE,
-			"bdb_modrdn: can't figure out "
+			LDAP_XSTRING(bdb_modrdn) ": can't figure out "
 			"type(s)/values(s) of newrdn\n", 
 			0, 0, 0 );
-#endif
 		rs->sr_err = LDAP_INVALID_DN_SYNTAX;
 		rs->sr_text = "unknown type(s) used in RDN";
 		goto return_results;
 	}
 
-#ifdef NEW_LOGGING
-	LDAP_LOG ( OPERATION, RESULTS, 
-		"bdb_modrdn: new_rdn_type=\"%s\", "
-		"new_rdn_val=\"%s\"\n",
-		new_rdn[ 0 ]->la_attr.bv_val, 
-		new_rdn[ 0 ]->la_value.bv_val, 0 );
-#else
 	Debug( LDAP_DEBUG_TRACE,
-		"bdb_modrdn: new_rdn_type=\"%s\", "
-		"new_rdn_val=\"%s\"\n",
+		LDAP_XSTRING(bdb_modrdn)
+		": new_rdn_type=\"%s\", new_rdn_val=\"%s\"\n",
 		new_rdn[ 0 ]->la_attr.bv_val,
 		new_rdn[ 0 ]->la_value.bv_val, 0 );
-#endif
 
 	if ( op->oq_modrdn.rs_deleteoldrdn ) {
-		if ( !old_rdn && ldap_bv2rdn_x( &op->o_req_dn, &old_rdn, (char **)&rs->sr_text,
-			LDAP_DN_FORMAT_LDAP, op->o_tmpmemctx ) )
+		if ( !old_rdn && ldap_bv2rdn_x( &op->o_req_dn, &old_rdn,
+			(char **)&rs->sr_text, LDAP_DN_FORMAT_LDAP, op->o_tmpmemctx ) )
 		{
-#ifdef NEW_LOGGING
-			LDAP_LOG ( OPERATION, ERR, 
-				"bdb_modrdn: can't figure out "
-				"type(s)/values(s) of old_rdn\n", 
-				0, 0, 0 );
-#else
 			Debug( LDAP_DEBUG_TRACE,
-				"bdb_modrdn: can't figure out "
+				LDAP_XSTRING(bdb_modrdn) ": can't figure out "
 				"the old_rdn type(s)/value(s)\n", 
 				0, 0, 0 );
-#endif
 			rs->sr_err = LDAP_OTHER;
 			rs->sr_text = "cannot parse RDN from old DN";
 			goto return_results;		
@@ -785,56 +621,40 @@ retry:	/* transaction retry */
 	}
 
 	if( op->o_preread ) {
+		if( preread_ctrl == NULL ) {
+			preread_ctrl = &ctrls[num_ctrls++];
+			ctrls[num_ctrls] = NULL;
+		}
 		if( slap_read_controls( op, rs, e,
-			&slap_pre_read_bv, &ctrls[num_ctrls] ) )
+			&slap_pre_read_bv, preread_ctrl ) )
 		{
-#ifdef NEW_LOGGING                                   
-			LDAP_LOG ( OPERATION, DETAIL1,
-				"<=- bdb_modrdn: post-read failed!\n", 0, 0, 0 );
-#else
 			Debug( LDAP_DEBUG_TRACE,        
-				"<=- bdb_modrdn: post-read failed!\n", 0, 0, 0 );
-#endif
+				"<=- " LDAP_XSTRING(bdb_modrdn)
+				": post-read failed!\n", 0, 0, 0 );
 			goto return_results;
 		}                   
-		ctrls[++num_ctrls] = NULL;
-		op->o_preread = 0;  /* prevent redo on retry */
 	}
 
 	/* nested transaction */
-	rs->sr_err = TXN_BEGIN( bdb->bi_dbenv, ltid, &lt2, 
-		bdb->bi_db_opflags );
+	rs->sr_err = TXN_BEGIN( bdb->bi_dbenv, ltid, &lt2, bdb->bi_db_opflags );
 	rs->sr_text = NULL;
 	if( rs->sr_err != 0 ) {
-#ifdef NEW_LOGGING
-		LDAP_LOG ( OPERATION, ERR, 
-			"bdb_modrdn: txn_begin(2) failed: %s (%d)\n", db_strerror(rs->sr_err), rs->sr_err, 0 );
-#else
 		Debug( LDAP_DEBUG_TRACE,
-			"bdb_modrdn: txn_begin(2) failed: %s (%d)\n",
+			LDAP_XSTRING(bdb_modrdn)
+			": txn_begin(2) failed: %s (%d)\n",
 			db_strerror(rs->sr_err), rs->sr_err, 0 );
-#endif
 		rs->sr_err = LDAP_OTHER;
 		rs->sr_text = "internal error";
 		goto return_results;
 	}
 
-	dummy = *e;
-	save = e;
-	e = &dummy;
-
-	/* delete old one */
+	/* delete old DN */
 	rs->sr_err = bdb_dn2id_delete( op, lt2, eip, e );
 	if ( rs->sr_err != 0 ) {
-#ifdef NEW_LOGGING
-		LDAP_LOG ( OPERATION, ERR, 
-			"<=- bdb_modrdn: dn2id del failed: %s (%d)\n",
-			db_strerror(rs->sr_err), rs->sr_err, 0 );
-#else
 		Debug(LDAP_DEBUG_TRACE,
-			"<=- bdb_modrdn: dn2id del failed: %s (%d)\n",
+			"<=- " LDAP_XSTRING(bdb_modrdn)
+			": dn2id del failed: %s (%d)\n",
 			db_strerror(rs->sr_err), rs->sr_err, 0 );
-#endif
 		switch( rs->sr_err ) {
 		case DB_LOCK_DEADLOCK:
 		case DB_LOCK_NOTGRANTED:
@@ -845,33 +665,19 @@ retry:	/* transaction retry */
 		goto return_results;
 	}
 
-	/* Binary format uses a single contiguous block, cannot
-	 * free individual fields. But if a previous modrdn has
-	 * already happened, must free the names. The frees are
-	 * done in bdb_cache_modrdn().
-	 */
-	if( e->e_nname.bv_val < e->e_bv.bv_val || e->e_nname.bv_val >
-		e->e_bv.bv_val + e->e_bv.bv_len ) {
-		e->e_name.bv_val = NULL;
-		e->e_nname.bv_val = NULL;
-	}
-	e->e_name = new_dn;
-	e->e_nname = new_ndn;
-	new_dn.bv_val = NULL;
-	new_ndn.bv_val = NULL;
+	/* copy the entry, then override some fields */
+	dummy = *e;
+	dummy.e_name = new_dn;
+	dummy.e_nname = new_ndn;
+	dummy.e_attrs = NULL;
 
-	/* add new one */
-	rs->sr_err = bdb_dn2id_add( op, lt2, neip ? neip : eip, e );
+	/* add new DN */
+	rs->sr_err = bdb_dn2id_add( op, lt2, neip ? neip : eip, &dummy );
 	if ( rs->sr_err != 0 ) {
-#ifdef NEW_LOGGING
-		LDAP_LOG ( OPERATION, ERR, 
-			"<=- bdb_modrdn: dn2id add failed: %s (%d)\n",
-			db_strerror(rs->sr_err), rs->sr_err, 0 );
-#else
 		Debug(LDAP_DEBUG_TRACE,
-			"<=- bdb_modrdn: dn2id add failed: %s (%d)\n",
+			"<=- " LDAP_XSTRING(bdb_modrdn)
+			": dn2id add failed: %s (%d)\n",
 			db_strerror(rs->sr_err), rs->sr_err, 0 );
-#endif
 		switch( rs->sr_err ) {
 		case DB_LOCK_DEADLOCK:
 		case DB_LOCK_NOTGRANTED:
@@ -882,29 +688,20 @@ retry:	/* transaction retry */
 		goto return_results;
 	}
 
-	if ( rs->sr_err == LDAP_SUCCESS && !op->o_noop && !op->o_no_psearch ) {
-		LDAP_LIST_FOREACH ( ps_list, &bdb->bi_psearch_list, o_ps_link ) {
-			bdb_psearch( op, rs, ps_list, e, LDAP_PSEARCH_BY_PREMODIFY );
-		}
-	}
+	dummy.e_attrs = e->e_attrs;
 
 	/* modify entry */
-	rs->sr_err = bdb_modify_internal( op, lt2, &mod[0], e,
+	rs->sr_err = bdb_modify_internal( op, lt2, &mod[0], &dummy,
 		&rs->sr_text, textbuf, textlen );
-
 	if( rs->sr_err != LDAP_SUCCESS ) {
-#ifdef NEW_LOGGING
-		LDAP_LOG ( OPERATION, ERR, 
-			"<=- bdb_modrdn: modify failed: %s (%d)\n",
-			db_strerror(rs->sr_err), rs->sr_err, 0 );
-#else
 		Debug(LDAP_DEBUG_TRACE,
-			"<=- bdb_modrdn: modify failed: %s (%d)\n",
+			"<=- " LDAP_XSTRING(bdb_modrdn)
+			": modify failed: %s (%d)\n",
 			db_strerror(rs->sr_err), rs->sr_err, 0 );
-#endif
 		if ( ( rs->sr_err == LDAP_INSUFFICIENT_ACCESS ) && opinfo.boi_err ) {
 			rs->sr_err = opinfo.boi_err;
 		}
+		if ( dummy.e_attrs == e->e_attrs ) dummy.e_attrs = NULL;
 		switch( rs->sr_err ) {
 		case DB_LOCK_DEADLOCK:
 		case DB_LOCK_NOTGRANTED:
@@ -913,36 +710,13 @@ retry:	/* transaction retry */
 		goto return_results;
 	}
 
-	if( op->o_postread ) {
-		if( slap_read_controls( op, rs, e,
-			&slap_post_read_bv, &ctrls[num_ctrls] ) )
-		{
-#ifdef NEW_LOGGING                                   
-			LDAP_LOG ( OPERATION, DETAIL1,
-				"<=- bdb_modrdn: post-read failed!\n", 0, 0, 0 );
-#else
-			Debug( LDAP_DEBUG_TRACE,        
-				"<=- bdb_modrdn: post-read failed!\n", 0, 0, 0 );
-#endif
-			goto return_results;
-		}                   
-		ctrls[++num_ctrls] = NULL;
-		op->o_postread = 0;  /* prevent redo on retry */
-		/* FIXME: should read entry on the last retry */
-	}
-
 	/* id2entry index */
-	rs->sr_err = bdb_id2entry_update( op->o_bd, lt2, e );
+	rs->sr_err = bdb_id2entry_update( op->o_bd, lt2, &dummy );
 	if ( rs->sr_err != 0 ) {
-#ifdef NEW_LOGGING
-		LDAP_LOG ( OPERATION, ERR, 
-			"<=- bdb_modrdn: id2entry failed: %s (%d)\n",
-			db_strerror(rs->sr_err), rs->sr_err, 0 );
-#else
 		Debug(LDAP_DEBUG_TRACE,
-			"<=- bdb_modrdn: id2entry failed: %s (%d)\n",
+			"<=- " LDAP_XSTRING(bdb_modrdn)
+			": id2entry failed: %s (%d)\n",
 			db_strerror(rs->sr_err), rs->sr_err, 0 );
-#endif
 		switch( rs->sr_err ) {
 		case DB_LOCK_DEADLOCK:
 		case DB_LOCK_NOTGRANTED:
@@ -952,56 +726,77 @@ retry:	/* transaction retry */
 		rs->sr_text = "entry update failed";
 		goto return_results;
 	}
+
+	if ( p_ndn.bv_len != 0 ) {
+		parent_is_glue = is_entry_glue(p);
+		rs->sr_err = bdb_cache_children( op, lt2, p );
+		if ( rs->sr_err != DB_NOTFOUND ) {
+			switch( rs->sr_err ) {
+			case DB_LOCK_DEADLOCK:
+			case DB_LOCK_NOTGRANTED:
+				goto retry;
+			case 0:
+				break;
+			default:
+				Debug(LDAP_DEBUG_ARGS,
+					"<=- " LDAP_XSTRING(bdb_modrdn)
+					": has_children failed: %s (%d)\n",
+					db_strerror(rs->sr_err), rs->sr_err, 0 );
+				rs->sr_err = LDAP_OTHER;
+				rs->sr_text = "internal error";
+				goto return_results;
+			}
+			parent_is_leaf = 1;
+		}
+		bdb_unlocked_cache_return_entry_r(&bdb->bi_cache, p);
+		p = NULL;
+	}
+
 	if ( TXN_COMMIT( lt2, 0 ) != 0 ) {
 		rs->sr_err = LDAP_OTHER;
 		rs->sr_text = "txn_commit(2) failed";
 		goto return_results;
 	}
 
-	if ( LDAP_STAILQ_EMPTY( &op->o_bd->be_syncinfo )) {
-		rc = bdb_csn_commit( op, rs, ltid, ei, &suffix_ei,
-			&ctxcsn_e, &ctxcsn_added, locker );
-		switch ( rc ) {
-		case BDB_CSN_ABORT :
-			goto return_results;
-		case BDB_CSN_RETRY :
-			goto retry;
+	if( op->o_postread ) {
+		if( postread_ctrl == NULL ) {
+			postread_ctrl = &ctrls[num_ctrls++];
+			ctrls[num_ctrls] = NULL;
 		}
+		if( slap_read_controls( op, rs, &dummy,
+			&slap_post_read_bv, postread_ctrl ) )
+		{
+			Debug( LDAP_DEBUG_TRACE,        
+				"<=- " LDAP_XSTRING(bdb_modrdn)
+				": post-read failed!\n", 0, 0, 0 );
+			goto return_results;
+		}                   
 	}
 
 	if( op->o_noop ) {
 		if(( rs->sr_err=TXN_ABORT( ltid )) != 0 ) {
 			rs->sr_text = "txn_abort (no-op) failed";
 		} else {
-			noop = 1;
 			rs->sr_err = LDAP_SUCCESS;
+			goto return_results;
 		}
 
 	} else {
-		char gid[DB_XIDDATASIZE];
+		rc = bdb_cache_modrdn( e, &op->orr_nnewrdn, &dummy, neip,
+			bdb->bi_dbenv, locker, &lock );
+		switch( rc ) {
+		case DB_LOCK_DEADLOCK:
+		case DB_LOCK_NOTGRANTED:
+			goto retry;
+		}
+		dummy.e_attrs = NULL;
+		new_dn.bv_val = NULL;
+		new_ndn.bv_val = NULL;
 
-		memset( gid, 0, sizeof(gid) );
-		snprintf( gid, sizeof( gid ), "%s-%08lx-%08lx",
-			bdb_uuid.bv_val, (long) op->o_connid, (long) op->o_opid );
-
-		if(( rs->sr_err=TXN_PREPARE( ltid, gid )) != 0 ) {
-			rs->sr_text = "txn_prepare failed";
+		if(( rs->sr_err=TXN_COMMIT( ltid, 0 )) != 0 ) {
+			rs->sr_text = "txn_commit failed";
 		} else {
-			bdb_cache_modrdn( save, &op->orr_nnewrdn, e, neip,
-				bdb->bi_dbenv, locker, &lock );
-
-			if ( LDAP_STAILQ_EMPTY( &op->o_bd->be_syncinfo )) {
-				if ( ctxcsn_added ) {
-					bdb_cache_add( bdb, suffix_ei, ctxcsn_e,
-						(struct berval *)&slap_ldapsync_cn_bv, locker );
-				}
-			}
-
-			if(( rs->sr_err=TXN_COMMIT( ltid, 0 )) != 0 ) {
-				rs->sr_text = "txn_commit failed";
-			} else {
-				rs->sr_err = LDAP_SUCCESS;
-			}
+			rs->sr_err = LDAP_SUCCESS;
 		}
 	}
  
@@ -1009,54 +804,37 @@ retry:	/* transaction retry */
 	op->o_private = NULL;
  
 	if( rs->sr_err != LDAP_SUCCESS ) {
-#ifdef NEW_LOGGING
-		LDAP_LOG ( OPERATION, RESULTS, "bdb_modrdn: %s : %s (%d)\n", 
+		Debug( LDAP_DEBUG_TRACE,
+			LDAP_XSTRING(bdb_modrdn) ": %s : %s (%d)\n",
 			rs->sr_text, db_strerror(rs->sr_err), rs->sr_err );
-#else
-		Debug( LDAP_DEBUG_TRACE, "bdb_add: %s : %s (%d)\n",
-			rs->sr_text, db_strerror(rs->sr_err), rs->sr_err );
-#endif
 		rs->sr_err = LDAP_OTHER;
 
 		goto return_results;
 	}
 
-#ifdef NEW_LOGGING
-	LDAP_LOG ( OPERATION, RESULTS, 
-		"bdb_modrdn: rdn modified%s id=%08lx dn=\"%s\"\n", 
-		op->o_noop ? " (no-op)" : "", e->e_id, e->e_dn );
-#else
 	Debug(LDAP_DEBUG_TRACE,
-		"bdb_modrdn: rdn modified%s id=%08lx dn=\"%s\"\n",
-		op->o_noop ? " (no-op)" : "", e->e_id, e->e_dn );
-#endif
+		LDAP_XSTRING(bdb_modrdn)
+		": rdn modified%s id=%08lx dn=\"%s\"\n",
+		op->o_noop ? " (no-op)" : "",
+		dummy.e_id, op->o_req_dn.bv_val );
 	rs->sr_text = NULL;
 	if( num_ctrls ) rs->sr_ctrls = ctrls;
 
 return_results:
-	send_ldap_result( op, rs );
-
-	if ( rs->sr_err == LDAP_SUCCESS && !op->o_noop ) {
-		/* Loop through in-scope entries for each psearch spec */
-		LDAP_LIST_FOREACH ( ps_list, &bdb->bi_psearch_list, o_ps_link ) {
-			bdb_psearch( op, rs, ps_list, e, LDAP_PSEARCH_BY_MODIFY );
-		}
-		pm_list = LDAP_LIST_FIRST(&op->o_pm_list);
-		while ( pm_list != NULL ) {
-			bdb_psearch(op, rs, pm_list->ps_op,
-						e, LDAP_PSEARCH_BY_SCOPEOUT);
-			pm_prev = pm_list;
-			LDAP_LIST_REMOVE ( pm_list, ps_link );
-			pm_list = LDAP_LIST_NEXT ( pm_list, ps_link );
-			ch_free( pm_prev );
-		}
+	if ( dummy.e_attrs ) {
+		attrs_free( dummy.e_attrs );
 	}
+	send_ldap_result( op, rs );
 
 	if( rs->sr_err == LDAP_SUCCESS && bdb->bi_txn_cp ) {
 		ldap_pvt_thread_yield();
 		TXN_CHECKPOINT( bdb->bi_dbenv,
 			bdb->bi_txn_cp_kbyte, bdb->bi_txn_cp_min, 0 );
 	}
+	
+	if ( rs->sr_err == LDAP_SUCCESS && parent_is_glue && parent_is_leaf ) {
+		op->o_delete_glue_parent = 1;
+	}
 
 done:
 	if( new_dn.bv_val != NULL ) free( new_dn.bv_val );
@@ -1101,16 +879,17 @@ done:
 	}
 
 	if( ltid != NULL ) {
-		pm_list = LDAP_LIST_FIRST(&op->o_pm_list);
-		while ( pm_list != NULL ) {
-			LDAP_LIST_REMOVE ( pm_list, ps_link );
-			pm_prev = pm_list;
-			pm_list = LDAP_LIST_NEXT ( pm_list, ps_link );
-			ch_free( pm_prev );
-		}
 		TXN_ABORT( ltid );
 		op->o_private = NULL;
 	}
 
-	return ( ( rs->sr_err == LDAP_SUCCESS ) ? noop : rs->sr_err );
+	if( preread_ctrl != NULL ) {
+		slap_sl_free( (*preread_ctrl)->ldctl_value.bv_val, op->o_tmpmemctx );
+		slap_sl_free( *preread_ctrl, op->o_tmpmemctx );
+	}
+	if( postread_ctrl != NULL ) {
+		slap_sl_free( (*postread_ctrl)->ldctl_value.bv_val, op->o_tmpmemctx );
+		slap_sl_free( *postread_ctrl, op->o_tmpmemctx );
+	}
+	return rs->sr_err;
 }
