@@ -1,7 +1,7 @@
-/* $OpenLDAP: pkg/ldap/servers/slapd/back-bdb/proto-bdb.h,v 1.77.2.6 2003/12/01 00:58:38 kurt Exp $ */
+/* $OpenLDAP: pkg/ldap/servers/slapd/back-bdb/proto-bdb.h,v 1.111.2.4 2005/03/14 22:24:59 kurt Exp $ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 2000-2003 The OpenLDAP Foundation.
+ * Copyright 2000-2005 The OpenLDAP Foundation.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -20,8 +20,10 @@ LDAP_BEGIN_DECL
 
 #ifdef BDB_HIER
 #define	BDB_SYMBOL(x)	LDAP_CONCAT(hdb_,x)
+#define BDB_UCTYPE	"HDB"
 #else
 #define BDB_SYMBOL(x)	LDAP_CONCAT(bdb_,x)
+#define BDB_UCTYPE	"BDB"
 #endif
 
 /*
@@ -31,6 +33,19 @@ LDAP_BEGIN_DECL
 #define bdb_attr_mask				BDB_SYMBOL(attr_mask)
 #define bdb_attr_index_config		BDB_SYMBOL(attr_index_config)
 #define bdb_attr_index_destroy		BDB_SYMBOL(attr_index_destroy)
+#define bdb_attr_index_unparse		BDB_SYMBOL(attr_index_unparse)
+
+#ifdef LDAP_COMP_MATCH
+#define bdb_attr_comp_ref			BDB_SYMBOL(attr_comp_ref)
+#define bdb_attr_mask_cr			BDB_SYMBOL(attr_mask_cr)
+void bdb_attr_comp_ref( struct bdb_info *bdb,
+	AttributeDescription *desc,
+	ComponentReference **cr );
+void bdb_attr_mask_cr( struct bdb_info *bdb,
+	AttributeDescription *desc,
+	slap_mask_t *indexmask,
+	ComponentReference **cr );
+#endif
 
 void bdb_attr_mask( struct bdb_info *bdb,
 	AttributeDescription *desc,
@@ -40,21 +55,16 @@ int bdb_attr_index_config LDAP_P(( struct bdb_info *bdb,
 	const char *fname, int lineno,
 	int argc, char **argv ));
 
+void bdb_attr_index_unparse LDAP_P(( struct bdb_info *bdb, BerVarray *bva ));
 void bdb_attr_index_destroy LDAP_P(( Avlnode *tree ));
 
 /*
- * ctxcsn.c
+ * config.c
  */
-#define bdb_csn_commit				BDB_SYMBOL(csn_commit)
-#define bdb_get_commit_csn			BDB_SYMBOL(get_commit_csn)
 
-int bdb_csn_commit LDAP_P(( Operation *op, SlapReply *rs, DB_TXN *tid,
-						EntryInfo *ei, EntryInfo **suffix_ei, Entry **ctxcsn_e,
-						int *ctxcsn_added, u_int32_t locker ));
+#define bdb_back_init_cf				BDB_SYMBOL(back_init_cf)
 
-int bdb_get_commit_csn LDAP_P(( Operation *op, SlapReply *rs,
-						struct berval **search_context_csn,
-						u_int32_t locker, DB_LOCK *ctxcsn_lock ));
+int bdb_back_init_cf( BackendInfo *bi );
 
 /*
  * dbcache.c
@@ -116,7 +126,6 @@ int bdb_dn2idl(
 
 #ifdef BDB_HIER
 #define bdb_dn2id_parent			BDB_SYMBOL(dn2id_parent)
-#define bdb_dup_compare				BDB_SYMBOL(dup_compare)
 #define bdb_fix_dn					BDB_SYMBOL(fix_dn)
 
 int bdb_dn2id_parent(
@@ -125,11 +134,6 @@ int bdb_dn2id_parent(
 	EntryInfo *ei,
 	ID *idp );
 
-int bdb_dup_compare(
-	DB *db,
-	const DBT *usrkey,
-	const DBT *curkey );
-
 int bdb_fix_dn( Entry *e, int checkit );
 #endif
 
@@ -139,7 +143,11 @@ int bdb_fix_dn( Entry *e, int checkit );
  */
 #define bdb_errcall					BDB_SYMBOL(errcall)
 
+#if DB_VERSION_FULL < 0x04030000
 void bdb_errcall( const char *pfx, char * msg );
+#else
+void bdb_errcall( const DB_ENV *env, const char *pfx, const char * msg );
+#endif
 
 #ifdef HAVE_EBCDIC
 #define ebcdic_dberror				BDB_SYMBOL(ebcdic_dberror)
@@ -183,11 +191,14 @@ int bdb_id2entry_delete(
 	DB_TXN *tid,
 	Entry *e);
 
+#ifdef SLAP_ZONE_ALLOC
+#else
 int bdb_id2entry(
 	BackendDB *be,
 	DB_TXN *tid,
 	ID id,
 	Entry **e);
+#endif
 
 #define bdb_entry_free				BDB_SYMBOL(entry_free)
 #define bdb_entry_return			BDB_SYMBOL(entry_return)
@@ -195,7 +206,11 @@ int bdb_id2entry(
 #define bdb_entry_get				BDB_SYMBOL(entry_get)
 
 void bdb_entry_free ( Entry *e );
+#ifdef SLAP_ZONE_ALLOC
+int bdb_entry_return( struct bdb_info *bdb, Entry *e, int seqno );
+#else
 int bdb_entry_return( Entry *e );
+#endif
 BI_entry_release_rw bdb_entry_release;
 BI_entry_get_rw bdb_entry_get;
 
@@ -203,7 +218,6 @@ BI_entry_get_rw bdb_entry_get;
 /*
  * idl.c
  */
-#ifdef SLAP_IDL_CACHE
 
 #define bdb_idl_cache_get			BDB_SYMBOL(idl_cache_get)
 #define bdb_idl_cache_put			BDB_SYMBOL(idl_cache_put)
@@ -228,7 +242,6 @@ bdb_idl_cache_del(
 	struct bdb_info	*bdb,
 	DB		*db,
 	DBT		*key );
-#endif
 
 #define bdb_idl_first				BDB_SYMBOL(idl_first)
 #define bdb_idl_next				BDB_SYMBOL(idl_next)
@@ -244,11 +257,13 @@ bdb_idl_cache_del(
 unsigned bdb_idl_search( ID *ids, ID id );
 
 int bdb_idl_fetch_key(
-	BackendDB *be,
-	DB *db,
-	DB_TXN *txn,
-	DBT *key,
-	ID *ids );
+	BackendDB	*be,
+	DB			*db,
+	DB_TXN		*tid,
+	DBT			*key,
+	ID			*ids,
+	DBC                     **saved_cursor,
+	int                     get_flag );
 
 int bdb_idl_insert( ID *ids, ID id );
 
@@ -280,13 +295,6 @@ ID bdb_idl_first( ID *ids, ID *cursor );
 ID bdb_idl_next( ID *ids, ID *cursor );
 
 
-#define bdb_bt_compare				BDB_SYMBOL(bt_compare)
-
-int bdb_bt_compare(
-	DB *db,
-	const DBT *a,
-	const DBT *b );
-
 
 /*
  * index.c
@@ -326,13 +334,6 @@ int bdb_index_entry LDAP_P(( Operation *op, DB_TXN *t, int r, Entry *e ));
 #define bdb_index_entry_del(op,t,e) \
 	bdb_index_entry((op),(t),SLAP_INDEX_DELETE_OP,(e))
 
-/*
- * init.c
- */
-#define bdb_uuid					BDB_SYMBOL(uuid)
-
-extern struct berval bdb_uuid;
-
 /*
  * key.c
  */
@@ -345,7 +346,9 @@ bdb_key_read(
 	DB *db,
 	DB_TXN *txn,
     struct berval *k,
-	ID *ids );
+	ID *ids,
+    DBC **saved_cursor,
+        int get_flags );
 
 extern int
 bdb_key_change(
@@ -379,13 +382,6 @@ int bdb_modify_internal(
 	char *textbuf,
 	size_t textlen );
 
-/*
- * passwd.c
- */
-#define bdb_exop_passwd				BDB_SYMBOL(exop_passwd)
-
-BI_op_extended bdb_exop_passwd;
-
 
 /*
  * cache.c
@@ -426,7 +422,7 @@ void bdb_unlocked_cache_return_entry_rw( Cache *cache, Entry *e, int rw );
 #define bdb_cache_add				BDB_SYMBOL(cache_add)
 #define bdb_cache_children			BDB_SYMBOL(cache_children)
 #define bdb_cache_delete			BDB_SYMBOL(cache_delete)
-#define bdb_cache_delete_cleanup		BDB_SYMBOL(cache_delete_cleanup)
+#define bdb_cache_delete_cleanup	BDB_SYMBOL(cache_delete_cleanup)
 #define bdb_cache_find_id			BDB_SYMBOL(cache_find_id)
 #define bdb_cache_find_info			BDB_SYMBOL(cache_find_info)
 #define bdb_cache_find_ndn			BDB_SYMBOL(cache_find_ndn)
@@ -491,7 +487,7 @@ int bdb_cache_delete(
 );
 void bdb_cache_delete_cleanup(
 	Cache	*cache,
-	Entry	*e
+	EntryInfo *ei
 );
 void bdb_cache_release_all( Cache *cache );
 void bdb_cache_delete_entry(
@@ -537,26 +533,6 @@ int bdb_locker_id( Operation *op, DB_ENV *env, int *locker );
 
 #endif
 
-/*
- * search.c
- */
-
-#define bdb_abandon					BDB_SYMBOL(abandon)
-#define bdb_cancel					BDB_SYMBOL(cancel)
-#define bdb_do_search				BDB_SYMBOL(do_search)
-
-BI_op_abandon bdb_abandon;
-BI_op_cancel bdb_cancel;
-
-int bdb_do_search(
-	Operation       *op,
-	SlapReply	*rs,
-	Operation       *ps_op,
-	Entry           *entry,
-	int             psearch_type
-);
-#define	bdb_psearch(op, rs, sop, e, ps_type)	bdb_do_search(op, rs, sop, e, ps_type)
-
 /*
  * trans.c
  */
@@ -565,6 +541,63 @@ int bdb_do_search(
 void
 bdb_trans_backoff( int num_retries );
 
+/*
+ * former external.h
+ */
+
+#define bdb_back_initialize		BDB_SYMBOL(back_initialize)
+#define bdb_db_config			BDB_SYMBOL(db_config)
+#define bdb_add				BDB_SYMBOL(add)
+#define bdb_bind			BDB_SYMBOL(bind)
+#define bdb_compare			BDB_SYMBOL(compare)
+#define bdb_delete			BDB_SYMBOL(delete)
+#define bdb_modify			BDB_SYMBOL(modify)
+#define bdb_modrdn			BDB_SYMBOL(modrdn)
+#define bdb_search			BDB_SYMBOL(search)
+#define bdb_extended			BDB_SYMBOL(extended)
+#define bdb_referrals			BDB_SYMBOL(referrals)
+#define bdb_operational			BDB_SYMBOL(operational)
+#define bdb_hasSubordinates		BDB_SYMBOL(hasSubordinates)
+#define bdb_tool_entry_open		BDB_SYMBOL(tool_entry_open)
+#define bdb_tool_entry_close		BDB_SYMBOL(tool_entry_close)
+#define bdb_tool_entry_next		BDB_SYMBOL(tool_entry_next)
+#define bdb_tool_entry_get		BDB_SYMBOL(tool_entry_get)
+#define bdb_tool_entry_put		BDB_SYMBOL(tool_entry_put)
+#define bdb_tool_entry_reindex		BDB_SYMBOL(tool_entry_reindex)
+#define bdb_tool_dn2id_get		BDB_SYMBOL(tool_dn2id_get)
+#define bdb_tool_id2entry_get		BDB_SYMBOL(tool_id2entry_get)
+#define bdb_tool_entry_modify		BDB_SYMBOL(tool_entry_modify)
+
+extern BI_init				bdb_back_initialize;
+
+extern BI_db_config			bdb_db_config;
+
+extern BI_op_add			bdb_add;
+extern BI_op_bind			bdb_bind;
+extern BI_op_compare			bdb_compare;
+extern BI_op_delete			bdb_delete;
+extern BI_op_modify			bdb_modify;
+extern BI_op_modrdn			bdb_modrdn;
+extern BI_op_search			bdb_search;
+extern BI_op_extended			bdb_extended;
+
+extern BI_chk_referrals			bdb_referrals;
+
+extern BI_operational			bdb_operational;
+
+extern BI_has_subordinates 		bdb_hasSubordinates;
+
+/* tools.c */
+extern BI_tool_entry_open		bdb_tool_entry_open;
+extern BI_tool_entry_close		bdb_tool_entry_close;
+extern BI_tool_entry_next		bdb_tool_entry_next;
+extern BI_tool_entry_get		bdb_tool_entry_get;
+extern BI_tool_entry_put		bdb_tool_entry_put;
+extern BI_tool_entry_reindex		bdb_tool_entry_reindex;
+extern BI_tool_dn2id_get		bdb_tool_dn2id_get;
+extern BI_tool_id2entry_get		bdb_tool_id2entry_get;
+extern BI_tool_entry_modify		bdb_tool_entry_modify;
+
 LDAP_END_DECL
 
 #endif /* _PROTO_BDB_H */
