@@ -1,8 +1,8 @@
 /* add.c - ldap ldbm back-end add routine */
-/* $OpenLDAP: pkg/ldap/servers/slapd/back-ldbm/add.c,v 1.73.2.4 2003/12/01 00:58:40 kurt Exp $ */
+/* $OpenLDAP: pkg/ldap/servers/slapd/back-ldbm/add.c,v 1.92.2.1 2005/01/20 18:03:59 kurt Exp $ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 1998-2003 The OpenLDAP Foundation.
+ * Copyright 1998-2005 The OpenLDAP Foundation.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -42,26 +42,20 @@ ldbm_back_add(
 	int subentry;
 #endif
 
-#ifdef NEW_LOGGING
-	LDAP_LOG( BACK_LDBM, ENTRY, "ldbm_back_add: %s\n", op->o_req_dn.bv_val, 0, 0 );
-#else
-	Debug(LDAP_DEBUG_ARGS, "==> ldbm_back_add: %s\n", op->o_req_dn.bv_val, 0, 0);
-#endif
+	Debug(LDAP_DEBUG_ARGS, "==> ldbm_back_add: %s\n",
+		op->o_req_dn.bv_val, 0, 0);
 	
-	rs->sr_err = entry_schema_check( op->o_bd, op->oq_add.rs_e, NULL, &rs->sr_text, textbuf, textlen );
+	rs->sr_err = entry_schema_check( op->o_bd, op->oq_add.rs_e, NULL,
+		&rs->sr_text, textbuf, textlen );
 
 	if ( rs->sr_err != LDAP_SUCCESS ) {
-#ifdef NEW_LOGGING
-		LDAP_LOG( BACK_LDBM, ERR, 
-			"ldbm_back_add: entry (%s) failed schema check.\n", op->o_req_dn.bv_val, 0, 0 );
-#else
 		Debug( LDAP_DEBUG_TRACE, "entry failed schema check: %s\n",
 			rs->sr_text, 0, 0 );
-#endif
 
 		send_ldap_result( op, rs );
 		return rs->sr_err;
 	}
+	rs->sr_text = NULL;
 
 #ifdef LDBM_SUBENTRIES
 	subentry = is_entry_subentry( op->oq_add.rs_e );
@@ -70,14 +64,8 @@ ldbm_back_add(
 	if ( !access_allowed( op, op->oq_add.rs_e,
 				entry, NULL, ACL_WRITE, NULL ) )
 	{
-#ifdef NEW_LOGGING
-		LDAP_LOG( BACK_LDBM, ERR, 
-			"ldbm_back_add: No write access to entry (%s).\n", 
-			op->o_req_dn.bv_val, 0, 0 );
-#else
 		Debug( LDAP_DEBUG_TRACE, "no write access to entry\n", 0,
 		    0, 0 );
-#endif
 
 		send_ldap_error( op, rs, LDAP_INSUFFICIENT_ACCESS,
 		    "no write access to entry" );
@@ -88,7 +76,8 @@ ldbm_back_add(
 	/* grab giant lock for writing */
 	ldap_pvt_thread_rdwr_wlock(&li->li_giant_rwlock);
 
-	if ( ( rs->sr_err = dn2id( op->o_bd, &op->o_req_ndn, &id ) ) || id != NOID ) {
+	rs->sr_err = dn2id( op->o_bd, &op->o_req_ndn, &id );
+	if ( rs->sr_err || id != NOID )	{
 		/* if (rs->sr_err) something bad happened to ldbm cache */
 		ldap_pvt_thread_rdwr_wunlock(&li->li_giant_rwlock);
 		rs->sr_err = rs->sr_err ? LDAP_OTHER : LDAP_ALREADY_EXISTS;
@@ -127,16 +116,11 @@ ldbm_back_add(
 
 			ldap_pvt_thread_rdwr_wunlock(&li->li_giant_rwlock);
 
-#ifdef NEW_LOGGING
-			LDAP_LOG( BACK_LDBM, ERR, 
-				"ldbm_back_add: Parent of (%s) does not exist.\n", 
-				op->o_req_dn.bv_val, 0, 0 );
-#else
 			Debug( LDAP_DEBUG_TRACE, "parent does not exist\n",
 				0, 0, 0 );
-#endif
 
-			rs->sr_text = rs->sr_ref ? "parent is referral" : "parent does not exist";
+			rs->sr_text = rs->sr_ref
+				? "parent is referral" : "parent does not exist";
 			rs->sr_err = LDAP_REFERRAL;
 			send_ldap_result( op, rs );
 
@@ -147,21 +131,13 @@ ldbm_back_add(
 			return rs->sr_err;
 		}
 
-		if ( ! access_allowed( op, p,
-			children, NULL, ACL_WRITE, NULL ) )
-		{
+		if ( ! access_allowed( op, p, children, NULL, ACL_WRITE, NULL ) ) {
 			/* free parent and writer lock */
 			cache_return_entry_w( &li->li_cache, p ); 
 			ldap_pvt_thread_rdwr_wunlock(&li->li_giant_rwlock);
 
-#ifdef NEW_LOGGING
-			LDAP_LOG( BACK_LDBM, ERR, 
-				"ldbm_back_add: No write access to parent (%s).\n", 
-				op->o_req_dn.bv_val, 0, 0 );
-#else
 			Debug( LDAP_DEBUG_TRACE, "no write access to parent\n", 0,
 			    0, 0 );
-#endif
 
 			send_ldap_error( op, rs, LDAP_INSUFFICIENT_ACCESS,
 			    "no write access to parent" );
@@ -171,13 +147,8 @@ ldbm_back_add(
 
 #ifdef LDBM_SUBENTRIES
 		if ( is_entry_subentry( p )) {
-#ifdef NEW_LOGGING
-			LDAP_LOG( OPERATION, DETAIL1,
-				"bdb_add: parent is subentry\n", 0, 0, 0 );
-#else
 			Debug( LDAP_DEBUG_TRACE, "bdb_add: parent is subentry\n",
 				0, 0, 0 );
-#endif
 			rs->sr_err = LDAP_OBJECT_CLASS_VIOLATION;
 			rs->sr_text = "parent is a subentry";
 			goto return_results;
@@ -191,13 +162,8 @@ ldbm_back_add(
 			cache_return_entry_w( &li->li_cache, p );
 			ldap_pvt_thread_rdwr_wunlock(&li->li_giant_rwlock);
 
-#ifdef NEW_LOGGING
-			LDAP_LOG(BACK_LDBM, ERR, 
-				"ldbm_back_add:  Parent is an alias.\n", 0, 0, 0 );
-#else
 			Debug( LDAP_DEBUG_TRACE, "parent is alias\n", 0,
 			    0, 0 );
-#endif
 
 			send_ldap_error( op, rs, LDAP_ALIAS_PROBLEM,
 			    "parent is an alias" );
@@ -216,13 +182,8 @@ ldbm_back_add(
 			cache_return_entry_w( &li->li_cache, p );
 			ldap_pvt_thread_rdwr_wunlock(&li->li_giant_rwlock);
 
-#ifdef NEW_LOGGING
-			LDAP_LOG( BACK_LDBM, ERR,
-				   "ldbm_back_add: Parent is referral.\n", 0, 0, 0 );
-#else
 			Debug( LDAP_DEBUG_TRACE, "parent is referral\n", 0,
 			    0, 0 );
-#endif
 			rs->sr_err = LDAP_REFERRAL;
 			send_ldap_result( op, rs );
 
@@ -243,19 +204,13 @@ ldbm_back_add(
 	} else {
 		assert( pdn.bv_val == NULL || *pdn.bv_val == '\0' );
 
-		if ( !be_isroot( op->o_bd, &op->o_ndn )
+		if (( !be_isroot(op) && !be_shadow_update(op) )
 			&& !is_entry_glue( op->oq_add.rs_e ))
 		{
 			ldap_pvt_thread_rdwr_wunlock(&li->li_giant_rwlock);
 
-#ifdef NEW_LOGGING
-			LDAP_LOG( BACK_LDBM, ERR,
-			   "ldbm_back_add: %s add denied.\n",
-			   pdn.bv_val == NULL ? "suffix" : "entry at root", 0, 0 );
-#else
 			Debug( LDAP_DEBUG_TRACE, "%s add denied\n",
 				pdn.bv_val == NULL ? "suffix" : "entry at root", 0, 0 );
-#endif
 
 			send_ldap_error( op, rs, LDAP_NO_SUCH_OBJECT, NULL );
 			return LDAP_NO_SUCH_OBJECT;
@@ -270,13 +225,8 @@ ldbm_back_add(
 
 		ldap_pvt_thread_rdwr_wunlock(&li->li_giant_rwlock);
 
-#ifdef NEW_LOGGING
-		LDAP_LOG( BACK_LDBM, ERR,
-			"ldbm_back_add: next_id failed.\n", 0, 0, 0 );
-#else
 		Debug( LDAP_DEBUG_ANY, "ldbm_add: next_id failed\n",
 			0, 0, 0 );
-#endif
 
 		send_ldap_error( op, rs, LDAP_OTHER,
 			"next_id add failed" );
@@ -287,7 +237,8 @@ ldbm_back_add(
 	/*
 	 * Try to add the entry to the cache, assign it a new dnid.
 	 */
-	rs->sr_err = cache_add_entry_rw(&li->li_cache, op->oq_add.rs_e, CACHE_WRITE_LOCK);
+	rs->sr_err = cache_add_entry_rw( &li->li_cache, op->oq_add.rs_e,
+		CACHE_WRITE_LOCK );
 
 	if ( rs->sr_err != 0 ) {
 		if( p != NULL) {
@@ -297,13 +248,8 @@ ldbm_back_add(
 
 		ldap_pvt_thread_rdwr_wunlock(&li->li_giant_rwlock);
 
-#ifdef NEW_LOGGING
-		LDAP_LOG( BACK_LDBM, ERR,
-			"ldbm_back_add: cache_add_entry_lock failed.\n", 0, 0, 0 );
-#else
 		Debug( LDAP_DEBUG_ANY, "cache_add_entry_lock failed\n", 0, 0,
 		    0 );
-#endif
 
 		rs->sr_text = rs->sr_err > 0 ? NULL : "cache add failed";
 		rs->sr_err = rs->sr_err > 0 ? LDAP_ALREADY_EXISTS : LDAP_OTHER;
@@ -316,13 +262,8 @@ ldbm_back_add(
 
 	/* attribute indexes */
 	if ( index_entry_add( op, op->oq_add.rs_e ) != LDAP_SUCCESS ) {
-#ifdef NEW_LOGGING
-		LDAP_LOG( BACK_LDBM, ERR,
-			"ldbm_back_add: index_entry_add failed.\n", 0, 0, 0 );
-#else
 		Debug( LDAP_DEBUG_TRACE, "index_entry_add failed\n", 0,
 		    0, 0 );
-#endif
 		
 		send_ldap_error( op, rs, LDAP_OTHER,
 			"index generation failed" );
@@ -331,14 +272,11 @@ ldbm_back_add(
 	}
 
 	/* dn2id index */
-	if ( dn2id_add( op->o_bd, &op->oq_add.rs_e->e_nname, op->oq_add.rs_e->e_id ) != 0 ) {
-#ifdef NEW_LOGGING
-		LDAP_LOG( BACK_LDBM, ERR,
-			"ldbm_back_add: dn2id_add failed.\n", 0, 0, 0 );
-#else
+	if ( dn2id_add( op->o_bd, &op->oq_add.rs_e->e_nname,
+		op->oq_add.rs_e->e_id ) != 0 )
+	{
 		Debug( LDAP_DEBUG_TRACE, "dn2id_add failed\n", 0,
 		    0, 0 );
-#endif
 		/* FIXME: delete attr indices? */
 
 		send_ldap_error( op, rs, LDAP_OTHER,
@@ -349,16 +287,12 @@ ldbm_back_add(
 
 	/* id2entry index */
 	if ( id2entry_add( op->o_bd, op->oq_add.rs_e ) != 0 ) {
-#ifdef NEW_LOGGING
-		LDAP_LOG( BACK_LDBM, ERR,
-			   "ldbm_back_add: id2entry_add failed.\n", 0, 0, 0 );
-#else
 		Debug( LDAP_DEBUG_TRACE, "id2entry_add failed\n", 0,
 		    0, 0 );
-#endif
 
 		/* FIXME: delete attr indices? */
-		(void) dn2id_delete( op->o_bd, &op->oq_add.rs_e->e_nname, op->oq_add.rs_e->e_id );
+		(void) dn2id_delete( op->o_bd, &op->oq_add.rs_e->e_nname,
+			op->oq_add.rs_e->e_id );
 		
 		send_ldap_error( op, rs, LDAP_OTHER,
 			"entry store failed" );
