@@ -1,8 +1,17 @@
 /* close.c - close ldbm backend */
-/* $OpenLDAP: pkg/ldap/servers/slapd/back-ldbm/close.c,v 1.16 2003/01/03 19:20:57 kurt Exp $ */
-/*
- * Copyright 1998-2003 The OpenLDAP Foundation, All Rights Reserved.
- * COPYING RESTRICTIONS APPLY, see COPYRIGHT file
+/* $OpenLDAP: pkg/ldap/servers/slapd/back-ldbm/close.c,v 1.19.2.1 2005/01/20 18:03:59 kurt Exp $ */
+/* This work is part of OpenLDAP Software <http://www.openldap.org/>.
+ *
+ * Copyright 1998-2005 The OpenLDAP Foundation.
+ * All rights reserved.
+ *
+ * Redistribution and use in source and binary forms, with or without
+ * modification, are permitted only as authorized by the OpenLDAP
+ * Public License.
+ *
+ * A copy of this license is available in the file LICENSE in the
+ * top-level directory of the distribution or, alternatively, at
+ * <http://www.OpenLDAP.org/license.html>.
  */
 
 #include "portable.h"
@@ -23,20 +32,10 @@ ldbm_back_db_close( Backend *be )
 		li->li_dbshutdown++;
 		ldap_pvt_thread_join( li->li_dbsynctid, (void *) NULL );
 	}
-#ifdef NEW_LOGGING
-	LDAP_LOG( BACK_LDBM, CRIT,
-		   "ldbm_back_db_close: ldbm backend syncing\n", 0, 0, 0 );
-#else
 	Debug( LDAP_DEBUG_TRACE, "ldbm backend syncing\n", 0, 0, 0 );
-#endif
 
 	ldbm_cache_flush_all( be );
-#ifdef NEW_LOGGING
-	LDAP_LOG( BACK_LDBM, CRIT,
-		   "ldbm_back_db_close: ldbm backend synch'ed\n", 0, 0, 0 );
-#else
 	Debug( LDAP_DEBUG_TRACE, "ldbm backend done syncing\n", 0, 0, 0 );
-#endif
 
 
 	cache_release_all( &((struct ldbminfo *) be->be_private)->li_cache );
