@@ -1,7 +1,9 @@
-/* $OpenLDAP: pkg/ldap/servers/slapd/back-meta/conn.c,v 1.24.2.1 2003/12/01 00:58:41 kurt Exp $ */
+/* $OpenLDAP: pkg/ldap/servers/slapd/back-meta/conn.c,v 1.31.2.2 2005/03/14 22:25:01 kurt Exp $ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 1999-2003 The OpenLDAP Foundation.
+ * Copyright 1999-2005 The OpenLDAP Foundation.
+ * Portions Copyright 2001-2003 Pierangelo Masarati.
+ * Portions Copyright 1999-2003 Howard Chu.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -17,69 +19,6 @@
  * in OpenLDAP Software and subsequently enhanced by Pierangelo
  * Masarati.
  */
-/* This is an altered version */
-/*
- * Copyright 2001, Pierangelo Masarati, All rights reserved. <ando@sys-net.it>
- *
- * This work has been developed to fulfill the requirements
- * of SysNet s.n.c. <http:www.sys-net.it> and it has been donated
- * to the OpenLDAP Foundation in the hope that it may be useful
- * to the Open Source community, but WITHOUT ANY WARRANTY.
- *
- * Permission is granted to anyone to use this software for any purpose
- * on any computer system, and to alter it and redistribute it, subject
- * to the following restrictions:
- *
- * 1. The author and SysNet s.n.c. are not responsible for the consequences
- *    of use of this software, no matter how awful, even if they arise from 
- *    flaws in it.
- *
- * 2. The origin of this software must not be misrepresented, either by
- *    explicit claim or by omission.  Since few users ever read sources,
- *    credits should appear in the documentation.
- *
- * 3. Altered versions must be plainly marked as such, and must not be
- *    misrepresented as being the original software.  Since few users
- *    ever read sources, credits should appear in the documentation.
- *    SysNet s.n.c. cannot be responsible for the consequences of the
- *    alterations.
- *
- * 4. This notice may not be removed or altered.
- *
- *
- * This software is based on the backend back-ldap, implemented
- * by Howard Chu <hyc@highlandsun.com>, and modified by Mark Valence
- * <kurash@sassafras.com>, Pierangelo Masarati <ando@sys-net.it> and other
- * contributors. The contribution of the original software to the present
- * implementation is acknowledged in this copyright statement.
- *
- * A special acknowledgement goes to Howard for the overall architecture
- * (and for borrowing large pieces of code), and to Mark, who implemented
- * from scratch the attribute/objectclass mapping.
- *
- * The original copyright statement follows.
- *
- * Copyright 1999, Howard Chu, All rights reserved. <hyc@highlandsun.com>
- *
- * Permission is granted to anyone to use this software for any purpose
- * on any computer system, and to alter it and redistribute it, subject
- * to the following restrictions:
- *
- * 1. The author is not responsible for the consequences of use of this
- *    software, no matter how awful, even if they arise from flaws in it.
- *
- * 2. The origin of this software must not be misrepresented, either by
- *    explicit claim or by omission.  Since few users ever read sources,
- *    credits should appear in the documentation.
- *
- * 3. Altered versions must be plainly marked as such, and must not be
- *    misrepresented as being the original software.  Since few users
- *    ever read sources, credits should appear in the
- *    documentation.
- *
- * 4. This notice may not be removed or altered.
- *
- */
 
 #include "portable.h"
 
@@ -114,7 +53,7 @@ meta_back_conn_cmp(
 	struct metaconn *lc1 = ( struct metaconn * )c1;
         struct metaconn *lc2 = ( struct metaconn * )c2;
 	
-	return SLAP_PTRCMP( lc1->conn, lc2->conn );
+	return SLAP_PTRCMP( lc1->mc_conn, lc2->mc_conn );
 }
 
 /*
@@ -132,7 +71,7 @@ meta_back_conn_dup(
 	struct metaconn *lc1 = ( struct metaconn * )c1;
 	struct metaconn *lc2 = ( struct metaconn * )c2;
 
-	return( ( lc1->conn == lc2->conn ) ? -1 : 0 );
+	return( ( lc1->mc_conn == lc2->mc_conn ) ? -1 : 0 );
 }
 
 /*
@@ -148,15 +87,15 @@ ravl_print( Avlnode *root, int depth )
 		return;
 	}
 	
-	ravl_print( root->avl_right, depth+1 );
+	ravl_print( root->avl_right, depth + 1 );
 	
 	for ( i = 0; i < depth; i++ ) {
 		printf( "    " );
 	}
 
-	printf( "c(%d) %d\n", ( ( struct metaconn * )root->avl_data )->conn->c_connid, root->avl_bf );
+	printf( "c(%d) %d\n", ( ( struct metaconn * )root->avl_data )->mc_conn->c_connid, root->avl_bf );
 	
-	ravl_print( root->avl_left, depth+1 );
+	ravl_print( root->avl_left, depth + 1 );
 }
 
 static void
@@ -197,23 +136,21 @@ metaconn_alloc( int ntargets )
 	/*
 	 * make it a null-terminated array ...
 	 */
-	lc->conns = ch_calloc( sizeof( struct metasingleconn ), ntargets+1 );
-	if ( lc->conns == NULL ) {
+	lc->mc_conns = ch_calloc( sizeof( struct metasingleconn ), ntargets+1 );
+	if ( lc->mc_conns == NULL ) {
 		free( lc );
 		return NULL;
 	}
-	lc->conns[ ntargets ].candidate = META_LAST_CONN;
+	lc->mc_conns[ ntargets ].msc_candidate = META_LAST_CONN;
 
 	for ( ; ntargets-- > 0; ) {
-		lc->conns[ ntargets ].ld = NULL;
-		lc->conns[ ntargets ].bound_dn.bv_val = NULL;
-		lc->conns[ ntargets ].bound_dn.bv_len = 0;
-		lc->conns[ ntargets ].cred.bv_val = NULL;
-		lc->conns[ ntargets ].cred.bv_len = 0;
-		lc->conns[ ntargets ].bound = META_UNBOUND;
+		lc->mc_conns[ ntargets ].msc_ld = NULL;
+		BER_BVZERO( &lc->mc_conns[ ntargets ].msc_bound_ndn );
+		BER_BVZERO( &lc->mc_conns[ ntargets ].msc_cred );
+		lc->mc_conns[ ntargets ].msc_bound = META_UNBOUND;
 	}
 
-	lc->bound_target = META_BOUND_NONE;
+	lc->mc_bound_target = META_BOUND_NONE;
 
 	return lc;
 }
@@ -232,8 +169,8 @@ metaconn_free(
 		return;
 	}
 	
-	if ( lc->conns ) {
-		ch_free( lc->conns );
+	if ( lc->mc_conns ) {
+		ch_free( lc->mc_conns );
 	}
 
 	free( lc );
@@ -249,8 +186,8 @@ init_one_conn(
 		Operation		*op,
 		SlapReply		*rs,
 		struct metatarget	*lt, 
-		struct metasingleconn	*lsc
-		)
+		struct metasingleconn	*lsc,
+		ldap_back_send_t	sendok )
 {
 	struct metainfo	*li = ( struct metainfo * )op->o_bd->be_private;
 	int		vers;
@@ -259,16 +196,16 @@ init_one_conn(
 	/*
 	 * Already init'ed
 	 */
-	if ( lsc->ld != NULL ) {
+	if ( lsc->msc_ld != NULL ) {
 		return LDAP_SUCCESS;
 	}
        
 	/*
 	 * Attempts to initialize the connection to the target ds
 	 */
-	rs->sr_err = ldap_initialize( &lsc->ld, lt->uri );
+	rs->sr_err = ldap_initialize( &lsc->msc_ld, lt->mt_uri );
 	if ( rs->sr_err != LDAP_SUCCESS ) {
-		return ldap_back_map_result( rs );
+		goto error_return;
 	}
 
 	/*
@@ -276,58 +213,168 @@ init_one_conn(
 	 * bound with a particular version, then so can we.
 	 */
 	vers = op->o_conn->c_protocol;
-	ldap_set_option( lsc->ld, LDAP_OPT_PROTOCOL_VERSION, &vers );
-	/* FIXME: configurable? */
-	ldap_set_option(lsc->ld, LDAP_OPT_REFERRALS, LDAP_OPT_ON);
+	ldap_set_option( lsc->msc_ld, LDAP_OPT_PROTOCOL_VERSION, &vers );
+
+	/* automatically chase referrals ("chase-referrals"/"dont-chase-referrals" statement) */
+	if ( LDAP_BACK_CHASE_REFERRALS( li ) ) {
+		ldap_set_option( lsc->msc_ld, LDAP_OPT_REFERRALS, LDAP_OPT_ON );
+	}
+
+#ifdef HAVE_TLS
+	/* start TLS ("start-tls"/"try-start-tls" statements) */
+	if ( ( LDAP_BACK_USE_TLS( li ) || ( op->o_conn->c_is_tls && LDAP_BACK_PROPAGATE_TLS( li ) ) )
+			&& !ldap_is_ldaps_url( lt->mt_uri ) )
+	{
+#if 1
+		/*
+		 * use asynchronous StartTLS
+		 * in case, chase referral (not implemented yet)
+		 */
+		int		msgid;
+
+		rs->sr_err = ldap_start_tls( lsc->msc_ld, NULL, NULL, &msgid );
+		if ( rs->sr_err == LDAP_SUCCESS ) {
+			LDAPMessage	*res = NULL;
+			int		rc, retries = 1;
+			struct timeval	tv = { 0, 0 };
+
+retry:;
+			rc = ldap_result( lsc->msc_ld, msgid, LDAP_MSG_ALL, &tv, &res );
+			if ( rc < 0 ) {
+				rs->sr_err = LDAP_OTHER;
+
+			} else if ( rc == 0 ) {
+				if ( retries ) {
+					retries--;
+					tv.tv_sec = 0;
+					tv.tv_usec = 100000;
+					goto retry;
+				}
+				rs->sr_err = LDAP_OTHER;
+
+			} else if ( rc == LDAP_RES_EXTENDED ) {
+				struct berval	*data = NULL;
+
+				rs->sr_err = ldap_parse_extended_result( lsc->msc_ld, res,
+						NULL, &data, 0 );
+				if ( rs->sr_err == LDAP_SUCCESS ) {
+					rs->sr_err = ldap_result2error( lsc->msc_ld, res, 1 );
+					res = NULL;
+					
+					/* FIXME: in case a referral 
+					 * is returned, should we try
+					 * using it instead of the 
+					 * configured URI? */
+					if ( rs->sr_err == LDAP_SUCCESS ) {
+						ldap_install_tls( lsc->msc_ld );
+
+					} else if ( rs->sr_err == LDAP_REFERRAL ) {
+						rs->sr_err = LDAP_OTHER;
+						rs->sr_text = "unwilling to chase referral returned by Start TLS exop";
+					}
+
+					if ( data ) {
+						if ( data->bv_val ) {
+							ber_memfree( data->bv_val );
+						}
+						ber_memfree( data );
+					}
+				}
+
+			} else {
+				rs->sr_err = LDAP_OTHER;
+			}
+
+			if ( res != NULL ) {
+				ldap_msgfree( res );
+			}
+		}
+#else
+		/*
+		 * use synchronous StartTLS
+		 */
+		rs->sr_err = ldap_start_tls_s( lsc->msc_ld, NULL, NULL );
+#endif
+
+		/* if StartTLS is requested, only attempt it if the URL
+		 * is not "ldaps://"; this may occur not only in case
+		 * of misconfiguration, but also when used in the chain 
+		 * overlay, where the "uri" can be parsed out of a referral */
+		if ( rs->sr_err == LDAP_SERVER_DOWN
+				|| ( rs->sr_err != LDAP_SUCCESS && LDAP_BACK_TLS_CRITICAL( li ) ) )
+		{
+			ldap_unbind_ext_s( lsc->msc_ld, NULL, NULL );
+			goto error_return;
+		}
+	}
+#endif /* HAVE_TLS */
 
 	/*
 	 * Set the network timeout if set
 	 */
 	if (li->network_timeout != 0){
-		struct timeval network_timeout;
+		struct timeval	network_timeout;
 
 		network_timeout.tv_usec = 0;
 		network_timeout.tv_sec = li->network_timeout;
 
-		ldap_set_option( lsc->ld, LDAP_OPT_NETWORK_TIMEOUT, (void *) &network_timeout);
+		ldap_set_option( lsc->msc_ld, LDAP_OPT_NETWORK_TIMEOUT,
+				(void *)&network_timeout );
 	}
 
 	/*
 	 * Sets a cookie for the rewrite session
 	 */
-	( void )rewrite_session_init( lt->rwmap.rwm_rw, op->o_conn );
+	( void )rewrite_session_init( lt->mt_rwmap.rwm_rw, op->o_conn );
 
 	/*
-	 * If the connection dn is not null, an attempt to rewrite it is made
+	 * If the connection DN is not null, an attempt to rewrite it is made
 	 */
-	if ( op->o_conn->c_dn.bv_len != 0 ) {
-		dc.rwmap = &lt->rwmap;
+	if ( !BER_BVISEMPTY( &op->o_conn->c_dn ) ) {
+		dc.rwmap = &lt->mt_rwmap;
 		dc.conn = op->o_conn;
 		dc.rs = rs;
-		dc.ctx = "bindDn";
+		dc.ctx = "bindDN";
 		
 		/*
 		 * Rewrite the bind dn if needed
 		 */
 		if ( ldap_back_dn_massage( &dc, &op->o_conn->c_dn,
-					&lsc->bound_dn) ) {
-			send_ldap_result( op, rs );
-			return rs->sr_err;
+					&lsc->msc_bound_ndn ) )
+		{
+			goto error_return;
 		}
 
-		assert( lsc->bound_dn.bv_val );
+		/* copy the DN idf needed */
+		if ( lsc->msc_bound_ndn.bv_val == op->o_conn->c_dn.bv_val ) {
+			ber_dupbv( &lsc->msc_bound_ndn, &op->o_conn->c_dn );
+		}
+
+		assert( !BER_BVISNULL( &lsc->msc_bound_ndn ) );
 
 	} else {
-		ber_str2bv( "", 0, 1, &lsc->bound_dn );
+		ber_str2bv( "", 0, 1, &lsc->msc_bound_ndn );
 	}
 
-	lsc->bound = META_UNBOUND;
+	lsc->msc_bound = META_UNBOUND;
 
-	/*
-	 * The candidate is activated
-	 */
-	lsc->candidate = META_CANDIDATE;
-	return LDAP_SUCCESS;
+error_return:;
+	if ( rs->sr_err != LDAP_SUCCESS ) {
+		rs->sr_err = slap_map_api2result( rs );
+		if ( sendok & LDAP_BACK_SENDERR ) {
+			send_ldap_result( op, rs );
+			rs->sr_text = NULL;
+		}
+
+	} else {
+
+		/*
+		 * The candidate is activated
+		 */
+		lsc->msc_candidate = META_CANDIDATE;
+	}
+
+	return rs->sr_err;
 }
 
 /*
@@ -344,19 +391,22 @@ init_one_conn(
  */
 struct metaconn *
 meta_back_getconn(
-	       	Operation 	*op,
-		SlapReply	*rs,
-		int 		op_type,
-		struct berval	*ndn,
-		int 		*candidate )
+	       	Operation 		*op,
+		SlapReply		*rs,
+		int 			op_type,
+		struct berval		*ndn,
+		int 			*candidate,
+		ldap_back_send_t	sendok )
 {
 	struct metainfo	*li = ( struct metainfo * )op->o_bd->be_private;
-	struct metaconn *lc, lc_curr;
-	int cached = -1, i = -1, err = LDAP_SUCCESS;
-	int new_conn = 0;
+	struct metaconn	*lc, lc_curr;
+	int		cached = META_TARGET_NONE,
+			i = META_TARGET_NONE,
+			err = LDAP_SUCCESS,
+			new_conn = 0;
 
 	/* Searches for a metaconn in the avl tree */
-	lc_curr.conn = op->o_conn;
+	lc_curr.mc_conn = op->o_conn;
 	ldap_pvt_thread_mutex_lock( &li->conn_mutex );
 	lc = (struct metaconn *)avl_find( li->conntree, 
 		(caddr_t)&lc_curr, meta_back_conn_cmp );
@@ -365,10 +415,37 @@ meta_back_getconn(
 	/* Looks like we didn't get a bind. Open a new session... */
 	if ( !lc ) {
 		lc = metaconn_alloc( li->ntargets );
-		lc->conn = op->o_conn;
+		lc->mc_conn = op->o_conn;
 		new_conn = 1;
 	}
 
+	/*
+	 * require all connections ...
+	 */
+	if ( op_type == META_OP_REQUIRE_ALL ) {
+		for ( i = 0; i < li->ntargets; i++ ) {
+
+			/*
+			 * The target is activated; if needed, it is
+			 * also init'd
+			 */
+			int lerr = init_one_conn( op, rs, li->targets[ i ],
+					&lc->mc_conns[ i ], sendok );
+			if ( lerr != LDAP_SUCCESS ) {
+				
+				/*
+				 * FIXME: in case one target cannot
+				 * be init'd, should the other ones
+				 * be tried?
+				 */
+				( void )meta_clear_one_candidate( &lc->mc_conns[ i ], 1 );
+				err = lerr;
+				continue;
+			}
+		}
+		goto done;
+	}
+	
 	/*
 	 * looks in cache, if any
 	 */
@@ -382,14 +459,14 @@ meta_back_getconn(
 		 * tries to get a unique candidate
 		 * (takes care of default target 
 		 */
-		if ( i < 0 ) {
+		if ( i == META_TARGET_NONE ) {
 			i = meta_back_select_unique_candidate( li, ndn );
 		}
 
 		/*
 		 * if any is found, inits the connection
 		 */
-		if ( i < 0 ) {
+		if ( i == META_TARGET_NONE ) {
 			if ( new_conn ) {
 				metaconn_free( lc );
 			}
@@ -398,15 +475,9 @@ meta_back_getconn(
 			return NULL;
 		}
 				
-#ifdef NEW_LOGGING
-		LDAP_LOG( BACK_META, INFO,
-			"meta_back_getconn: got target %d for ndn=\"%s\" from cache\n", 
-			i, ndn->bv_val, 0 );
-#else /* !NEW_LOGGING */
 		Debug( LDAP_DEBUG_CACHE,
-	"==>meta_back_getconn: got target %d for ndn=\"%s\" from cache\n%s",
-				i, ndn->bv_val, "" );
-#endif /* !NEW_LOGGING */
+	"==>meta_back_getconn: got target %d for ndn=\"%s\" from cache\n",
+				i, ndn->bv_val, 0 );
 
 		/*
 		 * Clear all other candidates
@@ -419,7 +490,7 @@ meta_back_getconn(
 		 * sends the appropriate result.
 		 */
 		err = init_one_conn( op, rs, li->targets[ i ],
-				&lc->conns[ i ] );
+				&lc->mc_conns[ i ], sendok );
 		if ( err != LDAP_SUCCESS ) {
 		
 			/*
@@ -427,7 +498,7 @@ meta_back_getconn(
 			 * be init'd, should the other ones
 			 * be tried?
 			 */
-			( void )meta_clear_one_candidate( &lc->conns[ i ], 1 );
+			( void )meta_clear_one_candidate( &lc->mc_conns[ i ], 1 );
 			if ( new_conn ) {
 				metaconn_free( lc );
 			}
@@ -438,38 +509,14 @@ meta_back_getconn(
 			*candidate = i;
 		}
 
-	/*
-	 * require all connections ...
-	 */
-	} else if (op_type == META_OP_REQUIRE_ALL) {
-		for ( i = 0; i < li->ntargets; i++ ) {
-
-			/*
-			 * The target is activated; if needed, it is
-			 * also init'd
-			 */
-			int lerr = init_one_conn( op, rs, li->targets[ i ],
-					&lc->conns[ i ] );
-			if ( lerr != LDAP_SUCCESS ) {
-				
-				/*
-				 * FIXME: in case one target cannot
-				 * be init'd, should the other ones
-				 * be tried?
-				 */
-				( void )meta_clear_one_candidate( &lc->conns[ i ], 1 );
-				err = lerr;
-				continue;
-			}
-		}
-
 	/*
 	 * if no unique candidate ...
 	 */
 	} else {
 		for ( i = 0; i < li->ntargets; i++ ) {
 			if ( i == cached 
-		|| meta_back_is_candidate( &li->targets[ i ]->suffix, ndn ) ) {
+				|| meta_back_is_candidate( &li->targets[ i ]->mt_nsuffix, ndn ) )
+			{
 
 				/*
 				 * The target is activated; if needed, it is
@@ -477,7 +524,7 @@ meta_back_getconn(
 				 */
 				int lerr = init_one_conn( op, rs,
 						li->targets[ i ],
-						&lc->conns[ i ] );
+						&lc->mc_conns[ i ], sendok );
 				if ( lerr != LDAP_SUCCESS ) {
 				
 					/*
@@ -485,7 +532,7 @@ meta_back_getconn(
 					 * be init'd, should the other ones
 					 * be tried?
 					 */
-					( void )meta_clear_one_candidate( &lc->conns[ i ], 1 );
+					( void )meta_clear_one_candidate( &lc->mc_conns[ i ], 1 );
 					err = lerr;
 					continue;
 				}
@@ -493,6 +540,7 @@ meta_back_getconn(
 		}
 	}
 
+done:;
 	/* clear out init_one_conn non-fatal errors */
 	rs->sr_err = LDAP_SUCCESS;
 	rs->sr_text = NULL;
@@ -512,14 +560,9 @@ meta_back_getconn(
 		
 		ldap_pvt_thread_mutex_unlock( &li->conn_mutex );
 
-#ifdef NEW_LOGGING
-		LDAP_LOG( BACK_META, INFO,
-			"meta_back_getconn: conn %ld inserted\n", lc->conn->c_connid, 0, 0);
-#else /* !NEW_LOGGING */
 		Debug( LDAP_DEBUG_TRACE,
-			"=>meta_back_getconn: conn %ld inserted\n%s%s",
-			lc->conn->c_connid, "", "" );
-#endif /* !NEW_LOGGING */
+			"=>meta_back_getconn: conn %ld inserted\n",
+			lc->mc_conn->c_connid, 0, 0 );
 		
 		/*
 		 * Err could be -1 in case a duplicate metaconn is inserted
@@ -530,15 +573,11 @@ meta_back_getconn(
 			metaconn_free( lc );
 			return NULL;
 		}
+
 	} else {
-#ifdef NEW_LOGGING
-		LDAP_LOG( BACK_META, INFO,
-			"meta_back_getconn: conn %ld fetched\n", lc->conn->c_connid, 0, 0 );
-#else /* !NEW_LOGGING */
 		Debug( LDAP_DEBUG_TRACE,
-			"=>meta_back_getconn: conn %ld fetched\n%s%s",
-			lc->conn->c_connid, "", "" );
-#endif /* !NEW_LOGGING */
+			"=>meta_back_getconn: conn %ld fetched\n",
+			lc->mc_conn->c_connid, 0, 0 );
 	}
 	
 	return lc;
