@@ -1,7 +1,9 @@
-/* $OpenLDAP: pkg/ldap/servers/slapd/back-meta/search.c,v 1.60.2.3 2003/12/01 00:58:41 kurt Exp $ */
+/* $OpenLDAP: pkg/ldap/servers/slapd/back-meta/search.c,v 1.84.2.4 2005/03/14 22:25:01 kurt Exp $ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 1999-2003 The OpenLDAP Foundation.
+ * Copyright 1999-2005 The OpenLDAP Foundation.
+ * Portions Copyright 2001-2003 Pierangelo Masarati.
+ * Portions Copyright 1999-2003 Howard Chu.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -17,69 +19,6 @@
  * in OpenLDAP Software and subsequently enhanced by Pierangelo
  * Masarati.
  */
-/* This is an altered version */
-/*
- * Copyright 2001, Pierangelo Masarati, All rights reserved. <ando@sys-net.it>
- *
- * This work has been developed to fulfill the requirements
- * of SysNet s.n.c. <http:www.sys-net.it> and it has been donated
- * to the OpenLDAP Foundation in the hope that it may be useful
- * to the Open Source community, but WITHOUT ANY WARRANTY.
- *
- * Permission is granted to anyone to use this software for any purpose
- * on any computer system, and to alter it and redistribute it, subject
- * to the following restrictions:
- *
- * 1. The author and SysNet s.n.c. are not responsible for the consequences
- *    of use of this software, no matter how awful, even if they arise from 
- *    flaws in it.
- *
- * 2. The origin of this software must not be misrepresented, either by
- *    explicit claim or by omission.  Since few users ever read sources,
- *    credits should appear in the documentation.
- *
- * 3. Altered versions must be plainly marked as such, and must not be
- *    misrepresented as being the original software.  Since few users
- *    ever read sources, credits should appear in the documentation.
- *    SysNet s.n.c. cannot be responsible for the consequences of the
- *    alterations.
- *
- * 4. This notice may not be removed or altered.
- *
- *
- * This software is based on the backend back-ldap, implemented
- * by Howard Chu <hyc@highlandsun.com>, and modified by Mark Valence
- * <kurash@sassafras.com>, Pierangelo Masarati <ando@sys-net.it> and other
- * contributors. The contribution of the original software to the present
- * implementation is acknowledged in this copyright statement.
- *
- * A special acknowledgement goes to Howard for the overall architecture
- * (and for borrowing large pieces of code), and to Mark, who implemented
- * from scratch the attribute/objectclass mapping.
- *
- * The original copyright statement follows.
- *
- * Copyright 1999, Howard Chu, All rights reserved. <hyc@highlandsun.com>
- *
- * Permission is granted to anyone to use this software for any purpose
- * on any computer system, and to alter it and redistribute it, subject
- * to the following restrictions:
- *
- * 1. The author is not responsible for the consequences of use of this
- *    software, no matter how awful, even if they arise from flaws in it.
- *
- * 2. The origin of this software must not be misrepresented, either by
- *    explicit claim or by omission.  Since few users ever read sources,
- *    credits should appear in the documentation.
- *
- * 3. Altered versions must be plainly marked as such, and must not be
- *    misrepresented as being the original software.  Since few users
- *    ever read sources, credits should appear in the
- *    documentation.
- *
- * 4. This notice may not be removed or altered.
- *                
- */
 
 #include "portable.h"
 
@@ -92,7 +31,6 @@
 #include "slap.h"
 #include "../back-ldap/back-ldap.h"
 #include "back-meta.h"
-#include "ldap_pvt.h"
 #undef ldap_debug	/* silence a warning in ldap-int.h */
 #include "ldap_log.h"
 #include "../../../libraries/libldap/ldap-int.h"
@@ -106,12 +44,6 @@ meta_send_entry(
 		LDAPMessage 	*e
 );
 
-static int
-is_one_level_rdn(
-		const char	*rdn,
-		int		from
-);
-
 int
 meta_back_search( Operation *op, SlapReply *rs )
 {
@@ -119,24 +51,19 @@ meta_back_search( Operation *op, SlapReply *rs )
 	struct metaconn *lc;
 	struct metasingleconn *lsc;
 	struct timeval	tv = { 0, 0 };
-	LDAPMessage	*res, *e;
-	int	rc = 0, *msgid, sres = LDAP_NO_SUCH_OBJECT;
+	LDAPMessage	*res = NULL, *e;
+	int	rc = 0, *msgid, sres = LDAP_SUCCESS;
 	char *err = NULL;
-	struct berval match = { 0, NULL }, mmatch = { 0, NULL };
+	struct berval match = BER_BVNULL, mmatch = BER_BVNULL;
 	BerVarray v2refs = NULL;
 		
 	int i, last = 0, candidates = 0, initial_candidates = 0,
 			candidate_match = 0;
-	struct slap_limits_set *limit = NULL;
-	int isroot = 0;
 	dncookie dc;
 
-	cache_manager*  cm = li->cm;
+	int	is_scope = 0,
+		is_filter = 0;
 
- 	if (cm->caching) {
- 		return meta_back_cache_search(op, rs);
-	}
-	
 	/*
 	 * controls are set in ldap_back_dobind()
 	 * 
@@ -144,16 +71,9 @@ meta_back_search( Operation *op, SlapReply *rs )
 	 * to map attrs and maybe rewrite value
 	 */
 	lc = meta_back_getconn( op, rs, META_OP_ALLOW_MULTIPLE, 
-			&op->o_req_ndn, NULL );
-	if ( !lc ) {
- 		send_ldap_result( op, rs );
-		return -1;
-	}
-
-	if ( !meta_back_dobind( lc, op ) ) {
-		rs->sr_err = LDAP_OTHER;
- 		send_ldap_result( op, rs );
-		return -1;
+			&op->o_req_ndn, NULL, LDAP_BACK_SENDERR );
+	if ( !lc || !meta_back_dobind( lc, op, LDAP_BACK_SENDERR ) ) {
+		return rs->sr_err;
 	}
 
 	/*
@@ -166,95 +86,47 @@ meta_back_search( Operation *op, SlapReply *rs )
 		return -1;
 	}
 	
-	/* if not root, get appropriate limits */
-	if ( be_isroot( op->o_bd, &op->o_ndn ) ) {
-		isroot = 1;
-	} else {
-		( void ) get_limits( op->o_bd, &op->o_ndn, &limit );
-	}
-
-	/* if no time limit requested, rely on remote server limits */
-	/* if requested limit higher than hard limit, abort */
-	if ( !isroot && op->oq_search.rs_tlimit > limit->lms_t_hard ) {
-		/* no hard limit means use soft instead */
-		if ( limit->lms_t_hard == 0
-				&& limit->lms_t_soft > -1
-				&& op->oq_search.rs_tlimit > limit->lms_t_soft ) {
-			op->oq_search.rs_tlimit = limit->lms_t_soft;
-			
-		/* positive hard limit means abort */
-		} else if ( limit->lms_t_hard > 0 ) {
-			rs->sr_err = LDAP_ADMINLIMIT_EXCEEDED;
-			send_ldap_result( op, rs );
-			rc = 0;
-			goto finish;
-		}
-		
-		/* negative hard limit means no limit */
-	}
-	
-	/* if no size limit requested, rely on remote server limits */
-	/* if requested limit higher than hard limit, abort */
-	if ( !isroot && op->oq_search.rs_slimit > limit->lms_s_hard ) {
-		/* no hard limit means use soft instead */
-		if ( limit->lms_s_hard == 0
-				&& limit->lms_s_soft > -1
-				&& op->oq_search.rs_slimit > limit->lms_s_soft ) {
-			op->oq_search.rs_slimit = limit->lms_s_soft;
-			
-		/* positive hard limit means abort */
-		} else if ( limit->lms_s_hard > 0 ) {
-			rs->sr_err = LDAP_ADMINLIMIT_EXCEEDED;
-			send_ldap_result( op, rs );
-			rc = 0;
-			goto finish;
-		}
-		
-		/* negative hard limit means no limit */
-	}
-
-
 	dc.conn = op->o_conn;
 	dc.rs = rs;
 
 	/*
 	 * Inits searches
 	 */
-	for ( i = 0, lsc = lc->conns; !META_LAST(lsc); ++i, ++lsc ) {
+	for ( i = 0, lsc = lc->mc_conns; !META_LAST( lsc ); ++i, ++lsc ) {
 		struct berval	realbase = op->o_req_dn;
-		int		realscope = op->oq_search.rs_scope;
+		int		realscope = op->ors_scope;
 		ber_len_t	suffixlen = 0;
-		struct berval	mbase = { 0, NULL }; 
-		struct berval	mfilter = { 0, NULL };
+		struct berval	mbase = BER_BVNULL; 
+		struct berval	mfilter = BER_BVNULL;
 		char		**mapped_attrs = NULL;
 
-		if ( lsc->candidate != META_CANDIDATE ) {
+		if ( lsc->msc_candidate != META_CANDIDATE ) {
 			msgid[ i ] = -1;
 			continue;
 		}
 
 		/* should we check return values? */
-		if ( op->oq_search.rs_deref != -1 ) {
-			ldap_set_option( lsc->ld, LDAP_OPT_DEREF,
-					( void * )&op->oq_search.rs_deref);
+		if ( op->ors_deref != -1 ) {
+			ldap_set_option( lsc->msc_ld, LDAP_OPT_DEREF,
+					( void * )&op->ors_deref);
 		}
-		if ( op->oq_search.rs_tlimit != -1 ) {
-			ldap_set_option( lsc->ld, LDAP_OPT_TIMELIMIT,
-					( void * )&op->oq_search.rs_tlimit);
+		if ( op->ors_tlimit != SLAP_NO_LIMIT ) {
+			ldap_set_option( lsc->msc_ld, LDAP_OPT_TIMELIMIT,
+					( void * )&op->ors_tlimit);
 		}
-		if ( op->oq_search.rs_slimit != -1 ) {
-			ldap_set_option( lsc->ld, LDAP_OPT_SIZELIMIT,
-					( void * )&op->oq_search.rs_slimit);
+		if ( op->ors_slimit != SLAP_NO_LIMIT ) {
+			ldap_set_option( lsc->msc_ld, LDAP_OPT_SIZELIMIT,
+					( void * )&op->ors_slimit);
 		}
 
-		dc.rwmap = &li->targets[ i ]->rwmap;
+		dc.rwmap = &li->targets[ i ]->mt_rwmap;
 
 		/*
 		 * modifies the base according to the scope, if required
 		 */
-		suffixlen = li->targets[ i ]->suffix.bv_len;
+		suffixlen = li->targets[ i ]->mt_nsuffix.bv_len;
 		if ( suffixlen > op->o_req_ndn.bv_len ) {
-			switch ( op->oq_search.rs_scope ) {
+			switch ( op->ors_scope ) {
 			case LDAP_SCOPE_SUBTREE:
 				/*
 				 * make the target suffix the new base
@@ -262,9 +134,11 @@ meta_back_search( Operation *op, SlapReply *rs )
 				 * illegal bases may be turned into 
 				 * the suffix of the target.
 				 */
-				if ( dnIsSuffix( &li->targets[ i ]->suffix,
+				if ( dnIsSuffix( &li->targets[ i ]->mt_nsuffix,
 						&op->o_req_ndn ) ) {
-					realbase = li->targets[ i ]->suffix;
+					realbase = li->targets[ i ]->mt_nsuffix;
+					is_scope++;
+
 				} else {
 					/*
 					 * this target is no longer candidate
@@ -275,18 +149,23 @@ meta_back_search( Operation *op, SlapReply *rs )
 				break;
 
 			case LDAP_SCOPE_ONELEVEL:
-				if ( is_one_level_rdn( li->targets[ i ]->suffix.bv_val,
-						suffixlen - op->o_req_ndn.bv_len - 1 ) 
-			&& dnIsSuffix( &li->targets[ i ]->suffix, &op->o_req_ndn ) ) {
+			{
+				struct berval	rdn = li->targets[ i ]->mt_nsuffix;
+				rdn.bv_len -= op->o_req_ndn.bv_len + STRLENOF( "," );
+				if ( dnIsOneLevelRDN( &rdn )
+						&& dnIsSuffix( &li->targets[ i ]->mt_nsuffix, &op->o_req_ndn ) )
+				{
 					/*
 					 * if there is exactly one level,
 					 * make the target suffix the new
 					 * base, and make scope "base"
 					 */
-					realbase = li->targets[ i ]->suffix;
+					realbase = li->targets[ i ]->mt_nsuffix;
 					realscope = LDAP_SCOPE_BASE;
+					is_scope++;
 					break;
 				} /* else continue with the next case */
+			}
 
 			case LDAP_SCOPE_BASE:
 				/*
@@ -296,6 +175,8 @@ meta_back_search( Operation *op, SlapReply *rs )
 				goto new_candidate;
 			}
 
+		} else {
+			is_scope++;
 		}
 
 		/*
@@ -333,9 +214,17 @@ meta_back_search( Operation *op, SlapReply *rs )
 		 * Maps filter
 		 */
 		rc = ldap_back_filter_map_rewrite( &dc,
-				op->oq_search.rs_filter,
+				op->ors_filter,
 				&mfilter, BACKLDAP_MAP );
-		if ( rc != 0 ) {
+		switch ( rc ) {
+		case LDAP_SUCCESS:
+			is_filter++;
+			break;
+
+		case LDAP_COMPARE_FALSE:
+			rc = 0;
+
+		default:
 			/*
 			 * this target is no longer candidate
 			 */
@@ -346,8 +235,8 @@ meta_back_search( Operation *op, SlapReply *rs )
 		/*
 		 * Maps required attributes
 		 */
-		rc = ldap_back_map_attrs( &li->targets[ i ]->rwmap.rwm_at,
-				op->oq_search.rs_attrs, BACKLDAP_MAP,
+		rc = ldap_back_map_attrs( &li->targets[ i ]->mt_rwmap.rwm_at,
+				op->ors_attrs, BACKLDAP_MAP,
 				&mapped_attrs );
 		if ( rc != LDAP_SUCCESS ) {
 			/*
@@ -360,26 +249,28 @@ meta_back_search( Operation *op, SlapReply *rs )
 		/*
 		 * Starts the search
 		 */
-		msgid[ i ] = ldap_search( lsc->ld, mbase.bv_val, realscope,
-				mfilter.bv_val, mapped_attrs,
-				op->oq_search.rs_attrsonly ); 
+		rc = ldap_search_ext( lsc->msc_ld,
+				mbase.bv_val, realscope, mfilter.bv_val,
+				mapped_attrs, op->ors_attrsonly,
+				op->o_ctrls, NULL,
+				NULL, op->ors_slimit, &msgid[ i ] ); 
 		if ( mapped_attrs ) {
 			free( mapped_attrs );
 			mapped_attrs = NULL;
 		}
-		if ( mfilter.bv_val != op->oq_search.rs_filterstr.bv_val ) {
+		if ( mfilter.bv_val != op->ors_filterstr.bv_val ) {
 			free( mfilter.bv_val );
-			mfilter.bv_val = NULL;
+			BER_BVZERO( &mfilter );
 		}
 		if ( mbase.bv_val != realbase.bv_val ) {
 			free( mbase.bv_val );
-			mbase.bv_val = NULL;
+			BER_BVZERO( &mbase );
 		}
 
-		if ( msgid[ i ] == -1 ) {
+		if ( rc != LDAP_SUCCESS ) {
 			continue;
 		}
-
+		
 		++candidates;
 
 new_candidate:;
@@ -405,19 +296,19 @@ new_candidate:;
 		/* check for abandon */
 		ab = op->o_abandon;
 
-		for ( i = 0, lsc = lc->conns; !META_LAST(lsc); lsc++, i++ ) {
+		for ( i = 0, lsc = lc->mc_conns; !META_LAST( lsc ); lsc++, i++ ) {
 			if ( msgid[ i ] == -1 ) {
 				continue;
 			}
 			
 			if ( ab ) {
-				ldap_abandon( lsc->ld, msgid[ i ] );
-				rc = 0;
+				ldap_abandon_ext( lsc->msc_ld, msgid[ i ], NULL, NULL );
+				rc = SLAPD_ABANDON;
 				break;
 			}
 
-			if ( op->oq_search.rs_slimit > 0
-					&& rs->sr_nentries == op->oq_search.rs_slimit ) {
+			if ( op->ors_slimit > 0
+					&& rs->sr_nentries == op->ors_slimit ) {
 				rs->sr_err = LDAP_SIZELIMIT_EXCEEDED;
 				rs->sr_v2ref = v2refs;
 				send_ldap_result( op, rs );
@@ -431,13 +322,19 @@ new_candidate:;
 			 * get a LDAP_TIMELIMIT_EXCEEDED from
 			 * one of them ...
 			 */
-			rc = ldap_result( lsc->ld, msgid[ i ],
+			rc = ldap_result( lsc->msc_ld, msgid[ i ],
 					0, &tv, &res );
 
 			if ( rc == 0 ) {
+				/* timeout exceeded */
+
+				/* FIXME: res should not need to be freed */
+				assert( res == NULL );
+
 				continue;
 
 			} else if ( rc == -1 ) {
+really_bad:;
 				/* something REALLY bad happened! */
 				( void )meta_clear_unused_candidates( li,
 						lc, -1, 0 );
@@ -446,12 +343,19 @@ new_candidate:;
 				send_ldap_result( op, rs );
 				
 				/* anything else needs be done? */
+
+				/* FIXME: res should not need to be freed */
+				assert( res == NULL );
+
 				goto finish;
 
 			} else if ( rc == LDAP_RES_SEARCH_ENTRY ) {
-				e = ldap_first_entry( lsc->ld, res );
+				e = ldap_first_entry( lsc->msc_ld, res );
 				meta_send_entry( op, rs, lc, i, e );
 
+				ldap_msgfree( res );
+				res = NULL;
+
 				/*
 				 * If scope is BASE, we need to jump out
 				 * as soon as one entry is found; if
@@ -459,26 +363,22 @@ new_candidate:;
 				 * this should correspond to the sole
 				 * entry that has the base DN
 				 */
-				if ( op->oq_search.rs_scope == LDAP_SCOPE_BASE
+				if ( op->ors_scope == LDAP_SCOPE_BASE
 						&& rs->sr_nentries > 0 ) {
 					candidates = 0;
 					sres = LDAP_SUCCESS;
 					break;
 				}
-				ldap_msgfree( res );
+
 				gotit = 1;
 
 			} else if ( rc == LDAP_RES_SEARCH_REFERENCE ) {
 				char		**references = NULL;
 				int		cnt;
 
-				/*
-				 * FIXME: should we collect references
-				 * and send them alltogether at the end?
-				 */
-
-				rc = ldap_parse_reference( lsc->ld, res,
+				rc = ldap_parse_reference( lsc->msc_ld, res,
 						&references, &rs->sr_ctrls, 1 );
+				res = NULL;
 
 				if ( rc != LDAP_SUCCESS ) {
 					continue;
@@ -488,24 +388,35 @@ new_candidate:;
 					continue;
 				}
 
+#ifdef ENABLE_REWRITE
+				dc.ctx = "referralDN";
+#else /* ! ENABLE_REWRITE */
+				dc.tofrom = 0;
+				dc.normalized = 0;
+#endif /* ! ENABLE_REWRITE */
 				for ( cnt = 0; references[ cnt ]; cnt++ )
-					/* NO OP */ ;
-				
-				rs->sr_ref = ch_calloc( cnt + 1, sizeof( struct berval ) );
+					;
+
+				rs->sr_ref = ch_calloc( sizeof( struct berval ), cnt + 1 );
 
 				for ( cnt = 0; references[ cnt ]; cnt++ ) {
-					rs->sr_ref[ cnt ].bv_val = references[ cnt ];
-					rs->sr_ref[ cnt ].bv_len = strlen( references[ cnt ] );
+					ber_str2bv( references[ cnt ], 0, 1, &rs->sr_ref[ cnt ] );
 				}
+				BER_BVZERO( &rs->sr_ref[ cnt ] );
 
-				/* ignore return value by now */
-				( void )send_search_reference( op, rs );
+				( void )ldap_back_referral_result_rewrite( &dc, rs->sr_ref );
+
+				if ( rs->sr_ref != NULL && !BER_BVISNULL( &rs->sr_ref[ 0 ] ) ) {
+					/* ignore return value by now */
+					( void )send_search_reference( op, rs );
+
+					ber_bvarray_free( rs->sr_ref );
+					rs->sr_ref = NULL;
+				}
 
 				/* cleanup */
 				if ( references ) {
 					ldap_value_free( references );
-					ch_free( rs->sr_ref );
-					rs->sr_ref = NULL;
 				}
 
 				if ( rs->sr_ctrls ) {
@@ -514,31 +425,31 @@ new_candidate:;
 				}
 
 			} else {
-				rs->sr_err = ldap_result2error( lsc->ld,
-						res, 1 );
-				sres = ldap_back_map_result( rs );
+				if ( ldap_parse_result( lsc->msc_ld, res,
+							&rs->sr_err,
+							NULL, NULL, NULL, NULL, 1 ) )
+				{
+					res = NULL;
+					goto really_bad;
+				}
+				res = NULL;
+
+				sres = slap_map_api2result( rs );
 				if ( err != NULL ) {
 					free( err );
 				}
-				ldap_get_option( lsc->ld,
+				ldap_get_option( lsc->msc_ld,
 						LDAP_OPT_ERROR_STRING, &err );
 				if ( match.bv_val != NULL ) {
 					free( match.bv_val );
 				}
-				ldap_get_option( lsc->ld,
+				ldap_get_option( lsc->msc_ld,
 						LDAP_OPT_MATCHED_DN, &match.bv_val );
 
-#ifdef NEW_LOGGING
-				LDAP_LOG( BACK_META, ERR,
-					"meta_back_search [%d] "
-					"match=\"%s\" err=\"%s\"\n",
-					i, match.bv_val, err );
-#else /* !NEW_LOGGING */
 				Debug( LDAP_DEBUG_ANY,
 					"=>meta_back_search [%d] "
 					"match=\"%s\" err=\"%s\"\n",
      					i, match.bv_val, err );	
-#endif /* !NEW_LOGGING */
 				candidate_match++;
 				last = i;
 				rc = 0;
@@ -558,8 +469,9 @@ new_candidate:;
 
 		if ( gotit == 0 ) {
 			tv.tv_sec = 0;
-                        tv.tv_usec = 100000;
+                        tv.tv_usec = 100000;	/* 0.1 s */
                         ldap_pvt_thread_yield();
+
 		} else {
 			tv.tv_sec = 0;
 			tv.tv_usec = 0;
@@ -581,8 +493,8 @@ new_candidate:;
 	 */
 	if ( candidate_match == initial_candidates
 			&& match.bv_val != NULL && *match.bv_val ) {
-		dc.ctx = "matchedDn";
-		dc.rwmap = &li->targets[ last ]->rwmap;
+		dc.ctx = "matchedDN";
+		dc.rwmap = &li->targets[ last ]->mt_rwmap;
 
 		if ( ldap_back_dn_massage( &dc, &match, &mmatch ) ) {
 			mmatch.bv_val = NULL;
@@ -596,6 +508,10 @@ new_candidate:;
 	 * FIXME: we should handle error codes and return the more 
 	 * important/reasonable
 	 */
+	if ( is_scope == 0 ) {
+		sres = LDAP_NO_SUCH_OBJECT;
+	}
+
 	if ( sres == LDAP_SUCCESS && v2refs ) {
 		sres = LDAP_REFERRAL;
 	}
@@ -637,10 +553,9 @@ meta_send_entry(
 {
 	struct metainfo 	*li = ( struct metainfo * )op->o_bd->be_private;
 	struct berval		a, mapped;
-	Entry 			ent;
+	Entry 			ent = {0};
 	BerElement 		ber = *e->lm_ber;
 	Attribute 		*attr, **attrp;
-	struct berval 		dummy = { 0, NULL };
 	struct berval 		*bv, bdn;
 	const char 		*text;
 	dncookie		dc;
@@ -652,7 +567,7 @@ meta_send_entry(
 	/*
 	 * Rewrite the dn of the result, if needed
 	 */
-	dc.rwmap = &li->targets[ target ]->rwmap;
+	dc.rwmap = &li->targets[ target ]->mt_rwmap;
 	dc.conn = op->o_conn;
 	dc.rs = rs;
 	dc.ctx = "searchResult";
@@ -670,7 +585,7 @@ meta_send_entry(
 	 * FIXME: should we log anything, or delegate to dnNormalize?
 	 */
 	if ( dnNormalize( 0, NULL, NULL, &ent.e_name, &ent.e_nname,
-		&op->o_tmpmemctx ) != LDAP_SUCCESS )
+		op->o_tmpmemctx ) != LDAP_SUCCESS )
 	{
 		return LDAP_INVALID_DN_SYNTAX;
 	}
@@ -683,16 +598,13 @@ meta_send_entry(
 				&ent.e_nname, target );
 	}
 
-	ent.e_id = 0;
-	ent.e_attrs = 0;
-	ent.e_private = 0;
 	attrp = &ent.e_attrs;
 
 	dc.ctx = "searchAttrDN";
 	while ( ber_scanf( &ber, "{m", &a ) != LBER_ERROR ) {
 		int		last = 0;
 
-		ldap_back_map( &li->targets[ target ]->rwmap.rwm_at, 
+		ldap_back_map( &li->targets[ target ]->mt_rwmap.rwm_at, 
 				&a, &mapped, BACKLDAP_REMAP );
 		if ( mapped.bv_val == NULL || mapped.bv_val[0] == '\0' ) {
 			continue;
@@ -708,14 +620,9 @@ meta_send_entry(
 				!= LDAP_SUCCESS) {
 			if ( slap_bv2undef_ad( &mapped, &attr->a_desc, &text ) 
 					!= LDAP_SUCCESS) {
-#ifdef NEW_LOGGING
-				LDAP_LOG( BACK_META, DETAIL1,
-					"slap_bv2undef_ad(%s): %s\n", mapped.bv_val, text, 0 );
-#else /* !NEW_LOGGING */
 				Debug( LDAP_DEBUG_ANY,
 						"slap_bv2undef_ad(%s): "
 						"%s\n%s", mapped.bv_val, text, "" );
-#endif /* !NEW_LOGGING */
 				ch_free( attr );
 				continue;
 			}
@@ -739,7 +646,7 @@ meta_send_entry(
 
 		if ( ber_scanf( &ber, "[W]", &attr->a_vals ) == LBER_ERROR 
 				|| attr->a_vals == NULL ) {
-			attr->a_vals = &dummy;
+			attr->a_vals = (struct berval *)&slap_dummy_bv;
 
 		} else if ( attr->a_desc == slap_schema.si_ad_objectClass
 				|| attr->a_desc == slap_schema.si_ad_structuralObjectClass ) {
@@ -747,7 +654,7 @@ meta_send_entry(
 			for ( last = 0; attr->a_vals[ last ].bv_val; ++last );
 
 			for ( bv = attr->a_vals; bv->bv_val; bv++ ) {
-				ldap_back_map( &li->targets[ target ]->rwmap.rwm_oc,
+				ldap_back_map( &li->targets[ target ]->mt_rwmap.rwm_oc,
 						bv, &mapped, BACKLDAP_REMAP );
 				if ( mapped.bv_val == NULL || mapped.bv_val[0] == '\0') {
 					free( bv->bv_val );
@@ -776,8 +683,12 @@ meta_send_entry(
 		 * everything pass thru the ldap backend.
 		 */
 		} else if ( attr->a_desc->ad_type->sat_syntax ==
-				slap_schema.si_syn_distinguishedName ) {
+				slap_schema.si_syn_distinguishedName )
+		{
 			ldap_dnattr_result_rewrite( &dc, attr->a_vals );
+
+		} else if ( attr->a_desc == slap_schema.si_ad_ref ) {
+			ldap_back_referral_result_rewrite( &dc, attr->a_vals );
 		}
 
 		if ( last && attr->a_desc->ad_type->sat_equality &&
@@ -791,7 +702,7 @@ meta_send_entry(
 					attr->a_desc->ad_type->sat_syntax,
 					attr->a_desc->ad_type->sat_equality,
 					&attr->a_vals[i], &attr->a_nvals[i],
-					op->o_tmpmemctx );
+					NULL );
 			}
 			attr->a_nvals[i].bv_val = NULL;
 			attr->a_nvals[i].bv_len = 0;
@@ -803,14 +714,15 @@ meta_send_entry(
 		attrp = &attr->a_next;
 	}
 	rs->sr_entry = &ent;
-	rs->sr_attrs = op->oq_search.rs_attrs;
+	rs->sr_attrs = op->ors_attrs;
+	rs->sr_flags = 0;
 	send_search_entry( op, rs );
 	rs->sr_entry = NULL;
 	rs->sr_attrs = NULL;
 	while ( ent.e_attrs ) {
 		attr = ent.e_attrs;
 		ent.e_attrs = attr->a_next;
-		if ( attr->a_vals != &dummy ) {
+		if ( attr->a_vals != &slap_dummy_bv ) {
 			ber_bvarray_free( attr->a_vals );
 		}
 		free( attr );
@@ -826,18 +738,4 @@ meta_send_entry(
 	return LDAP_SUCCESS;
 }
 
-static int
-is_one_level_rdn(
-		const char 	*rdn,
-		int 		from
-)
-{
-	for ( ; from--; ) {
-		if ( DN_SEPARATOR( rdn[ from ] ) ) {
-			return 0;
-		}
-	}
-
-	return 1;
-}
 
