@@ -1,8 +1,8 @@
 /* cancel.c - LDAP cancel extended operation */
-/* $OpenLDAP: pkg/ldap/servers/slapd/cancel.c,v 1.10.2.2 2003/12/01 00:58:33 kurt Exp $ */
+/* $OpenLDAP: pkg/ldap/servers/slapd/cancel.c,v 1.16.2.2 2005/03/14 22:24:58 kurt Exp $ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 1998-2003 The OpenLDAP Foundation.
+ * Copyright 1998-2005 The OpenLDAP Foundation.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -25,8 +25,6 @@
 
 #include "slap.h"
 
-#ifdef LDAP_EXOP_X_CANCEL
-
 #include <lber_pvt.h>
 #include <lutil.h>
 
@@ -34,7 +32,6 @@ int cancel_extop( Operation *op, SlapReply *rs )
 {
 	Operation *o;
 	int rc;
-	int found = 0;
 	int opid;
 	BerElement *ber;
 	int i;
@@ -68,64 +65,56 @@ int cancel_extop( Operation *op, SlapReply *rs )
 	LDAP_STAILQ_FOREACH( o, &op->o_conn->c_pending_ops, o_next ) {
 		if ( o->o_msgid == opid ) {
 			LDAP_STAILQ_REMOVE( &op->o_conn->c_pending_ops, o, slap_op, o_next );
+			LDAP_STAILQ_NEXT(o, o_next) = NULL;
+			op->o_conn->c_n_ops_pending--;
 			slap_op_free( o );
-			found = 1;
-			break;
+			ldap_pvt_thread_mutex_unlock( &op->o_conn->c_mutex );
+			return LDAP_SUCCESS;
 		}
 	}
-	ldap_pvt_thread_mutex_unlock( &op->o_conn->c_mutex );
 
-	if ( found ) return LDAP_SUCCESS;
-
-	found = 0;
-	ldap_pvt_thread_mutex_lock( &op->o_conn->c_mutex );
 	LDAP_STAILQ_FOREACH( o, &op->o_conn->c_ops, o_next ) {
 		if ( o->o_msgid == opid ) {
-			found = 1;
+			o->o_abandon = 1;
 			break;
 		}
 	}
 
-	if ( !found ) {
+	ldap_pvt_thread_mutex_unlock( &op->o_conn->c_mutex );
+
+	if ( o ) {
+		if ( o->o_cancel != SLAP_CANCEL_NONE ) {
+			rs->sr_text = "message ID already being cancelled";
+			return LDAP_PROTOCOL_ERROR;
+		}
+
+		o->o_cancel = SLAP_CANCEL_REQ;
+
 		for ( i = 0; i < nbackends; i++ ) {
 			op->o_bd = &backends[i];
 			if( !op->o_bd->be_cancel ) continue;
 
-			ldap_pvt_thread_mutex_unlock( &op->o_conn->c_mutex );
-
 			op->oq_cancel.rs_msgid = opid;
 			if ( op->o_bd->be_cancel( op, rs ) == LDAP_SUCCESS ) {
 				return LDAP_SUCCESS;
 			}
-			ldap_pvt_thread_mutex_lock( &op->o_conn->c_mutex );
 		}
-		ldap_pvt_thread_mutex_unlock( &op->o_conn->c_mutex );
-		rs->sr_text = "message ID not found";
-	 	return LDAP_NO_SUCH_OPERATION;
-	}
 
-	if ( op->o_cancel != SLAP_CANCEL_NONE ) {
-		ldap_pvt_thread_mutex_unlock( &op->o_conn->c_mutex );
-		rs->sr_text = "message ID already being cancelled";
-		return LDAP_PROTOCOL_ERROR;
-	}
+		while ( o->o_cancel == SLAP_CANCEL_REQ ) {
+			ldap_pvt_thread_yield();
+		}
 
-	op->o_cancel = SLAP_CANCEL_REQ;
-	ldap_pvt_thread_mutex_unlock( &op->o_conn->c_mutex );
+		if ( o->o_cancel == SLAP_CANCEL_ACK ) {
+			rc = LDAP_SUCCESS;
+		} else {
+			rc = o->o_cancel;
+		}
 
-	while ( op->o_cancel == SLAP_CANCEL_REQ ) {
-		ldap_pvt_thread_yield();
-	}
-
-	if ( op->o_cancel == SLAP_CANCEL_ACK ) {
-		rc = LDAP_SUCCESS;
+		o->o_cancel = SLAP_CANCEL_DONE;
 	} else {
-		rc = op->o_cancel;
+		rs->sr_text = "message ID not found";
+	 	rc = LDAP_NO_SUCH_OPERATION;
 	}
 
-	op->o_cancel = SLAP_CANCEL_DONE;
-
 	return rc;
 }
-
-#endif /* LDAP_EXOP_X_CANCEL */
