@@ -1,7 +1,7 @@
-/* $OpenLDAP: pkg/ldap/servers/slapd/controls.c,v 1.72.2.6 2003/12/01 00:58:33 kurt Exp $ */
+/* $OpenLDAP: pkg/ldap/servers/slapd/controls.c,v 1.125.2.5 2005/03/14 22:24:58 kurt Exp $ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 1998-2003 The OpenLDAP Foundation.
+ * Copyright 1998-2005 The OpenLDAP Foundation.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -35,21 +35,24 @@ static SLAP_CTRL_PARSE_FN parsePagedResults;
 static SLAP_CTRL_PARSE_FN parseValuesReturnFilter;
 static SLAP_CTRL_PARSE_FN parsePermissiveModify;
 static SLAP_CTRL_PARSE_FN parseDomainScope;
-
-#ifdef LDAP_CONTROL_SUBENTRIES
+static SLAP_CTRL_PARSE_FN parseTreeDelete;
+static SLAP_CTRL_PARSE_FN parseSearchOptions;
 static SLAP_CTRL_PARSE_FN parseSubentries;
-#endif
-static SLAP_CTRL_PARSE_FN parseLDAPsync;
 
 #undef sc_mask /* avoid conflict with Irix 6.5 <sys/signal.h> */
 
 const struct berval slap_pre_read_bv = BER_BVC(LDAP_CONTROL_PRE_READ);
 const struct berval slap_post_read_bv = BER_BVC(LDAP_CONTROL_POST_READ);
 
+struct slap_control_ids slap_cids;
+
 struct slap_control {
 	/* Control OID */
 	char *sc_oid;
 
+	/* The controlID for this control */
+	int sc_cid;
+
 	/* Operations supported by control */
 	slap_mask_t sc_mask;
 
@@ -68,7 +71,19 @@ static LDAP_SLIST_HEAD(ControlsList, slap_control) controls_list
 /*
  * all known request control OIDs should be added to this list
  */
-char **slap_known_controls = NULL;
+/*
+ * NOTE: initialize num_known_controls to 1 so that cid = 0 always
+ * addresses an undefined control; this allows to safely test for 
+ * well known controls even if they are not registered, e.g. if 
+ * they get moved to modules.  An example is sc_LDAPsync, which 
+ * is implemented in the syncprov overlay and thus, if configured 
+ * as dynamic module, may not be registered.  One side effect is that 
+ * slap_known_controls[0] == NULL, so it should always be used 
+ * starting from 1.
+ * FIXME: should we define the "undefined control" oid?
+ */
+char *slap_known_controls[SLAP_MAX_CIDS+1];
+static int num_known_controls = 1;
 
 static char *proxy_authz_extops[] = {
 	LDAP_EXOP_MODIFY_PASSWD,
@@ -77,54 +92,75 @@ static char *proxy_authz_extops[] = {
 };
 
 static struct slap_control control_defs[] = {
-	{ LDAP_CONTROL_ASSERT,
+	{  LDAP_CONTROL_ASSERT,
+ 		(int)offsetof(struct slap_control_ids, sc_assert),
 		SLAP_CTRL_HIDE|SLAP_CTRL_ACCESS, NULL,
 		parseAssert, LDAP_SLIST_ENTRY_INITIALIZER(next) },
 	{ LDAP_CONTROL_PRE_READ,
+ 		(int)offsetof(struct slap_control_ids, sc_preRead),
 		SLAP_CTRL_HIDE|SLAP_CTRL_DELETE|SLAP_CTRL_MODIFY|SLAP_CTRL_RENAME, NULL,
 		parsePreRead, LDAP_SLIST_ENTRY_INITIALIZER(next) },
 	{ LDAP_CONTROL_POST_READ,
+ 		(int)offsetof(struct slap_control_ids, sc_postRead),
 		SLAP_CTRL_HIDE|SLAP_CTRL_ADD|SLAP_CTRL_MODIFY|SLAP_CTRL_RENAME, NULL,
 		parsePostRead, LDAP_SLIST_ENTRY_INITIALIZER(next) },
  	{ LDAP_CONTROL_VALUESRETURNFILTER,
- 		SLAP_CTRL_SEARCH, NULL,
+ 		(int)offsetof(struct slap_control_ids, sc_valuesReturnFilter),
+ 		SLAP_CTRL_GLOBAL|SLAP_CTRL_SEARCH, NULL,
 		parseValuesReturnFilter, LDAP_SLIST_ENTRY_INITIALIZER(next) },
-#ifdef LDAP_CONTROL_PAGEDRESULTS
 	{ LDAP_CONTROL_PAGEDRESULTS,
+ 		(int)offsetof(struct slap_control_ids, sc_pagedResults),
 		SLAP_CTRL_SEARCH, NULL,
 		parsePagedResults, LDAP_SLIST_ENTRY_INITIALIZER(next) },
-#endif
 #ifdef LDAP_CONTROL_X_DOMAIN_SCOPE
 	{ LDAP_CONTROL_X_DOMAIN_SCOPE,
-		SLAP_CTRL_FRONTEND|SLAP_CTRL_SEARCH, NULL,
+ 		(int)offsetof(struct slap_control_ids, sc_domainScope),
+		SLAP_CTRL_GLOBAL|SLAP_CTRL_SEARCH, NULL,
 		parseDomainScope, LDAP_SLIST_ENTRY_INITIALIZER(next) },
 #endif
 #ifdef LDAP_CONTROL_X_PERMISSIVE_MODIFY
 	{ LDAP_CONTROL_X_PERMISSIVE_MODIFY,
+ 		(int)offsetof(struct slap_control_ids, sc_permissiveModify),
 		SLAP_CTRL_MODIFY, NULL,
 		parsePermissiveModify, LDAP_SLIST_ENTRY_INITIALIZER(next) },
 #endif
+#ifdef SLAP_CONTROL_X_TREE_DELETE
+	{ LDAP_CONTROL_X_TREE_DELETE,
+ 		(int)offsetof(struct slap_control_ids, sc_treeDelete),
+		SLAP_CTRL_HIDE|SLAP_CTRL_DELETE, NULL,
+		parseTreeDelete, LDAP_SLIST_ENTRY_INITIALIZER(next) },
+#endif
+#ifdef LDAP_CONTORL_X_SEARCH_OPTIONS
+	{ LDAP_CONTORL_X_SEARCH_OPTIONS,
+ 		(int)offsetof(struct slap_control_ids, sc_searchOptions),
+		SLAP_CTRL_GLOBAL|SLAP_CTRL_SEARCH, NULL,
+		parseSearchOptions, LDAP_SLIST_ENTRY_INITIALIZER(next) },
+#endif
 #ifdef LDAP_CONTROL_SUBENTRIES
 	{ LDAP_CONTROL_SUBENTRIES,
+ 		(int)offsetof(struct slap_control_ids, sc_subentries),
 		SLAP_CTRL_SEARCH, NULL,
 		parseSubentries, LDAP_SLIST_ENTRY_INITIALIZER(next) },
 #endif
 	{ LDAP_CONTROL_NOOP,
-		SLAP_CTRL_ACCESS, NULL,
+ 		(int)offsetof(struct slap_control_ids, sc_noOp),
+		SLAP_CTRL_HIDE|SLAP_CTRL_ACCESS, NULL,
 		parseNoOp, LDAP_SLIST_ENTRY_INITIALIZER(next) },
-	{ LDAP_CONTROL_SYNC,
-		SLAP_CTRL_HIDE|SLAP_CTRL_SEARCH, NULL,
-		parseLDAPsync, LDAP_SLIST_ENTRY_INITIALIZER(next) },
+#ifdef LDAP_CONTROL_MODIFY_INCREMENT
 	{ LDAP_CONTROL_MODIFY_INCREMENT,
+ 		(int)offsetof(struct slap_control_ids, sc_modifyIncrement),
 		SLAP_CTRL_HIDE|SLAP_CTRL_MODIFY, NULL,
 		parseModifyIncrement, LDAP_SLIST_ENTRY_INITIALIZER(next) },
+#endif
 	{ LDAP_CONTROL_MANAGEDSAIT,
+ 		(int)offsetof(struct slap_control_ids, sc_manageDSAit),
 		SLAP_CTRL_ACCESS, NULL,
 		parseManageDSAit, LDAP_SLIST_ENTRY_INITIALIZER(next) },
 	{ LDAP_CONTROL_PROXY_AUTHZ,
-		SLAP_CTRL_FRONTEND|SLAP_CTRL_ACCESS, proxy_authz_extops,
+ 		(int)offsetof(struct slap_control_ids, sc_proxyAuthz),
+		SLAP_CTRL_GLOBAL|SLAP_CTRL_ACCESS, proxy_authz_extops,
 		parseProxyAuthz, LDAP_SLIST_ENTRY_INITIALIZER(next) },
-	{ NULL, 0, NULL, 0, LDAP_SLIST_ENTRY_INITIALIZER(next) }
+	{ NULL, 0, 0, NULL, 0, LDAP_SLIST_ENTRY_INITIALIZER(next) }
 };
 
 /*
@@ -137,19 +173,34 @@ int
 register_supported_control(const char *controloid,
 	slap_mask_t controlmask,
 	char **controlexops,
-	SLAP_CTRL_PARSE_FN *controlparsefn)
+	SLAP_CTRL_PARSE_FN *controlparsefn,
+	int *controlcid)
 {
 	struct slap_control *sc;
 	int i;
 
-	if ( controloid == NULL ) {
-		return LDAP_PARAM_ERROR;
+	if ( num_known_controls >= SLAP_MAX_CIDS ) {
+		Debug( LDAP_DEBUG_ANY, "Too many controls registered."
+			" Recompile slapd with SLAP_MAX_CIDS defined > %d\n",
+		SLAP_MAX_CIDS, 0, 0 );
+		return LDAP_OTHER;
+	}
+
+	if ( controloid == NULL ) return LDAP_PARAM_ERROR;
+
+	/* sanity check - should never happen */
+	for ( i = 0; slap_known_controls[ i ]; i++ ) {
+		if ( strcmp( controloid, slap_known_controls[ i ] ) == 0 ) {
+			Debug( LDAP_DEBUG_ANY,
+				"Control %s already registered.\n",
+				controloid, 0, 0 );
+			return LDAP_PARAM_ERROR;
+		}
 	}
 
 	sc = (struct slap_control *)SLAP_MALLOC( sizeof( *sc ) );
-	if ( sc == NULL ) {
-		return LDAP_NO_MEMORY;
-	}
+	if ( sc == NULL ) return LDAP_NO_MEMORY;
+
 	sc->sc_oid = ch_strdup( controloid );
 	sc->sc_mask = controlmask;
 	if ( controlexops != NULL ) {
@@ -163,32 +214,15 @@ register_supported_control(const char *controloid,
 	}
 	sc->sc_parse = controlparsefn;
 
+	if ( controlcid ) *controlcid = num_known_controls;
+	sc->sc_cid = num_known_controls;
+
 	/* Update slap_known_controls, too. */
-	if ( slap_known_controls == NULL ) {
-		slap_known_controls = (char **)SLAP_MALLOC( 2 * sizeof(char *) );
-		if ( slap_known_controls == NULL ) {
-			if ( sc->sc_extendedops != NULL ) ldap_charray_free( sc->sc_extendedops );
-			ch_free( sc );
-			return LDAP_NO_MEMORY;
-		}
-		slap_known_controls[0] = ch_strdup( sc->sc_oid );
-		slap_known_controls[1] = NULL;
-	} else {
-		for ( i = 0; slap_known_controls[i] != NULL; i++ )
-			;
-		slap_known_controls = (char **)SLAP_REALLOC( slap_known_controls, (i + 2) * sizeof(char *) );
-		if ( slap_known_controls == NULL ) {
-			if ( sc->sc_extendedops != NULL ) ldap_charray_free( sc->sc_extendedops );
-			ch_free( sc );
-			return LDAP_NO_MEMORY;
-		}
-		slap_known_controls[i++] = ch_strdup( sc->sc_oid );
-		slap_known_controls[i] = NULL;
-	}
+	slap_known_controls[num_known_controls-1] = sc->sc_oid;
+	slap_known_controls[num_known_controls++] = NULL;
 
 	LDAP_SLIST_NEXT( sc, sc_next ) = NULL;
 	LDAP_SLIST_INSERT_HEAD( &controls_list, sc, sc_next );
-
 	return LDAP_SUCCESS;
 }
 
@@ -204,11 +238,11 @@ slap_controls_init( void )
 	rc = LDAP_SUCCESS;
 
 	for ( i = 0; control_defs[i].sc_oid != NULL; i++ ) {
+		int *cid = (int *)(((char *)&slap_cids) + control_defs[i].sc_cid );
 		rc = register_supported_control( control_defs[i].sc_oid,
 			control_defs[i].sc_mask, control_defs[i].sc_extendedops,
-			control_defs[i].sc_parse );
-		if ( rc != LDAP_SUCCESS )
-			break;
+			control_defs[i].sc_parse, cid );
+		if ( rc != LDAP_SUCCESS ) break;
 	}
 
 	return rc;
@@ -232,7 +266,6 @@ controls_destroy( void )
 		}
 		ch_free( sc );
 	}
-	ldap_charray_free( slap_known_controls );
 }
 
 /*
@@ -334,10 +367,51 @@ find_ctrl( const char *oid )
 	return NULL;
 }
 
+int
+slap_find_control_id(
+	const char *oid,
+	int *cid )
+{
+	struct slap_control *ctrl = find_ctrl( oid );
+	if ( ctrl ) {
+		if ( cid ) *cid = ctrl->sc_cid;
+		return LDAP_SUCCESS;
+	}
+	return LDAP_CONTROL_NOT_FOUND;
+}
+
+int
+slap_global_control( Operation *op, const char *oid, int *cid )
+{
+	struct slap_control *ctrl = find_ctrl( oid );
+
+	if ( ctrl == NULL ) {
+		/* should not be reachable */
+		Debug( LDAP_DEBUG_ANY,
+			"slap_global_control: unrecognized control: %s\n",      
+			oid, 0, 0 );
+		return LDAP_CONTROL_NOT_FOUND;
+	}
+
+	if ( cid ) *cid = ctrl->sc_cid;
+
+	if ( ( ctrl->sc_mask & SLAP_CTRL_GLOBAL ) ||
+			( ( op->o_tag & LDAP_REQ_SEARCH ) &&
+			( ctrl->sc_mask & SLAP_CTRL_GLOBAL_SEARCH ) ) )
+	{
+		return LDAP_COMPARE_TRUE;
+	}
+
+	Debug( LDAP_DEBUG_TRACE,
+		"slap_global_control: unavailable control: %s\n",      
+		oid, 0, 0 );
+
+	return LDAP_COMPARE_FALSE;
+}
+
 void slap_free_ctrls(
 	Operation *op,
-	LDAPControl **ctrls
-)
+	LDAPControl **ctrls )
 {
 	int i;
 
@@ -377,13 +451,8 @@ int get_ctrls(
 		goto return_results;
 	}
 
-#ifdef NEW_LOGGING
-	LDAP_LOG( OPERATION, ENTRY,
-		"get_ctrls: conn %lu\n", op->o_connid, 0, 0 );
-#else
 	Debug( LDAP_DEBUG_TRACE,
 		"=> get_ctrls\n", 0, 0, 0 );
-#endif
 
 	if( op->o_protocol < LDAP_VERSION3 ) {
 		rs->sr_err = SLAPD_DISCONNECT;
@@ -441,13 +510,8 @@ int get_ctrls(
 		c->ldctl_oid = bv.bv_val;
 
 		if( tag == LBER_ERROR ) {
-#ifdef NEW_LOGGING
-			LDAP_LOG( OPERATION, INFO, "get_ctrls: conn %lu get OID failed.\n",
-				op->o_connid, 0, 0 );
-#else
 			Debug( LDAP_DEBUG_TRACE, "=> get_ctrls: get oid failed.\n",
 				0, 0, 0 );
-#endif
 
 			slap_free_ctrls( op, op->o_ctrls );
 			op->o_ctrls = NULL;
@@ -456,15 +520,9 @@ int get_ctrls(
 			goto return_results;
 
 		} else if( c->ldctl_oid == NULL ) {
-#ifdef NEW_LOGGING
-			LDAP_LOG( OPERATION, INFO,
-				"get_ctrls: conn %lu got emtpy OID.\n",
-				op->o_connid, 0, 0 );
-#else
 			Debug( LDAP_DEBUG_TRACE,
 				"get_ctrls: conn %lu got emtpy OID.\n",
 				op->o_connid, 0, 0 );
-#endif
 
 			slap_free_ctrls( op, op->o_ctrls );
 			op->o_ctrls = NULL;
@@ -480,14 +538,8 @@ int get_ctrls(
 			tag = ber_scanf( ber, "b", &crit );
 
 			if( tag == LBER_ERROR ) {
-#ifdef NEW_LOGGING
-				LDAP_LOG( OPERATION, INFO, 
-					"get_ctrls: conn %lu get crit failed.\n", 
-					op->o_connid, 0, 0 );
-#else
 				Debug( LDAP_DEBUG_TRACE, "=> get_ctrls: get crit failed.\n",
 					0, 0, 0 );
-#endif
 				slap_free_ctrls( op, op->o_ctrls );
 				op->o_ctrls = NULL;
 				rs->sr_err = SLAPD_DISCONNECT;
@@ -503,17 +555,10 @@ int get_ctrls(
 			tag = ber_scanf( ber, "m", &c->ldctl_value );
 
 			if( tag == LBER_ERROR ) {
-#ifdef NEW_LOGGING
-				LDAP_LOG( OPERATION, INFO, "get_ctrls: conn %lu: "
-					"%s (%scritical): get value failed.\n",
-					op->o_connid, c->ldctl_oid,
-					c->ldctl_iscritical ? "" : "non" );
-#else
 				Debug( LDAP_DEBUG_TRACE, "=> get_ctrls: conn %lu: "
 					"%s (%scritical): get value failed.\n",
 					op->o_connid, c->ldctl_oid,
 					c->ldctl_iscritical ? "" : "non" );
-#endif
 				slap_free_ctrls( op, op->o_ctrls );
 				op->o_ctrls = NULL;
 				rs->sr_err = SLAPD_DISCONNECT;
@@ -522,15 +567,9 @@ int get_ctrls(
 			}
 		}
 
-#ifdef NEW_LOGGING
-		LDAP_LOG( OPERATION, INFO, 
-			"get_ctrls: conn %lu oid=\"%s\" (%scritical)\n",
-			op->o_connid, c->ldctl_oid, c->ldctl_iscritical ? "" : "non" );
-#else
 		Debug( LDAP_DEBUG_TRACE,
 			"=> get_ctrls: oid=\"%s\" (%scritical)\n",
 			c->ldctl_oid, c->ldctl_iscritical ? "" : "non", 0 );
-#endif
 
 		sc = find_ctrl( c->ldctl_oid );
 		if( sc != NULL ) {
@@ -570,8 +609,8 @@ int get_ctrls(
 				if( sc->sc_extendedops != NULL ) {
 					int i;
 					for( i=0; sc->sc_extendedops[i] != NULL; i++ ) {
-						if( strcmp( op->ore_reqoid.bv_val, sc->sc_extendedops[i] )
-							== 0 )
+						if( strcmp( op->ore_reqoid.bv_val,
+							sc->sc_extendedops[i] ) == 0 )
 						{
 							tagmask=0L;
 							break;
@@ -587,6 +626,7 @@ int get_ctrls(
 
 			if (( sc->sc_mask & tagmask ) == tagmask ) {
 				/* available extension */
+				int	rc;
 
 				if( !sc->sc_parse ) {
 					rs->sr_err = LDAP_OTHER;
@@ -594,19 +634,11 @@ int get_ctrls(
 					goto return_results;
 				}
 
-				rs->sr_err = sc->sc_parse( op, rs, c );
-
-				if( rs->sr_err != LDAP_SUCCESS ) goto return_results;
-
-				if ( sc->sc_mask & SLAP_CTRL_FRONTEND ) {
-					/* kludge to disable backend_control() check */
-					c->ldctl_iscritical = 0;
-
-				} else if ( tagmask == SLAP_CTRL_SEARCH &&
-					sc->sc_mask & SLAP_CTRL_FRONTEND_SEARCH )
-				{
-					/* kludge to disable backend_control() check */
-					c->ldctl_iscritical = 0;
+				rc = sc->sc_parse( op, rs, c );
+				if ( rc ) {
+					assert( rc != LDAP_UNAVAILABLE_CRITICAL_EXTENSION );
+					rs->sr_err = rc;
+					goto return_results;
 				}
 
 			} else if( c->ldctl_iscritical ) {
@@ -622,18 +654,13 @@ int get_ctrls(
 			rs->sr_text = "critical extension is not recognized";
 			goto return_results;
 		}
+next_ctrl:;
 	}
 
 return_results:
-#ifdef NEW_LOGGING
-	LDAP_LOG( OPERATION, RESULTS, 
-		"get_ctrls: n=%d rc=%d err=\"%s\"\n",
-		nctrls, rs->sr_err, rs->sr_text ? rs->sr_text : "" );
-#else
 	Debug( LDAP_DEBUG_TRACE,
 		"<= get_ctrls: n=%d rc=%d err=\"%s\"\n",
 		nctrls, rs->sr_err, rs->sr_text ? rs->sr_text : "");
-#endif
 
 	if( sendres && rs->sr_err != LDAP_SUCCESS ) {
 		if( rs->sr_err == SLAPD_DISCONNECT ) {
@@ -654,7 +681,7 @@ static int parseModifyIncrement (
 	LDAPControl *ctrl )
 {
 #if 0
-	if ( op->o_parseModifyIncrement != SLAP_NO_CONTROL ) {
+	if ( op->o_modifyIncrement != SLAP_CONTROL_NONE ) {
 		rs->sr_text = "modifyIncrement control specified multiple times";
 		return LDAP_PROTOCOL_ERROR;
 	}
@@ -666,9 +693,9 @@ static int parseModifyIncrement (
 	}
 
 #if 0
-	op->o_parseModifyIncrement = ctrl->ldctl_iscritical
-		? SLAP_CRITICAL_CONTROL
-		: SLAP_NONCRITICAL_CONTROL;
+	op->o_modifyIncrement = ctrl->ldctl_iscritical
+		? SLAP_CONTROL_CRITICAL
+		: SLAP_CONTROL_NONCRITICAL;
 #endif
 
 	return LDAP_SUCCESS;
@@ -679,7 +706,7 @@ static int parseManageDSAit (
 	SlapReply *rs,
 	LDAPControl *ctrl )
 {
-	if ( op->o_managedsait != SLAP_NO_CONTROL ) {
+	if ( op->o_managedsait != SLAP_CONTROL_NONE ) {
 		rs->sr_text = "manageDSAit control specified multiple times";
 		return LDAP_PROTOCOL_ERROR;
 	}
@@ -690,8 +717,8 @@ static int parseManageDSAit (
 	}
 
 	op->o_managedsait = ctrl->ldctl_iscritical
-		? SLAP_CRITICAL_CONTROL
-		: SLAP_NONCRITICAL_CONTROL;
+		? SLAP_CONTROL_CRITICAL
+		: SLAP_CONTROL_NONCRITICAL;
 
 	return LDAP_SUCCESS;
 }
@@ -701,42 +728,28 @@ static int parseProxyAuthz (
 	SlapReply *rs,
 	LDAPControl *ctrl )
 {
-	int rc;
-	struct berval dn = { 0, NULL };
+	int		rc;
+	struct berval	dn = BER_BVNULL;
 
-	if ( op->o_proxy_authz != SLAP_NO_CONTROL ) {
+	if ( op->o_proxy_authz != SLAP_CONTROL_NONE ) {
 		rs->sr_text = "proxy authorization control specified multiple times";
 		return LDAP_PROTOCOL_ERROR;
 	}
 
 	op->o_proxy_authz = ctrl->ldctl_iscritical
-		? SLAP_CRITICAL_CONTROL
-		: SLAP_NONCRITICAL_CONTROL;
+		? SLAP_CONTROL_CRITICAL
+		: SLAP_CONTROL_NONCRITICAL;
 
-#ifdef NEW_LOGGING
-	LDAP_LOG( OPERATION, ARGS, 
-		"parseProxyAuthz: conn %lu authzid=\"%s\"\n", 
-		op->o_connid,
-		ctrl->ldctl_value.bv_len ?  ctrl->ldctl_value.bv_val : "anonymous",
-		0 );
-#else
 	Debug( LDAP_DEBUG_ARGS,
 		"parseProxyAuthz: conn %lu authzid=\"%s\"\n", 
 		op->o_connid,
 		ctrl->ldctl_value.bv_len ?  ctrl->ldctl_value.bv_val : "anonymous",
 		0 );
-#endif
 
 	if( ctrl->ldctl_value.bv_len == 0 ) {
-#ifdef NEW_LOGGING
-		LDAP_LOG( OPERATION, RESULTS, 
-			"parseProxyAuthz: conn=%lu anonymous\n", 
-			op->o_connid, 0, 0 );
-#else
 		Debug( LDAP_DEBUG_TRACE,
 			"parseProxyAuthz: conn=%lu anonymous\n", 
 			op->o_connid, 0, 0 );
-#endif
 
 		/* anonymous */
 		free( op->o_dn.bv_val );
@@ -750,11 +763,11 @@ static int parseProxyAuthz (
 		return LDAP_SUCCESS;
 	}
 
-	rc = slap_sasl_getdn( op->o_conn, op,
-		ctrl->ldctl_value.bv_val, ctrl->ldctl_value.bv_len,
-		NULL, &dn, SLAP_GETDN_AUTHZID );
+	rc = slap_sasl_getdn( op->o_conn, op, &ctrl->ldctl_value,
+			NULL, &dn, SLAP_GETDN_AUTHZID );
 
-	if( rc != LDAP_SUCCESS || !dn.bv_len ) {
+	/* FIXME: empty DN in proxyAuthz control should be legal... */
+	if( rc != LDAP_SUCCESS /* || !dn.bv_len */ ) {
 		if ( dn.bv_val ) {
 			ch_free( dn.bv_val );
 		}
@@ -762,17 +775,10 @@ static int parseProxyAuthz (
 		return LDAP_PROXY_AUTHZ_FAILURE;
 	}
 
-#ifdef NEW_LOGGING
-	LDAP_LOG( OPERATION, RESULTS, 
-		"parseProxyAuthz: conn=%lu \"%s\"\n", 
-		op->o_connid,
-		dn.bv_len ? dn.bv_val : "(NULL)", 0 );
-#else
 	Debug( LDAP_DEBUG_TRACE,
 		"parseProxyAuthz: conn=%lu \"%s\"\n", 
 		op->o_connid,
 		dn.bv_len ? dn.bv_val : "(NULL)", 0 );
-#endif
 
 	rc = slap_sasl_authorized( op, &op->o_ndn, &dn );
 
@@ -788,6 +794,9 @@ static int parseProxyAuthz (
 	op->o_dn.bv_val = NULL;
 	op->o_ndn = dn;
 
+	Statslog( LDAP_DEBUG_STATS, "%s PROXYAUTHZ dn=\"%s\"\n",
+	    op->o_log_prefix, dn.bv_val, 0, 0, 0 );
+
 	/*
 	 * NOTE: since slap_sasl_getdn() returns a normalized dn,
 	 * from now on op->o_dn is normalized
@@ -802,7 +811,7 @@ static int parseNoOp (
 	SlapReply *rs,
 	LDAPControl *ctrl )
 {
-	if ( op->o_noop != SLAP_NO_CONTROL ) {
+	if ( op->o_noop != SLAP_CONTROL_NONE ) {
 		rs->sr_text = "noop control specified multiple times";
 		return LDAP_PROTOCOL_ERROR;
 	}
@@ -813,29 +822,30 @@ static int parseNoOp (
 	}
 
 	op->o_noop = ctrl->ldctl_iscritical
-		? SLAP_CRITICAL_CONTROL
-		: SLAP_NONCRITICAL_CONTROL;
+		? SLAP_CONTROL_CRITICAL
+		: SLAP_CONTROL_NONCRITICAL;
 
 	return LDAP_SUCCESS;
 }
 
-#ifdef LDAP_CONTROL_PAGEDRESULTS
 static int parsePagedResults (
 	Operation *op,
 	SlapReply *rs,
 	LDAPControl *ctrl )
 {
-	ber_tag_t tag;
-	ber_int_t size;
-	BerElement *ber;
-	struct berval cookie = { 0, NULL };
+	int		rc = LDAP_SUCCESS;
+	ber_tag_t	tag;
+	ber_int_t	size;
+	BerElement	*ber;
+	struct berval	cookie = BER_BVNULL;
+	PagedResultsState	*ps;
 
-	if ( op->o_pagedresults != SLAP_NO_CONTROL ) {
+	if ( op->o_pagedresults != SLAP_CONTROL_NONE ) {
 		rs->sr_text = "paged results control specified multiple times";
 		return LDAP_PROTOCOL_ERROR;
 	}
 
-	if ( ctrl->ldctl_value.bv_len == 0 ) {
+	if ( BER_BVISEMPTY( &ctrl->ldctl_value ) ) {
 		rs->sr_text = "paged results control value is empty (or absent)";
 		return LDAP_PROTOCOL_ERROR;
 	}
@@ -849,58 +859,96 @@ static int parsePagedResults (
 	 * }
 	 */
 	ber = ber_init( &ctrl->ldctl_value );
-	if( ber == NULL ) {
+	if ( ber == NULL ) {
 		rs->sr_text = "internal error";
 		return LDAP_OTHER;
 	}
 
 	tag = ber_scanf( ber, "{im}", &size, &cookie );
-	(void) ber_free( ber, 1 );
 
-	if( tag == LBER_ERROR ) {
+	if ( tag == LBER_ERROR ) {
 		rs->sr_text = "paged results control could not be decoded";
-		return LDAP_PROTOCOL_ERROR;
+		rc = LDAP_PROTOCOL_ERROR;
+		goto done;
 	}
 
-	if( size < 0 ) {
+	if ( size < 0 ) {
 		rs->sr_text = "paged results control size invalid";
-		return LDAP_PROTOCOL_ERROR;
+		rc = LDAP_PROTOCOL_ERROR;
+		goto done;
 	}
 
-	if( cookie.bv_len ) {
+#if 0
+	/* defer cookie decoding/checks to backend... */
+	if ( cookie.bv_len ) {
 		PagedResultsCookie reqcookie;
 		if( cookie.bv_len != sizeof( reqcookie ) ) {
 			/* bad cookie */
 			rs->sr_text = "paged results cookie is invalid";
-			return LDAP_PROTOCOL_ERROR;
+			rc = LDAP_PROTOCOL_ERROR;
+			goto done;
 		}
 
 		AC_MEMCPY( &reqcookie, cookie.bv_val, sizeof( reqcookie ));
 
-		if( reqcookie > op->o_pagedresults_state.ps_cookie ) {
+		if ( reqcookie > op->o_pagedresults_state.ps_cookie ) {
 			/* bad cookie */
 			rs->sr_text = "paged results cookie is invalid";
-			return LDAP_PROTOCOL_ERROR;
+			rc = LDAP_PROTOCOL_ERROR;
+			goto done;
 
-		} else if( reqcookie < op->o_pagedresults_state.ps_cookie ) {
+		} else if ( reqcookie < op->o_pagedresults_state.ps_cookie ) {
 			rs->sr_text = "paged results cookie is invalid or old";
-			return LDAP_UNWILLING_TO_PERFORM;
+			rc = LDAP_UNWILLING_TO_PERFORM;
+			goto done;
 		}
+
 	} else {
 		/* Initial request.  Initialize state. */
+#if 0
+		if ( op->o_conn->c_pagedresults_state.ps_cookie != 0 ) {
+			/* There's another pagedResults control on the
+			 * same connection; reject new pagedResults controls 
+			 * (allowed by RFC2696) */
+			rs->sr_text = "paged results cookie unavailable; try later";
+			rc = LDAP_UNWILLING_TO_PERFORM;
+			goto done;
+		}
+#endif
 		op->o_pagedresults_state.ps_cookie = 0;
-		op->o_pagedresults_state.ps_id = NOID;
+		op->o_pagedresults_state.ps_count = 0;
+	}
+#endif
+
+	ps = op->o_tmpalloc( sizeof(PagedResultsState), op->o_tmpmemctx );
+	*ps = op->o_conn->c_pagedresults_state;
+	ps->ps_size = size;
+	op->o_pagedresults_state = ps;
+
+	/* NOTE: according to RFC 2696 3.:
+
+    If the page size is greater than or equal to the sizeLimit value, the
+    server should ignore the control as the request can be satisfied in a
+    single page.
+	 
+	 * NOTE: this assumes that the op->ors_slimit be set
+	 * before the controls are parsed.     
+	 */
+		
+	if ( op->ors_slimit > 0 && size >= op->ors_slimit ) {
+		op->o_pagedresults = SLAP_CONTROL_IGNORED;
+
+	} else if ( ctrl->ldctl_iscritical ) {
+		op->o_pagedresults = SLAP_CONTROL_CRITICAL;
+
+	} else {
+		op->o_pagedresults = SLAP_CONTROL_NONCRITICAL;
 	}
 
-	op->o_pagedresults_size = size;
-
-	op->o_pagedresults = ctrl->ldctl_iscritical
-		? SLAP_CRITICAL_CONTROL
-		: SLAP_NONCRITICAL_CONTROL;
-
-	return LDAP_SUCCESS;
+done:;
+	(void)ber_free( ber, 1 );
+	return rc;
 }
-#endif
 
 static int parseAssert (
 	Operation *op,
@@ -908,10 +956,10 @@ static int parseAssert (
 	LDAPControl *ctrl )
 {
 	BerElement	*ber;
-	struct berval	fstr = { 0, NULL };
+	struct berval	fstr = BER_BVNULL;
 	const char *err_msg = "";
 
-	if ( op->o_assert != SLAP_NO_CONTROL ) {
+	if ( op->o_assert != SLAP_CONTROL_NONE ) {
 		rs->sr_text = "assert control specified multiple times";
 		return LDAP_PROTOCOL_ERROR;
 	}
@@ -927,7 +975,7 @@ static int parseAssert (
 		return LDAP_OTHER;
 	}
 	
-	rs->sr_err = get_filter( op, ber, &(op->o_assertion), &rs->sr_text);
+	rs->sr_err = get_filter( op, ber, (Filter **)&(op->o_assertion), &rs->sr_text);
 
 	if( rs->sr_err != LDAP_SUCCESS ) {
 		if( rs->sr_err == SLAPD_DISCONNECT ) {
@@ -946,20 +994,14 @@ static int parseAssert (
 #ifdef LDAP_DEBUG
 	filter2bv_x( op, op->o_assertion, &fstr );
 
-#ifdef NEW_LOGGING
-	LDAP_LOG( OPERATION, ARGS, 
-		"parseAssert: conn %ld assert: %s\n", 
-		op->o_connid, fstr.bv_len ? fstr.bv_val : "empty" , 0 );
-#else
 	Debug( LDAP_DEBUG_ARGS, "parseAssert: conn %ld assert: %s\n",
 		op->o_connid, fstr.bv_len ? fstr.bv_val : "empty" , 0 );
-#endif
 	op->o_tmpfree( fstr.bv_val, op->o_tmpmemctx );
 #endif
 
 	op->o_assert = ctrl->ldctl_iscritical
-		? SLAP_CRITICAL_CONTROL
-		: SLAP_NONCRITICAL_CONTROL;
+		? SLAP_CONTROL_CRITICAL
+		: SLAP_CONTROL_NONCRITICAL;
 
 	rs->sr_err = LDAP_SUCCESS;
 	return LDAP_SUCCESS;
@@ -974,7 +1016,7 @@ static int parsePreRead (
 	AttributeName *an = NULL;
 	BerElement	*ber;
 
-	if ( op->o_preread != SLAP_NO_CONTROL ) {
+	if ( op->o_preread != SLAP_CONTROL_NONE ) {
 		rs->sr_text = "preread control specified multiple times";
 		return LDAP_PROTOCOL_ERROR;
 	}
@@ -998,15 +1040,24 @@ static int parsePreRead (
 	}
 
 	for( i=0; i<siz; i++ ) {
-		const char *dummy;
+		int		rc = LDAP_SUCCESS;
+		const char	*dummy = NULL;
+
 		an[i].an_desc = NULL;
 		an[i].an_oc = NULL;
-		slap_bv2ad( &an[i].an_name, &an[i].an_desc, &dummy );
+		an[i].an_oc_exclude = 0;
+		rc = slap_bv2ad( &an[i].an_name, &an[i].an_desc, &dummy );
+		if ( rc != LDAP_SUCCESS && ctrl->ldctl_iscritical ) {
+			rs->sr_text = dummy
+				? dummy
+				: "postread control: unknown attributeType";
+			return rc;
+		}
 	}
 
 	op->o_preread = ctrl->ldctl_iscritical
-		? SLAP_CRITICAL_CONTROL
-		: SLAP_NONCRITICAL_CONTROL;
+		? SLAP_CONTROL_CRITICAL
+		: SLAP_CONTROL_NONCRITICAL;
 
 	op->o_preread_attrs = an;
 
@@ -1023,7 +1074,7 @@ static int parsePostRead (
 	AttributeName *an = NULL;
 	BerElement	*ber;
 
-	if ( op->o_postread != SLAP_NO_CONTROL ) {
+	if ( op->o_postread != SLAP_CONTROL_NONE ) {
 		rs->sr_text = "postread control specified multiple times";
 		return LDAP_PROTOCOL_ERROR;
 	}
@@ -1047,15 +1098,24 @@ static int parsePostRead (
 	}
 
 	for( i=0; i<siz; i++ ) {
-		const char *dummy;
+		int		rc = LDAP_SUCCESS;
+		const char	*dummy = NULL;
+
 		an[i].an_desc = NULL;
 		an[i].an_oc = NULL;
-		slap_bv2ad( &an[i].an_name, &an[i].an_desc, &dummy );
+		an[i].an_oc_exclude = 0;
+		rc = slap_bv2ad( &an[i].an_name, &an[i].an_desc, &dummy );
+		if ( rc != LDAP_SUCCESS && ctrl->ldctl_iscritical ) {
+			rs->sr_text = dummy
+				? dummy
+				: "postread control: unknown attributeType";
+			return rc;
+		}
 	}
 
 	op->o_postread = ctrl->ldctl_iscritical
-		? SLAP_CRITICAL_CONTROL
-		: SLAP_NONCRITICAL_CONTROL;
+		? SLAP_CONTROL_CRITICAL
+		: SLAP_CONTROL_NONCRITICAL;
 
 	op->o_postread_attrs = an;
 
@@ -1069,10 +1129,10 @@ int parseValuesReturnFilter (
 	LDAPControl *ctrl )
 {
 	BerElement	*ber;
-	struct berval	fstr = { 0, NULL };
+	struct berval	fstr = BER_BVNULL;
 	const char *err_msg = "";
 
-	if ( op->o_valuesreturnfilter != SLAP_NO_CONTROL ) {
+	if ( op->o_valuesreturnfilter != SLAP_CONTROL_NONE ) {
 		rs->sr_text = "valuesReturnFilter control specified multiple times";
 		return LDAP_PROTOCOL_ERROR;
 	}
@@ -1088,7 +1148,7 @@ int parseValuesReturnFilter (
 		return LDAP_OTHER;
 	}
 	
-	rs->sr_err = get_vrFilter( op, ber, &(op->o_vrFilter), &rs->sr_text);
+	rs->sr_err = get_vrFilter( op, ber, (ValuesReturnFilter **)&(op->o_vrFilter), &rs->sr_text);
 
 	if( rs->sr_err != LDAP_SUCCESS ) {
 		if( rs->sr_err == SLAPD_DISCONNECT ) {
@@ -1105,20 +1165,14 @@ int parseValuesReturnFilter (
 		vrFilter2bv( op, op->o_vrFilter, &fstr );
 	}
 
-#ifdef NEW_LOGGING
-	LDAP_LOG( OPERATION, ARGS, 
-		"parseValuesReturnFilter: conn %d	vrFilter: %s\n", 
-		op->o_connid, fstr.bv_len ? fstr.bv_val : "empty" , 0 );
-#else
 	Debug( LDAP_DEBUG_ARGS, "	vrFilter: %s\n",
 		fstr.bv_len ? fstr.bv_val : "empty", 0, 0 );
-#endif
 	op->o_tmpfree( fstr.bv_val, op->o_tmpmemctx );
 #endif
 
 	op->o_valuesreturnfilter = ctrl->ldctl_iscritical
-		? SLAP_CRITICAL_CONTROL
-		: SLAP_NONCRITICAL_CONTROL;
+		? SLAP_CONTROL_CRITICAL
+		: SLAP_CONTROL_NONCRITICAL;
 
 	rs->sr_err = LDAP_SUCCESS;
 	return LDAP_SUCCESS;
@@ -1130,25 +1184,27 @@ static int parseSubentries (
 	SlapReply *rs,
 	LDAPControl *ctrl )
 {
-	if ( op->o_subentries != SLAP_NO_CONTROL ) {
+	if ( op->o_subentries != SLAP_CONTROL_NONE ) {
 		rs->sr_text = "subentries control specified multiple times";
 		return LDAP_PROTOCOL_ERROR;
 	}
 
 	/* FIXME: should use BER library */
 	if( ( ctrl->ldctl_value.bv_len != 3 )
-		&& ( ctrl->ldctl_value.bv_val[0] != 0x01 )
-		&& ( ctrl->ldctl_value.bv_val[1] != 0x01 ))
+		|| ( ctrl->ldctl_value.bv_val[0] != 0x01 )
+		|| ( ctrl->ldctl_value.bv_val[1] != 0x01 ))
 	{
 		rs->sr_text = "subentries control value encoding is bogus";
 		return LDAP_PROTOCOL_ERROR;
 	}
 
 	op->o_subentries = ctrl->ldctl_iscritical
-		? SLAP_CRITICAL_CONTROL
-		: SLAP_NONCRITICAL_CONTROL;
+		? SLAP_CONTROL_CRITICAL
+		: SLAP_CONTROL_NONCRITICAL;
 
-	op->o_subentries_visibility = (ctrl->ldctl_value.bv_val[2] != 0x00);
+	if ( (void *)(ctrl->ldctl_value.bv_val[2] != 0x00)) {
+		set_subentries_visibility( op );
+	}
 
 	return LDAP_SUCCESS;
 }
@@ -1160,7 +1216,7 @@ static int parsePermissiveModify (
 	SlapReply *rs,
 	LDAPControl *ctrl )
 {
-	if ( op->o_permissive_modify != SLAP_NO_CONTROL ) {
+	if ( op->o_permissive_modify != SLAP_CONTROL_NONE ) {
 		rs->sr_text = "permissiveModify control specified multiple times";
 		return LDAP_PROTOCOL_ERROR;
 	}
@@ -1171,8 +1227,8 @@ static int parsePermissiveModify (
 	}
 
 	op->o_permissive_modify = ctrl->ldctl_iscritical
-		? SLAP_CRITICAL_CONTROL
-		: SLAP_NONCRITICAL_CONTROL;
+		? SLAP_CONTROL_CRITICAL
+		: SLAP_CONTROL_NONCRITICAL;
 
 	return LDAP_SUCCESS;
 }
@@ -1184,7 +1240,7 @@ static int parseDomainScope (
 	SlapReply *rs,
 	LDAPControl *ctrl )
 {
-	if ( op->o_domain_scope != SLAP_NO_CONTROL ) {
+	if ( op->o_domain_scope != SLAP_CONTROL_NONE ) {
 		rs->sr_text = "domainScope control specified multiple times";
 		return LDAP_PROTOCOL_ERROR;
 	}
@@ -1195,98 +1251,86 @@ static int parseDomainScope (
 	}
 
 	op->o_domain_scope = ctrl->ldctl_iscritical
-		? SLAP_CRITICAL_CONTROL
-		: SLAP_NONCRITICAL_CONTROL;
+		? SLAP_CONTROL_CRITICAL
+		: SLAP_CONTROL_NONCRITICAL;
 
 	return LDAP_SUCCESS;
 }
 #endif
 
-static int parseLDAPsync (
+#ifdef LDAP_CONTROL_X_TREE_DELETE
+static int parseTreeDelete (
 	Operation *op,
 	SlapReply *rs,
 	LDAPControl *ctrl )
 {
-	ber_tag_t tag;
-	BerElement *ber;
-	ber_int_t mode;
-	ber_len_t len;
-	struct slap_session_entry *se;
-
-	if ( op->o_sync != SLAP_NO_CONTROL ) {
-		rs->sr_text = "LDAP Sync control specified multiple times";
+	if ( op->o_tree_delete != SLAP_CONTROL_NONE ) {
+		rs->sr_text = "treeDelete control specified multiple times";
 		return LDAP_PROTOCOL_ERROR;
 	}
 
+	if ( ctrl->ldctl_value.bv_len ) {
+		rs->sr_text = "treeDelete control value not empty";
+		return LDAP_PROTOCOL_ERROR;
+	}
+
+	op->o_tree_delete = ctrl->ldctl_iscritical
+		? SLAP_CONTROL_CRITICAL
+		: SLAP_CONTROL_NONCRITICAL;
+
+	return LDAP_SUCCESS;
+}
+#endif
+
+#ifdef LDAP_CONTORL_X_SEARCH_OPTIONS
+static int parseSearchOptions (
+	Operation *op,
+	SlapReply *rs,
+	LDAPControl *ctrl )
+{
+	BerElement *ber;
+	ber_int_t search_flags;
+
 	if ( ctrl->ldctl_value.bv_len == 0 ) {
-		rs->sr_text = "LDAP Sync control value is empty (or absent)";
+		rs->sr_text = "searchOptions control value not empty";
 		return LDAP_PROTOCOL_ERROR;
 	}
 
-	/* Parse the control value
-	 *      syncRequestValue ::= SEQUENCE {
-	 *              mode   ENUMERATED {
-	 *                      -- 0 unused
-	 *                      refreshOnly		(1),
-	 *                      -- 2 reserved
-	 *                      refreshAndPersist	(3)
-	 *              },
-	 *              cookie  syncCookie OPTIONAL
-	 *      }
-	 */
-
 	ber = ber_init( &ctrl->ldctl_value );
 	if( ber == NULL ) {
 		rs->sr_text = "internal error";
 		return LDAP_OTHER;
 	}
 
-	if ( (tag = ber_scanf( ber, "{i" /*}*/, &mode )) == LBER_ERROR ) {
-		rs->sr_text = "LDAP Sync control : mode decoding error";
+	if ( (tag = ber_scanf( ber, "{i}", &search_flags )) == LBER_ERROR ) {
+		rs->sr_text = "searchOptions control decoding error";
 		return LDAP_PROTOCOL_ERROR;
 	}
 
-	switch( mode ) {
-	case LDAP_SYNC_REFRESH_ONLY:
-		mode = SLAP_SYNC_REFRESH;
-		break;
-	case LDAP_SYNC_REFRESH_AND_PERSIST:
-		mode = SLAP_SYNC_REFRESH_AND_PERSIST;
-		break;
-	default:
-		rs->sr_text = "LDAP Sync control : unknown update mode";
-		return LDAP_PROTOCOL_ERROR;
-	}
-
-	tag = ber_peek_tag( ber, &len );
-
-	if ( tag == LDAP_TAG_SYNC_COOKIE ) {
-		struct berval tmp_bv;	
-		if (( ber_scanf( ber, /*{*/ "o", &tmp_bv )) == LBER_ERROR ) {
-			rs->sr_text = "LDAP Sync control : cookie decoding error";
-			return LDAP_PROTOCOL_ERROR;
-		}
-		ber_bvarray_add( &op->o_sync_state.octet_str, &tmp_bv );
-		slap_parse_sync_cookie( &op->o_sync_state );
-	}
-	if ( tag == LDAP_TAG_RELOAD_HINT ) {
-		if (( ber_scanf( ber, /*{*/ "b", &op->o_sync_rhint )) == LBER_ERROR ) {
-			rs->sr_text = "LDAP Sync control : rhint decoding error";
-			return LDAP_PROTOCOL_ERROR;
-		}
-	}
-	if (( ber_scanf( ber, /*{*/ "}")) == LBER_ERROR ) {
-			rs->sr_text = "LDAP Sync control : decoding error";
-			return LDAP_PROTOCOL_ERROR;
-	}
-
 	(void) ber_free( ber, 1 );
 
-	op->o_sync_mode = (char) mode;
+	if ( search_flags & LDAP_SEARCH_FLAG_DOMAIN_SCOPE ) {
+		if ( op->o_domain_scope != SLAP_CONTROL_NONE ) {
+			rs->sr_text = "searchOptions control specified multiple times "
+				"or with domainScope control";
+			return LDAP_PROTOCOL_ERROR;
+		}
 
-	op->o_sync = ctrl->ldctl_iscritical
-		? SLAP_CRITICAL_CONTROL
-		: SLAP_NONCRITICAL_CONTROL;
+		op->o_domain_scope = ctrl->ldctl_iscritical
+			? SLAP_CONTROL_CRITICAL
+			: SLAP_CONTROL_NONCRITICAL;
+	}
+
+	if ( search_flags & ~(LDAP_SEARCH_FLAG_DOMAIN_SCOPE) ) {
+		/* Other search flags not recognised so far,
+		 * including:
+		 *		LDAP_SEARCH_FLAG_PHANTOM_ROOM
+		 */
+		rs->sr_text = "searchOptions contained unrecongized flag";
+		return LDAP_UNWILLING_TO_PERFORM;
+	}
 
 	return LDAP_SUCCESS;
 }
+#endif
+
