@@ -1,7 +1,7 @@
-/* $OpenLDAP: pkg/ldap/servers/slapd/main.c,v 1.157.2.3 2003/12/01 00:58:35 kurt Exp $ */
+/* $OpenLDAP: pkg/ldap/servers/slapd/main.c,v 1.198.2.3 2005/03/17 22:29:49 kurt Exp $ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 1998-2003 The OpenLDAP Foundation.
+ * Copyright 1998-2005 The OpenLDAP Foundation.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -34,14 +34,12 @@
 #include <ac/wait.h>
 #include <ac/errno.h>
 
-#include "ldap_pvt.h"
-
 #include "slap.h"
 #include "lutil.h"
 #include "ldif.h"
 
 #ifdef LDAP_SLAPI
-#include "slapi.h"
+#include "slapi/slapi.h"
 #endif
 
 #ifdef LDAP_SIGCHLD
@@ -64,6 +62,31 @@ static struct sockaddr_in	bind_addr;
 #define MAIN_RETURN(x) return(x)
 #endif
 
+typedef int (MainFunc) LDAP_P(( int argc, char *argv[] ));
+extern MainFunc slapadd, slapcat, slapdn, slapindex, slappasswd,
+	slaptest, slapauth, slapacl;
+
+static struct {
+	char *name;
+	MainFunc *func;
+} tools[] = {
+	{"slapadd", slapadd},
+	{"slapcat", slapcat},
+	{"slapdn", slapdn},
+	{"slapindex", slapindex},
+	{"slappasswd", slappasswd},
+	{"slaptest", slaptest},
+	{"slapauth", slapauth},
+	{"slapacl", slapacl},
+	/* NOTE: new tools must be added in chronological order,
+	 * not in alphabetical order, because for backwards
+	 * compatibility name[4] is used to identify the
+	 * tools; so name[4]=='a' must refer to "slapadd" and
+	 * not to "slapauth".  Alphabetical order can be used
+	 * for tools whose name[4] is not used yet */
+	{NULL, NULL}
+};
+
 /*
  * when more than one slapd is running on one machine, each one might have
  * it's own LOCAL for syslogging and must have its own pid/args files
@@ -75,7 +98,6 @@ const char Versionstr[] =
 #endif
 
 #ifdef LOG_LOCAL4
-
 #define DEFAULT_SYSLOG_USER  LOG_LOCAL4
 
 typedef struct _str2intDispatch {
@@ -84,7 +106,6 @@ typedef struct _str2intDispatch {
 	int	 intVal;
 } STRDISP, *STRDISP_P;
 
-
 /* table to compute syslog-options to integer */
 static STRDISP	syslog_types[] = {
 	{ "LOCAL0", sizeof("LOCAL0"), LOG_LOCAL0 },
@@ -95,14 +116,73 @@ static STRDISP	syslog_types[] = {
 	{ "LOCAL5", sizeof("LOCAL5"), LOG_LOCAL5 },
 	{ "LOCAL6", sizeof("LOCAL6"), LOG_LOCAL6 },
 	{ "LOCAL7", sizeof("LOCAL7"), LOG_LOCAL7 },
+#ifdef LOG_USER
+	{ "USER", sizeof("USER"), LOG_USER },
+#endif
+#ifdef LOG_DAEMON
+	{ "DAEMON", sizeof("DAEMON"), LOG_DAEMON },
+#endif
 	{ NULL, 0, 0 }
 };
 
-static int   cnvt_str2int( char *, STRDISP_P, int );
-
+static int cnvt_str2int( char *, STRDISP_P, int );
 #endif	/* LOG_LOCAL4 */
 
-static int check_config = 0;
+#define CHECK_NONE	0x00
+#define CHECK_CONFIG	0x01
+static int check = CHECK_NONE;
+static int version = 0;
+
+void *slap_tls_ctx;
+
+static int
+slapd_opt_slp( const char *val, void *arg )
+{
+#ifdef HAVE_SLP
+	/* NULL is default */
+	if ( val == NULL || strcasecmp( val, "on" ) == 0 ) {
+		slapd_register_slp = 1;
+
+	} else if ( strcasecmp( val, "off" ) == 0 ) {
+		slapd_register_slp = 0;
+
+	/* NOTE: add support for URL specification? */
+
+	} else {
+		fprintf(stderr, "unrecognized value \"%s\" for SLP option\n", val );
+		return -1;
+	}
+
+	return 0;
+		
+#else
+	fputs( "slapd: SLP support is not available\n", stderr );
+	return 0;
+#endif
+}
+
+/*
+ * Option helper structure:
+ * 
+ * oh_nam	is left-hand part of <option>[=<value>]
+ * oh_fnc	is handler function
+ * oh_arg	is an optional arg to oh_fnc
+ * oh_usage	is the one-line usage string related to the option,
+ *		which is assumed to start with <option>[=<value>]
+ *
+ * please leave valid options in the structure, and optionally #ifdef
+ * their processing inside the helper, so that reasonable and helpful
+ * error messages can be generated if a disabled option is requested.
+ */
+struct option_helper {
+	struct berval	oh_name;
+	int		(*oh_fnc)(const char *val, void *arg);
+	void		*oh_arg;
+	const char	*oh_usage;
+} option_helpers[] = {
+	{ BER_BVC("slp"),	slapd_opt_slp,	NULL, "slp[={on|off}] enable/disable SLP" },
+	{ BER_BVNULL }
+};
 
 static void
 usage( char *name )
@@ -112,6 +192,8 @@ usage( char *name )
 	fprintf( stderr,
 		"\t-4\t\tIPv4 only\n"
 		"\t-6\t\tIPv6 only\n"
+		"\t-T {add|auth|cat|dn|index|passwd|test}\n"
+		"\t\t\tRun in Tool mode\n"
 		"\t-c cookie\tSync cookie of consumer\n"
 		"\t-d level\tDebug level" "\n"
 		"\t-f filename\tConfiguration file\n"
@@ -123,14 +205,26 @@ usage( char *name )
 		"\t-l facility\tSyslog facility (default: LOCAL4)\n"
 #endif
 		"\t-n serverName\tService name\n"
+		"\t-o <opt>[=val] generic means to specify options" );
+	if ( !BER_BVISNULL( &option_helpers[0].oh_name ) ) {
+		int	i;
+
+		fprintf( stderr, "; supported options:\n" );
+		for ( i = 0; !BER_BVISNULL( &option_helpers[i].oh_name ); i++) {
+			fprintf( stderr, "\t\t%s\n", option_helpers[i].oh_usage );
+		}
+	} else {
+		fprintf( stderr, "\n" );
+	}
+	fprintf( stderr,	
 #ifdef HAVE_CHROOT
 		"\t-r directory\tSandbox directory to chroot to\n"
 #endif
 		"\t-s level\tSyslog level\n"
-		"\t-t\t\tCheck configuration file and exit\n"
 #if defined(HAVE_SETUID) && defined(HAVE_SETGID)
 		"\t-u user\t\tUser (id or name) to run as\n"
 #endif
+		"\t-V\t\tprint version info (-VV only)\n"
     );
 }
 
@@ -159,13 +253,14 @@ int main( int argc, char **argv )
 
 #ifdef HAVE_NT_SERVICE_MANAGER
 	char		*configfile = ".\\slapd.conf";
+	char		*configdir = ".\\slapd.d";
 #else
 	char		*configfile = SLAPD_DEFAULT_CONFIGFILE;
+	char		*configdir = SLAPD_DEFAULT_CONFIGDIR;
 #endif
-	char	    *serverName = NULL;
+	char	    *serverName;
 	int	    serverMode = SLAP_SERVER_MODE;
 
-	struct berval cookie = { 0, NULL };
 	struct sync_cookie *scp = NULL;
 	struct sync_cookie *scp_entry = NULL;
 
@@ -175,71 +270,69 @@ int main( int argc, char **argv )
 		leakfile = stderr;
 	}
 #endif
+	char	*serverNamePrefix = "";
+	size_t	l;
 
-	sl_mem_init();
+	slap_sl_mem_init();
+
+	serverName = lutil_progname( "slapd", argc, argv );
+
+	if ( strcmp( serverName, "slapd" ) ) {
+		for (i=0; tools[i].name; i++) {
+			if ( !strcmp( serverName, tools[i].name ) ) {
+				rc = tools[i].func(argc, argv);
+				MAIN_RETURN(rc);
+			}
+		}
+	}
 
 #ifdef HAVE_NT_SERVICE_MANAGER
 	{
 		int *i;
 		char *newConfigFile;
+		char *newConfigDir;
 		char *newUrls;
 		char *regService = NULL;
 
 		if ( is_NT_Service ) {
-			serverName = argv[0];
 			lutil_CommenceStartupProcessing( serverName, slap_sig_shutdown );
 			if ( strcmp(serverName, SERVICE_NAME) )
 			    regService = serverName;
 		}
 
 		i = (int*)lutil_getRegParam( regService, "DebugLevel" );
-		if ( i != NULL ) 
-		{
+		if ( i != NULL ) {
 			slap_debug = *i;
-#ifdef NEW_LOGGING
-			lutil_log_initialize( argc, argv );
-			LDAP_LOG( SLAPD, INFO, 
-				"main: new debug level from registry is: %d\n", 
-				slap_debug, 0, 0 );
-#else
-			Debug( LDAP_DEBUG_ANY, "new debug level from registry is: %d\n", slap_debug, 0, 0 );
-#endif
+			Debug( LDAP_DEBUG_ANY,
+				"new debug level from registry is: %d\n", slap_debug, 0, 0 );
 		}
 
 		newUrls = (char *) lutil_getRegParam(regService, "Urls");
-		if (newUrls)
-		{
+		if (newUrls) {
 		    if (urls)
 			ch_free(urls);
 
 		    urls = ch_strdup(newUrls);
-#ifdef NEW_LOGGING
-		    LDAP_LOG( SLAPD, INFO, 
-				"main: new urls from registry: %s\n", urls, 0, 0 );
-#else
 		    Debug(LDAP_DEBUG_ANY, "new urls from registry: %s\n",
-			  urls, 0, 0);
-#endif
-
+				urls, 0, 0);
 		}
 
 		newConfigFile = (char*)lutil_getRegParam( regService, "ConfigFile" );
-		if ( newConfigFile != NULL ) 
-		{
+		if ( newConfigFile != NULL ) {
 			configfile = newConfigFile;
-#ifdef NEW_LOGGING
-			LDAP_LOG( SLAPD, INFO, 
-				"main: new config file from registry is: %s\n", configfile, 0, 0 );
-#else
 			Debug ( LDAP_DEBUG_ANY, "new config file from registry is: %s\n", configfile, 0, 0 );
-#endif
+		}
 
+		newConfigDir = (char*)lutil_getRegParam( regService, "ConfigDir" );
+		if ( newConfigDir != NULL ) {
+			configdir = newConfigDir;
+			Debug ( LDAP_DEBUG_ANY, "new config dir from registry is: %s\n", configdir, 0, 0 );
 		}
 	}
 #endif
 
 	while ( (i = getopt( argc, argv,
-			     "c:d:f:h:s:n:t"
+			     "c:d:f:F:h:n:o:s:StT:V"
 #if LDAP_PF_INET6
 				"46"
 #endif
@@ -271,21 +364,14 @@ int main( int argc, char **argv )
 		case 'c':	/* provide sync cookie, override if exist in replica */
 			scp = (struct sync_cookie *) ch_calloc( 1,
 										sizeof( struct sync_cookie ));
-			ber_str2bv( optarg, strlen( optarg ), 1, &cookie );
-			ber_bvarray_add( &scp->octet_str, &cookie );
+			ber_str2bv( optarg, 0, 1, &scp->octet_str );
 			slap_parse_sync_cookie( scp );
 
 			LDAP_STAILQ_FOREACH( scp_entry, &slap_sync_cookie, sc_next ) {
 				if ( scp->rid == scp_entry->rid ) {
-#ifdef NEW_LOGGING
-					LDAP_LOG( OPERATION, CRIT,
-							"main: duplicated replica id in cookies\n",
-							0, 0, 0 );
-#else
 					Debug( LDAP_DEBUG_ANY,
 						    "main: duplicated replica id in cookies\n",
 							0, 0, 0 );
-#endif
 					slap_sync_cookie_free( scp, 1 );
 					goto destroy;
 				}
@@ -308,6 +394,44 @@ int main( int argc, char **argv )
 			configfile = ch_strdup( optarg );
 			break;
 
+		case 'F':	/* use config dir */
+			configdir = ch_strdup( optarg );
+			break;
+
+		case 'o': {
+			char		*val = strchr( optarg, '=' );
+			struct berval	opt;
+			int		i;
+
+			opt.bv_val = optarg;
+			
+			if ( val ) {
+				opt.bv_len = ( val - optarg );
+				val++;
+			
+			} else {
+				opt.bv_len = strlen( optarg );
+			}
+
+			for ( i = 0; !BER_BVISNULL( &option_helpers[i].oh_name ); i++ ) {
+				if ( ber_bvstrcasecmp( &option_helpers[i].oh_name, &opt ) == 0 ) {
+					assert( option_helpers[i].oh_fnc != NULL );
+					if ( (*option_helpers[i].oh_fnc)( val, option_helpers[i].oh_arg ) == -1 ) {
+						/* we assume the option parsing helper
+						 * issues appropriate and self-explanatory
+						 * error messages... */
+						goto stop;
+					}
+					break;
+				}
+			}
+
+			if ( BER_BVISNULL( &option_helpers[i].oh_name ) ) {
+				goto unhandled_option;
+			}
+			break;
+		}
+
 		case 's':	/* set syslog level */
 			ldap_syslog = atoi( optarg );
 			break;
@@ -339,15 +463,46 @@ int main( int argc, char **argv )
 #endif /* SETUID && GETUID */
 
 		case 'n':  /* NT service name */
-			if( serverName != NULL ) free( serverName );
 			serverName = ch_strdup( optarg );
 			break;
 
 		case 't':
-			check_config++;
+			/* deprecated; use slaptest instead */
+			fprintf( stderr, "option -t deprecated; "
+				"use slaptest command instead\n" );
+			check |= CHECK_CONFIG;
 			break;
 
+		case 'V':
+			version++;
+			break;
+
+		case 'T':
+			/* try full option string first */
+			for ( i = 0; tools[i].name; i++ ) {
+				if ( strcmp( optarg, &tools[i].name[4] ) == 0 ) {
+					rc = tools[i].func( argc, argv );
+					MAIN_RETURN( rc );
+				}
+			}
+
+			/* try bits of option string (backward compatibility for single char) */
+			l = strlen( optarg );
+			for ( i = 0; tools[i].name; i++ ) {
+				if ( strncmp( optarg, &tools[i].name[4], l ) == 0 ) {
+					rc = tools[i].func( argc, argv );
+					MAIN_RETURN( rc );
+				}
+			}
+			
+			/* issue error */
+			serverName = optarg;
+			serverNamePrefix = "slap";
+			fprintf( stderr, "program name \"%s%s\" unrecognized; "
+					"aborting...\n", serverNamePrefix, serverName );
+			/* FALLTHRU */
 		default:
+unhandled_option:;
 			usage( argv[0] );
 			rc = 1;
 			SERVICE_EXIT( ERROR_SERVICE_SPECIFIC_ERROR, 15 );
@@ -355,36 +510,39 @@ int main( int argc, char **argv )
 		}
 	}
 
-#ifdef NEW_LOGGING
-	lutil_log_initialize( argc, argv );
-#else
-	lutil_set_debug_level( "slapd", slap_debug );
+	(void) ldap_pvt_thread_initialize();
+
 	ber_set_option(NULL, LBER_OPT_DEBUG_LEVEL, &slap_debug);
 	ldap_set_option(NULL, LDAP_OPT_DEBUG_LEVEL, &slap_debug);
 	ldif_debug = slap_debug;
-#endif
 
-#ifdef NEW_LOGGING
-	LDAP_LOG( SLAPD, INFO, "%s", Versionstr, 0, 0 );
-#else
-	Debug( LDAP_DEBUG_TRACE, "%s", Versionstr, 0, 0 );
-#endif
-
-	if( serverName == NULL ) {
-		if ( (serverName = strrchr( argv[0], *LDAP_DIRSEP )) == NULL ) {
-			serverName = argv[0];
-		} else {
-			serverName = serverName + 1;
-		}
+	if ( version ) {
+		fprintf( stderr, "%s\n", Versionstr );
+		if ( version > 1 ) goto stop;
 	}
 
-#ifdef LOG_LOCAL4
-	openlog( serverName, OPENLOG_OPTIONS, syslogUser );
-#elif LOG_DEBUG
-	openlog( serverName, OPENLOG_OPTIONS );
+	{
+		char *logName;
+#ifdef HAVE_EBCDIC
+		logName = ch_strdup( serverName );
+		__atoe( logName );
+#else
+		logName = serverName;
 #endif
 
-	if( !check_config && slapd_daemon_init( urls ) != 0 ) {
+#ifdef LOG_LOCAL4
+		openlog( logName, OPENLOG_OPTIONS, syslogUser );
+#elif LOG_DEBUG
+		openlog( logName, OPENLOG_OPTIONS );
+#endif
+#ifdef HAVE_EBCDIC
+		free( logName );
+#endif
+	}
+
+	Debug( LDAP_DEBUG_ANY, "%s", Versionstr, 0, 0 );
+
+	if( check == CHECK_NONE && slapd_daemon_init( urls ) != 0 ) {
 		rc = 1;
 		SERVICE_EXIT( ERROR_SERVICE_SPECIFIC_ERROR, 16 );
 		goto stop;
@@ -424,13 +582,9 @@ int main( int argc, char **argv )
 #endif
 
 	if ( slap_schema_init( ) != 0 ) {
-#ifdef NEW_LOGGING
-		LDAP_LOG( OPERATION, CRIT, "main: schema initialization error\n", 0, 0, 0 );
-#else
 		Debug( LDAP_DEBUG_ANY,
 		    "schema initialization error\n",
 		    0, 0, 0 );
-#endif
 
 		goto destroy;
 	}
@@ -442,13 +596,9 @@ int main( int argc, char **argv )
 	}
 
 	if ( slap_controls_init( ) != 0 ) {
-#ifdef NEW_LOGGING
-		LDAP_LOG( OPERATION, CRIT, "main: controls initialization error\n", 0, 0, 0 );
-#else
 		Debug( LDAP_DEBUG_ANY,
 		    "controls initialization error\n",
 		    0, 0, 0 );
-#endif
 
 		goto destroy;
 	}
@@ -464,55 +614,54 @@ int main( int argc, char **argv )
 #endif
 
 #ifdef LDAP_SLAPI
-	if ( slapi_init() != 0 ) {
-#ifdef NEW_LOGGING
-		LDAP_LOG( OPERATION, CRIT, "main: slapi initialization error\n", 0, 0, 0 );
-#else
+	if ( slapi_int_initialize() != 0 ) {
 		Debug( LDAP_DEBUG_ANY,
 		    "slapi initialization error\n",
 		    0, 0, 0 );
-#endif
 
 		goto destroy;
 	}
 #endif /* LDAP_SLAPI */
 
-	if ( read_config( configfile, 0 ) != 0 ) {
+	if ( frontend_init() ) {
+		goto destroy;
+	}
+
+	if ( overlay_init() ) {
+		goto destroy;
+	}
+
+#ifdef SLAP_DYNACL
+	if ( acl_init() ) {
+		goto destroy;
+	}
+#endif /* SLAP_DYNACL */
+
+	if ( read_config( configfile, configdir ) != 0 ) {
 		rc = 1;
 		SERVICE_EXIT( ERROR_SERVICE_SPECIFIC_ERROR, 19 );
 
-		if ( check_config ) {
+		if ( check & CHECK_CONFIG ) {
 			fprintf( stderr, "config check failed\n" );
 		}
 
 		goto destroy;
 	}
 
-	if ( check_config ) {
-		rc = 0;
+	if ( check & CHECK_CONFIG ) {
 		fprintf( stderr, "config check succeeded\n" );
-		goto destroy;
-	}
 
-	if ( glue_sub_init( ) != 0 ) {
-#ifdef NEW_LOGGING
-		LDAP_LOG( SLAPD, CRIT, "main: subordinate config error\n", 0, 0, 0 );
-#else
-		Debug( LDAP_DEBUG_ANY,
-		    "subordinate config error\n",
-		    0, 0, 0 );
-#endif
-		goto destroy;
+		check &= ~CHECK_CONFIG;
+		if ( check == CHECK_NONE ) {
+			rc = 0;
+			goto destroy;
+		}
 	}
 
 	if ( slap_schema_check( ) != 0 ) {
-#ifdef NEW_LOGGING
-		LDAP_LOG( SLAPD, CRIT, "main: schema prep error\n", 0, 0, 0 );
-#else
 		Debug( LDAP_DEBUG_ANY,
 		    "schema prep error\n",
 		    0, 0, 0 );
-#endif
 
 		goto destroy;
 	}
@@ -520,30 +669,36 @@ int main( int argc, char **argv )
 #ifdef HAVE_TLS
 	rc = ldap_pvt_tls_init();
 	if( rc != 0) {
-#ifdef NEW_LOGGING
-		LDAP_LOG( SLAPD, CRIT, "main: tls init failed: %d\n", rc, 0, 0 );
-#else
 		Debug( LDAP_DEBUG_ANY,
 		    "main: TLS init failed: %d\n",
 		    0, 0, 0 );
-#endif
 		rc = 1;
 		SERVICE_EXIT( ERROR_SERVICE_SPECIFIC_ERROR, 20 );
 		goto destroy;
 	}
 
-	rc = ldap_pvt_tls_init_def_ctx();
-	if( rc != 0) {
-#ifdef NEW_LOGGING
-		LDAP_LOG( SLAPD, CRIT, "main: tls init def ctx failed: %d\n", rc, 0, 0 );
-#else
-		Debug( LDAP_DEBUG_ANY,
-		    "main: TLS init def ctx failed: %d\n",
-		    rc, 0, 0 );
-#endif
-		rc = 1;
-		SERVICE_EXIT( ERROR_SERVICE_SPECIFIC_ERROR, 20 );
-		goto destroy;
+	{
+		void *def_ctx = NULL;
+
+		/* Save existing default ctx, if any */
+		ldap_pvt_tls_get_option( NULL, LDAP_OPT_X_TLS_CTX, &def_ctx );
+
+		/* Force new ctx to be created */
+		ldap_pvt_tls_set_option( NULL, LDAP_OPT_X_TLS_CTX, NULL );
+
+		rc = ldap_pvt_tls_init_def_ctx();
+		if( rc != 0) {
+			Debug( LDAP_DEBUG_ANY,
+			    "main: TLS init def ctx failed: %d\n",
+			    rc, 0, 0 );
+			rc = 1;
+			SERVICE_EXIT( ERROR_SERVICE_SPECIFIC_ERROR, 20 );
+			goto destroy;
+		}
+		/* Retrieve slapd's own ctx */
+		ldap_pvt_tls_get_option( NULL, LDAP_OPT_X_TLS_CTX, &slap_tls_ctx );
+		/* Restore previous ctx */
+		ldap_pvt_tls_set_option( NULL, LDAP_OPT_X_TLS_CTX, def_ctx );
 	}
 #endif
 
@@ -558,6 +713,9 @@ int main( int argc, char **argv )
 #endif
 	(void) SIGNAL( SIGINT, slap_sig_shutdown );
 	(void) SIGNAL( SIGTERM, slap_sig_shutdown );
+#ifdef SIGTRAP
+	(void) SIGNAL( SIGTRAP, slap_sig_shutdown );
+#endif
 #ifdef LDAP_SIGCHLD
 	(void) SIGNAL( LDAP_SIGCHLD, wait4child );
 #endif
@@ -586,11 +744,7 @@ int main( int argc, char **argv )
 		goto shutdown;
 	}
 
-#ifdef NEW_LOGGING
-	LDAP_LOG( SLAPD, INFO, "main: slapd starting.\n", 0, 0, 0 );
-#else
 	Debug( LDAP_DEBUG_ANY, "slapd starting\n", 0, 0, 0 );
-#endif
 
 
 	if ( slapd_pid_file != NULL ) {
@@ -662,11 +816,7 @@ stop:
 	lutil_LogStoppedEvent( serverName );
 #endif
 
-#ifdef NEW_LOGGING
-	LDAP_LOG( SLAPD, CRIT, "main: slapd stopped.\n", 0, 0, 0 );
-#else
 	Debug( LDAP_DEBUG_ANY, "slapd stopped.\n", 0, 0, 0 );
-#endif
 
 
 #ifdef HAVE_NT_SERVICE_MANAGER
