@@ -1,8 +1,8 @@
 /* root_dse.c - Provides the Root DSA-Specific Entry */
-/* $OpenLDAP: pkg/ldap/servers/slapd/root_dse.c,v 1.81.2.6 2003/12/01 00:58:36 kurt Exp $ */
+/* $OpenLDAP: pkg/ldap/servers/slapd/root_dse.c,v 1.95.2.3 2005/03/14 22:33:12 kurt Exp $ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 1999-2003 The OpenLDAP Foundation.
+ * Copyright 1999-2005 The OpenLDAP Foundation.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -25,17 +25,20 @@
 #include "lber_pvt.h"
 
 #ifdef LDAP_SLAPI
-#include "slapi.h"
+#include "slapi/slapi.h"
 #endif
 
 static struct berval supportedFeatures[] = {
-	BER_BVC(LDAP_FEATURE_ALL_OPERATIONAL_ATTRS), /* All Op Attrs (+) */
-	BER_BVC(LDAP_FEATURE_OBJECTCLASS_ATTRS), /* OCs in Attrs List (+person) */
-	BER_BVC(LDAP_FEATURE_ABSOLUTE_FILTERS), /* (&) and (|) search filters */
+	BER_BVC(LDAP_FEATURE_ALL_OP_ATTRS),			/* All Op Attrs (+) */
+	BER_BVC(LDAP_FEATURE_OBJECTCLASS_ATTRS),	/* OCs in Attrs List (@class) */
+	BER_BVC(LDAP_FEATURE_ABSOLUTE_FILTERS),		/* (&) and (|) search filters */
 	BER_BVC(LDAP_FEATURE_LANGUAGE_TAG_OPTIONS), /* Language Tag Options */
-	BER_BVC(LDAP_FEATURE_LANGUAGE_RANGE_OPTIONS), /* Language Range Options */
-#ifdef LDAP_DEVEL
-	BER_BVC(LDAP_FEATURE_MODIFY_INCREMENT), /* Modify/increment */
+	BER_BVC(LDAP_FEATURE_LANGUAGE_RANGE_OPTIONS),/* Language Range Options */
+#ifdef LDAP_FEATURE_SUBORDINATE_SCOPE
+	BER_BVC(LDAP_FEATURE_SUBORDINATE_SCOPE),	/* "children" search scope */
+#endif
+#ifdef LDAP_FEATURE_MODIFY_INCREMENT
+	BER_BVC(LDAP_FEATURE_MODIFY_INCREMENT),		/* Modify/increment */
 #endif
 	{0,NULL}
 };
@@ -70,6 +73,8 @@ root_dse_info(
 		= slap_schema.si_ad_supportedFeatures;
 	AttributeDescription *ad_monitorContext
 		= slap_schema.si_ad_monitorContext;
+	AttributeDescription *ad_configContext
+		= slap_schema.si_ad_configContext;
 	AttributeDescription *ad_ref
 		= slap_schema.si_ad_ref;
 
@@ -79,13 +84,8 @@ root_dse_info(
 	e = (Entry *) SLAP_CALLOC( 1, sizeof(Entry) );
 
 	if( e == NULL ) {
-#ifdef NEW_LOGGING
-		LDAP_LOG( OPERATION, ERR,
-			"root_dse_info: SLAP_CALLOC failed", 0, 0, 0 );
-#else
 		Debug( LDAP_DEBUG_ANY,
 			"root_dse_info: SLAP_CALLOC failed", 0, 0, 0 );
-#endif
 		return LDAP_OTHER;
 	}
 
@@ -117,7 +117,12 @@ root_dse_info(
 	}
 
 	for ( i = 0; i < nbackends; i++ ) {
-		if ( backends[i].be_flags & SLAP_BFLAG_MONITOR ) {
+		if ( backends[i].be_suffix == NULL
+				|| backends[i].be_nsuffix == NULL ) {
+			/* no suffix! */
+			continue;
+		}
+		if ( SLAP_MONITOR( &backends[i] )) {
 			vals[0] = backends[i].be_suffix[0];
 			nvals[0] = backends[i].be_nsuffix[0];
 			if( attr_merge( e, ad_monitorContext, vals, nvals ) ) {
@@ -125,7 +130,15 @@ root_dse_info(
 			}
 			continue;
 		}
-		if ( SLAP_GLUE_SUBORDINATE( &backends[i] ) ) {
+		if ( SLAP_CONFIG( &backends[i] )) {
+			vals[0] = backends[i].be_suffix[0];
+			nvals[0] = backends[i].be_nsuffix[0];
+			if( attr_merge( e, ad_configContext, vals, nvals ) ) {
+				return LDAP_OTHER;
+			}
+			continue;
+		}
+		if ( SLAP_GLUE_SUBORDINATE( &backends[i] ) && !SLAP_GLUE_ADVERTISE( &backends[i] ) ) {
 			continue;
 		}
 		for ( j = 0; backends[i].be_suffix[j].bv_val != NULL; j++ ) {
@@ -151,7 +164,7 @@ root_dse_info(
 
 #ifdef LDAP_SLAPI
 	/* netscape supportedExtension */
-	for ( i = 0; (bv = ns_get_supported_extop(i)) != NULL; i++ ) {
+	for ( i = 0; (bv = slapi_int_get_supported_extop(i)) != NULL; i++ ) {
 		vals[0] = *bv;
 		if( attr_merge( e, ad_supportedExtension, vals, NULL )) {
 			return LDAP_OTHER;
@@ -238,13 +251,8 @@ int read_root_dse_file( const char *fname )
 
 	usr_attr = (Entry *) SLAP_CALLOC( 1, sizeof(Entry) );
 	if( usr_attr == NULL ) {
-#ifdef NEW_LOGGING
-		LDAP_LOG( OPERATION, ERR,
-			"read_root_dse_file: SLAP_CALLOC failed", 0, 0, 0 );
-#else
 		Debug( LDAP_DEBUG_ANY,
 			"read_root_dse_file: SLAP_CALLOC failed", 0, 0, 0 );
-#endif
 		fclose( fp );
 		return LDAP_OTHER;
 	}
