@@ -1,8 +1,8 @@
 /* schema_check.c - routines to enforce schema definitions */
-/* $OpenLDAP: pkg/ldap/servers/slapd/schema_check.c,v 1.76.2.3 2003/12/01 00:58:36 kurt Exp $ */
+/* $OpenLDAP: pkg/ldap/servers/slapd/schema_check.c,v 1.94.2.2 2005/03/17 22:29:51 kurt Exp $ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 1998-2003 The OpenLDAP Foundation.
+ * Copyright 1998-2005 The OpenLDAP Foundation.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -23,7 +23,6 @@
 #include <ac/socket.h>
 
 #include "slap.h"
-#include "ldap_pvt.h"
 
 static char * oc_check_required(
 	Entry *e,
@@ -104,14 +103,9 @@ entry_schema_check(
 				"attribute '%s' cannot have multiple values",
 				type );
 
-#ifdef NEW_LOGGING
-			LDAP_LOG( OPERATION, INFO, 
-				"entry_schema_check: dn=\"%s\" %s\n", e->e_dn, textbuf, 0 );
-#else
 			Debug( LDAP_DEBUG_ANY,
 			    "Entry (%s), %s\n",
 			    e->e_dn, textbuf, 0 );
-#endif
 
 			return LDAP_CONSTRAINT_VIOLATION;
 		}
@@ -123,15 +117,9 @@ entry_schema_check(
 	/* find the structural object class attribute */
 	asc = attr_find( e->e_attrs, ad_structuralObjectClass );
 	if ( asc == NULL ) {
-#ifdef NEW_LOGGING
-		LDAP_LOG( OPERATION, INFO, 
-			"entry_schema_check: No structuralObjectClass for entry (%s)\n", 
-			e->e_dn, 0, 0 );
-#else
 		Debug( LDAP_DEBUG_ANY,
 			"No structuralObjectClass for entry (%s)\n",
 		    e->e_dn, 0, 0 );
-#endif
 
 		*text = "no structuralObjectClass operational attribute";
 		return LDAP_OTHER;
@@ -147,14 +135,9 @@ entry_schema_check(
 			"unrecognized structuralObjectClass '%s'",
 			asc->a_vals[0].bv_val );
 
-#ifdef NEW_LOGGING
-		LDAP_LOG( OPERATION, INFO, 
-			"entry_schema_check: dn (%s), %s\n", e->e_dn, textbuf, 0 );
-#else
 		Debug( LDAP_DEBUG_ANY,
 			"entry_check_schema(%s): %s\n",
 			e->e_dn, textbuf, 0 );
-#endif
 
 		return LDAP_OBJECT_CLASS_VIOLATION;
 	}
@@ -164,14 +147,9 @@ entry_schema_check(
 			"structuralObjectClass '%s' is not STRUCTURAL",
 			asc->a_vals[0].bv_val );
 
-#ifdef NEW_LOGGING
-		LDAP_LOG( OPERATION, INFO, 
-			"entry_schema_check: dn (%s), %s\n", e->e_dn, textbuf, 0 );
-#else
 		Debug( LDAP_DEBUG_ANY,
 			"entry_check_schema(%s): %s\n",
 			e->e_dn, textbuf, 0 );
-#endif
 
 		return LDAP_OTHER;
 	}
@@ -181,14 +159,9 @@ entry_schema_check(
 			"structuralObjectClass '%s' is OBSOLETE",
 			asc->a_vals[0].bv_val );
 
-#ifdef NEW_LOGGING
-		LDAP_LOG( OPERATION, INFO, 
-			"entry_schema_check: dn (%s), %s\n", e->e_dn, textbuf, 0 );
-#else
 		Debug( LDAP_DEBUG_ANY,
 			"entry_check_schema(%s): %s\n",
 			e->e_dn, textbuf, 0 );
-#endif
 
 		return LDAP_OBJECT_CLASS_VIOLATION;
 	}
@@ -196,14 +169,8 @@ entry_schema_check(
 	/* find the object class attribute */
 	aoc = attr_find( e->e_attrs, ad_objectClass );
 	if ( aoc == NULL ) {
-#ifdef NEW_LOGGING
-		LDAP_LOG( OPERATION, INFO, 
-			"entry_schema_check: No objectClass for entry (%s).\n", 
-			e->e_dn, 0, 0 );
-#else
 		Debug( LDAP_DEBUG_ANY, "No objectClass for entry (%s)\n",
 		    e->e_dn, 0, 0 );
-#endif
 
 		*text = "no objectClass attribute";
 		return LDAP_OBJECT_CLASS_VIOLATION;
@@ -258,14 +225,9 @@ entry_schema_check(
 				"content rule '%s' is obsolete",
 				ldap_contentrule2name( &cr->scr_crule ));
 
-#ifdef NEW_LOGGING
-			LDAP_LOG( OPERATION, INFO, 
-				"entry_schema_check: dn=\"%s\" %s", e->e_dn, textbuf, 0 );
-#else
 			Debug( LDAP_DEBUG_ANY,
 				"Entry (%s): %s\n",
 				e->e_dn, textbuf, 0 );
-#endif
 
 			return LDAP_OBJECT_CLASS_VIOLATION;
 		}
@@ -286,14 +248,9 @@ entry_schema_check(
 					ldap_contentrule2name( &cr->scr_crule ),
 					at->sat_cname.bv_val );
 
-#ifdef NEW_LOGGING
-				LDAP_LOG( OPERATION, INFO, 
-					"entry_schema_check: dn=\"%s\" %s", e->e_dn, textbuf, 0 );
-#else
 				Debug( LDAP_DEBUG_ANY,
 					"Entry (%s): %s\n",
 					e->e_dn, textbuf, 0 );
-#endif
 
 				return LDAP_OBJECT_CLASS_VIOLATION;
 			}
@@ -315,14 +272,9 @@ entry_schema_check(
 					ldap_contentrule2name( &cr->scr_crule ),
 					at->sat_cname.bv_val );
 
-#ifdef NEW_LOGGING
-				LDAP_LOG( OPERATION, INFO, 
-					"entry_schema_check: dn=\"%s\" %s", e->e_dn, textbuf, 0 );
-#else
 				Debug( LDAP_DEBUG_ANY,
 					"Entry (%s): %s\n",
 					e->e_dn, textbuf, 0 );
-#endif
 
 				return LDAP_OBJECT_CLASS_VIOLATION;
 			}
@@ -336,14 +288,9 @@ entry_schema_check(
 				"unrecognized objectClass '%s'",
 				aoc->a_vals[i].bv_val );
 
-#ifdef NEW_LOGGING
-			LDAP_LOG( OPERATION, INFO, 
-				"entry_schema_check: dn (%s), %s\n", e->e_dn, textbuf, 0 );
-#else
 			Debug( LDAP_DEBUG_ANY,
 				"entry_check_schema(%s): %s\n",
 				e->e_dn, textbuf, 0 );
-#endif
 
 			return LDAP_OBJECT_CLASS_VIOLATION;
 		}
@@ -354,14 +301,9 @@ entry_schema_check(
 				"objectClass '%s' is OBSOLETE",
 				aoc->a_vals[i].bv_val );
 
-#ifdef NEW_LOGGING
-			LDAP_LOG( OPERATION, INFO, 
-				"entry_schema_check: dn (%s), %s\n", e->e_dn, textbuf, 0 );
-#else
 			Debug( LDAP_DEBUG_ANY,
 				"entry_check_schema(%s): %s\n",
 				e->e_dn, textbuf, 0 );
-#endif
 
 			return LDAP_OBJECT_CLASS_VIOLATION;
 		}
@@ -389,15 +331,9 @@ entry_schema_check(
 								"unrecognized objectClass '%s'",
 								aoc->a_vals[i].bv_val );
 
-#ifdef NEW_LOGGING
-							LDAP_LOG( OPERATION, INFO, 
-								"entry_schema_check: dn (%s), %s\n",
-								e->e_dn, textbuf, 0 );
-#else
 							Debug( LDAP_DEBUG_ANY,
 								"entry_check_schema(%s): %s\n",
 								e->e_dn, textbuf, 0 );
-#endif
 
 							return LDAP_OBJECT_CLASS_VIOLATION;
 						}
@@ -422,15 +358,9 @@ entry_schema_check(
 						"abstract objectClass '%s' not allowed",
 						aoc->a_vals[i].bv_val );
 
-#ifdef NEW_LOGGING
-					LDAP_LOG( OPERATION, INFO, 
-						"entry_schema_check: dn (%s), %s\n", 
-						e->e_dn, textbuf, 0 );
-#else
 					Debug( LDAP_DEBUG_ANY,
 						"entry_check_schema(%s): %s\n",
 						e->e_dn, textbuf, 0 );
-#endif
 
 					return LDAP_OBJECT_CLASS_VIOLATION;
 				}
@@ -466,15 +396,9 @@ entry_schema_check(
 						ldap_contentrule2name( &cr->scr_crule ),
 						oc->soc_cname.bv_val );
 
-#ifdef NEW_LOGGING
-					LDAP_LOG( OPERATION, INFO, 
-						"entry_schema_check: dn=\"%s\" %s",
-						e->e_dn, textbuf, 0 );
-#else
 					Debug( LDAP_DEBUG_ANY,
 						"Entry (%s): %s\n",
 						e->e_dn, textbuf, 0 );
-#endif
 
 					return LDAP_OBJECT_CLASS_VIOLATION;
 				}
@@ -486,14 +410,9 @@ entry_schema_check(
 					"object class '%s' requires attribute '%s'",
 					aoc->a_vals[i].bv_val, s );
 
-#ifdef NEW_LOGGING
-				LDAP_LOG( OPERATION, INFO, 
-					"entry_schema_check: dn=\"%s\" %s", e->e_dn, textbuf, 0 );
-#else
 				Debug( LDAP_DEBUG_ANY,
 					"Entry (%s): %s\n",
 					e->e_dn, textbuf, 0 );
-#endif
 
 				return LDAP_OBJECT_CLASS_VIOLATION;
 			}
@@ -505,6 +424,7 @@ entry_schema_check(
 	}
 
 	if( extensible ) {
+		*text = NULL;
 		return LDAP_SUCCESS;
 	}
 
@@ -544,19 +464,15 @@ entry_schema_check(
 				"attribute '%s' not allowed",
 				type );
 
-#ifdef NEW_LOGGING
-			LDAP_LOG( OPERATION, INFO, 
-				"entry_schema_check: dn=\"%s\" %s\n", e->e_dn, textbuf, 0);
-#else
 			Debug( LDAP_DEBUG_ANY,
 			    "Entry (%s), %s\n",
 			    e->e_dn, textbuf, 0 );
-#endif
 
 			return ret;
 		}
 	}
 
+	*text = NULL;
 	return LDAP_SUCCESS;
 }
 
@@ -570,15 +486,9 @@ oc_check_required(
 	int		i;
 	Attribute	*a;
 
-#ifdef NEW_LOGGING
-	LDAP_LOG( OPERATION, ENTRY, 
-		"oc_check_required: dn (%s), objectClass \"%s\"\n", 
-		e->e_dn, ocname->bv_val, 0 );
-#else
 	Debug( LDAP_DEBUG_TRACE,
 		"oc_check_required entry (%s), objectClass \"%s\"\n",
 		e->e_dn, ocname->bv_val, 0 );
-#endif
 
 
 	/* check for empty oc_required */
@@ -611,14 +521,9 @@ int oc_check_allowed(
 {
 	int		i, j;
 
-#ifdef NEW_LOGGING
-	LDAP_LOG( OPERATION, ENTRY, 
-		"oc_check_allowed: type \"%s\"\n", at->sat_cname.bv_val, 0, 0 );
-#else
 	Debug( LDAP_DEBUG_TRACE,
 		"oc_check_allowed type \"%s\"\n",
 		at->sat_cname.bv_val, 0, 0 );
-#endif
 
 	/* always allow objectClass attribute */
 	if ( strcasecmp( at->sat_cname.bv_val, "objectClass" ) == 0 ) {
@@ -781,6 +686,8 @@ int structural_class(
 		return LDAP_OBJECT_CLASS_VIOLATION;
 	}
 
+	*text = NULL;
+
 	return LDAP_SUCCESS;
 }
 
@@ -810,12 +717,12 @@ int mods_structural_class(
 		return LDAP_OBJECT_CLASS_VIOLATION;
 	}
 
-	if( ocmod->sml_bvalues == NULL || ocmod->sml_bvalues[0].bv_val == NULL ) {
+	if( ocmod->sml_values == NULL || ocmod->sml_values[0].bv_val == NULL ) {
 		*text = "objectClass attribute has no values";
 		return LDAP_OBJECT_CLASS_VIOLATION;
 	}
 
-	return structural_class( ocmod->sml_bvalues, sc, NULL,
+	return structural_class( ocmod->sml_values, sc, NULL,
 		text, textbuf, textlen );
 }
 
@@ -881,7 +788,7 @@ entry_naming_check(
 
 		if( desc->ad_type->sat_obsolete ) {
 			snprintf( textbuf, textlen, 
-				"naming attribute '%s' is collective",
+				"naming attribute '%s' is obsolete",
 				ava->la_attr.bv_val );
 			rc = LDAP_NAMING_VIOLATION;
 			break;
