@@ -1,8 +1,8 @@
 /* schemaparse.c - routines to parse config file objectclass definitions */
-/* $OpenLDAP: pkg/ldap/servers/slapd/schemaparse.c,v 1.64.2.3 2003/12/01 00:58:37 kurt Exp $ */
+/* $OpenLDAP: pkg/ldap/servers/slapd/schemaparse.c,v 1.71.2.1 2005/01/20 18:03:56 kurt Exp $ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 1998-2003 The OpenLDAP Foundation.
+ * Copyright 1998-2005 The OpenLDAP Foundation.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -126,8 +126,7 @@ parse_cr(
     const char	*fname,
     int		lineno,
     char	*line,
-    char	**argv
-)
+    char	**argv )
 {
 	LDAPContentRule *cr;
 	int		code;
@@ -165,8 +164,7 @@ parse_oc(
     const char	*fname,
     int		lineno,
     char	*line,
-    char	**argv
-)
+    char	**argv )
 {
 	LDAPObjectClass *oc;
 	int		code;
@@ -247,8 +245,7 @@ parse_at(
     const char	*fname,
     int		lineno,
     char	*line,
-    char	**argv
-)
+    char	**argv )
 {
 	LDAPAttributeType *at;
 	int		code;
