@@ -1,7 +1,7 @@
-/* $OpenLDAP: pkg/ldap/servers/slapd/slapi/printmsg.c,v 1.7.2.1 2003/12/01 00:58:45 kurt Exp $ */
+/* $OpenLDAP: pkg/ldap/servers/slapd/slapi/printmsg.c,v 1.13.2.1 2005/01/20 18:04:05 kurt Exp $ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 2002-2003 The OpenLDAP Foundation.
+ * Copyright 2002-2005 The OpenLDAP Foundation.
  * Portions Copyright 1997,2002-2003 IBM Corporation.
  * All rights reserved.
  *
@@ -40,7 +40,7 @@ char			*slapi_log_file = NULL;
 int			slapi_log_level = SLAPI_LOG_PLUGIN;
 
 int 
-vLogError(
+slapi_int_log_error(
 	int		level, 	
 	char		*subsystem, 
 	char		*fmt, 
